/// \file
/// \author Matthew Allen
/// \brief Growable, type-safe array.

#ifndef _GARRAY_H_
#define _GARRAY_H_

#include <stdlib.h>
#include <assert.h>
#include "LgiDefs.h"

#define GARRAY_MIN_SIZE			16

/// \brief Growable type-safe array.
///
/// You can store simple objects inline in this array, but all their contents are initialized 
/// to the octet 0x00. Which limits use to objects that don't have a virtual table and
/// don't need construction (constructors are not called).
///
/// The objects are copied around during use so you can't assume their pointer
/// will remain the same over time either. However when objects are deleted from the
/// array their destructors WILL be called. This allows you to have simple objects that
/// have dynamically allocated pointers that are freed properly. A good example of this
/// type of object is the GVariant class.
///
/// If you want to store objects with a virtual table, or that need their constructor
/// to be called then you should create the GArray with pointers to the objects instead
/// of inline objects. And to clean up the memory you can call GArray::DeleteObjects or
/// GArray::DeleteArrays.
template <class Type>
class GArray
{
	Type *p;
	uint32 len;
	uint32 alloc;

public:
	/// Constructor
	GArray(int PreAlloc = 0)
	{
		p = 0;
		len = 0;
		alloc = PreAlloc;
		if (alloc)
		{
			int Bytes = sizeof(Type) * alloc;
			p = (Type*) malloc(Bytes);
			if (p)
			{
				memset(p, 0, Bytes);
			}
			else
			{
				alloc = 0;
			}
		}
	}

	/// Destructor	
	~GArray()
	{
		Length(0);
	}	
	
	/// Returns the number of used entries
	int Length()
	{
		return len;
	}

	/// Sets the length of available entries
	bool Length(uint32 i)
	{
		if (i > 0)
		{
			uint nalloc = i + GARRAY_MIN_SIZE;
			if (nalloc != alloc)
			{
				Type *np = (Type*)malloc(sizeof(Type) * nalloc);
				if (NOT np)
				{
					return false;
				}

				if (p)
				{
					// copy across common elements
					memcpy(np, p, min(len, i) * sizeof(Type));
					free(p);
				}
				p = np;
				alloc = nalloc;
			}

			if (i > len)
			{
				// zero new elements
				memset(p + len, 0, sizeof(Type) * (i - len));
			}

			len = i;
		}
		else
		{
			if (p)
			{
				int Length = len;
				for (uint i=0; i<Length; i++)
				{
					p[i].~Type();
				}
				free(p);
				p = 0;
			}
			len = alloc = 0;
		}

		return true;
	}

	GArray<Type> &operator =(GArray<Type> &a)
	{
		Length(a.Length());
		if (p AND a.p)
		{
			memcpy(p, a.p, sizeof(Type) * len);
		}
		return *this;
	}
	
	/// \brief Returns a reference a given entry.
	///
	/// If the entry is off the end of the array, it will grow to make it valid.
	Type &operator [](uint32 i)
	{
		static Type *t = 0;
		if (i < 0)
		{
			return *t;
		}
		
		if (i > 5000000)
		{
			#ifdef _DEBUG
			#ifdef _MSC_VER
			_asm int 3
			#else
			assert(0);
			#endif
			#endif
			
			return *t;
		}

		if (i >= alloc)
		{
			// increase array length
			uint nalloc = max(alloc, GARRAY_MIN_SIZE);
			while (nalloc <= i)
			{
				nalloc <<= 1;
			}

			// alloc new array
			Type *np = (Type*) malloc(sizeof(Type) * nalloc);
			if (np)
			{
				// clear new cells
				memset(np + len, 0, (nalloc - len) * sizeof(Type));
				if (p)
				{
					// copy across old cells
					memcpy(np, p, len * sizeof(Type));
					
					// clear old array
					free(p);
				}
				
				// new values
				p = np;
				alloc = nalloc;
			}
			else
			{
				static Type *t = 0;
				return *t;
			}
		}

		// adjust length of the the array
		if (i + 1 > len)
		{
			len = i + 1;
		}
		
		return p[i];
	}
	
	/// Delete all the entries as if they are pointers to objects
	void DeleteObjects()
	{
		for (uint i=0; i<len; i++)
		{
			DeleteObj(p[i]);
		}
		Length(0);
	}
	
	/// Delete all the entries as if they are pointers to arrays
	void DeleteArrays()
	{
		for (int i=0; i<len; i++)
		{
			DeleteArray(p[i]);
		}
		Length(0);
	}
	
	/// Find the index of entry 'n'
	int IndexOf(Type n)
	{
		for (uint i=0; i<len; i++)
		{
			if (p[i] == n) return i;
		}
		
		return -1;
	}

	/// Returns true if the item 'n' is in the array
	bool HasItem(Type n)
	{
		return IndexOf(n) >= 0;
	}
	
	/// Deletes an entry
	bool DeleteAt
	(
		/// The index of the entry to delete
		uint Index,
		/// true if the order of the array matters, otherwise false.
		bool Ordered = false
	)
	{
		if (p AND Index >= 0)
		{
			if (Index < len - 1)
			{
				if (Ordered)
				{
					memmove(p + Index, p + Index + 1, (len - Index - 1) * sizeof(Type) );
				}
				else
				{
					p[Index] = p[len-1];
				}
			}
			len--;
			return true;
		}
		
		return false;
	}

	/// Deletes the entry 'n'
	bool Delete
	(
		/// The value of the entry to delete
		Type n,
		/// true if the order of the array matters, otherwise false.
		bool Ordered = false
	)
	{
		int i = IndexOf(n);
		if (p AND i >= 0)
		{
			return DeleteAt(i, Ordered);
		}
		
		return false;
	}

	/// Appends an element
	void Add
	(
		/// Item to insert
		Type n
	)
	{
		(*this)[len] = n;
	}

	/// Inserts an element into the array
	bool AddAt
	(
		/// Item to insert before
		int Index,
		/// Item to insert
		Type n
	)
	{
		// Make room
		if (Length(len + 1))
		{
			// Shift elements after insert point up one
			memmove(p + Index + 1, p + Index, (len - Index - 1) * sizeof(Type) );

			// Insert item
			p[Index] = n;

			return true;
		}

		return false;
	}

	/// Sorts the array
	void Sort(int (*Compare)(Type*, Type*))
	{
		typedef int (*qsort_compare)(const void *, const void *);
		qsort(p, len, sizeof(Type), (qsort_compare)Compare);
	}
};

#endif
