#ifndef __IFTP_H
#define __IFTP_H

#include "INet.h"
#include "GDateTime.h"

#define IFTP_READ			0x0400
#define IFTP_WRITE			0x0200
#define IFTP_EXECUTE		0x0100
#define IFTP_DIR			0x0800

#define IFTP_GRP_READ		0x0040
#define IFTP_GRP_WRITE		0x0020
#define IFTP_GRP_EXECUTE	0x0010

#define IFTP_GLOB_READ		0x0004
#define IFTP_GLOB_WRITE		0x0002
#define IFTP_GLOB_EXECUTE	0x0001

#define IFTP_ARCHIVE		0x1000
#define IFTP_HIDDEN			0x2000
#define IFTP_SYSTEM			0x4000

#define IFTP_SYM_LINK		0x8000

// Options
#define OPT_LogOpen			"LogOpen"

// Messages
#define M_CREATE_FOLDER		(M_USER+300)
#define M_DELETE			(M_USER+301)
#define M_RENAME			(M_USER+302)
#define M_PERMISSIONS		(M_USER+303)
#define M_RUN				(M_USER+304)
#define M_OPEN_AS_TEXT		(M_USER+305)
#define M_TRANSFER			(M_USER+306)
#define M_REFRESH			(M_USER+307)
#define M_UP				(M_USER+308)

// Classes
enum FtpOpenStatus
{
	FO_ConnectFailed,
	FO_LoginFailed,
	FO_Error,
	FO_Connected
};

class IFtpDataPipe
{
public:
	virtual int WritePipe(GStream *c, char *Data, int Len, int Flags) = 0;
	virtual int ReadPipe(GStream *c, char *Data, int Len, int Flags) = 0;
};

class IFtpEntry
{
public:
	int Attributes;
	int64 Size;
	char *Name;
	char *Path;
	char *User;
	char *Group;
	GDateTime Date;

	IFtpEntry();
	IFtpEntry(char *Entry);
	IFtpEntry(IFtpEntry *Entry);
	virtual ~IFtpEntry();

	bool IsDir() { return (Attributes & IFTP_DIR) != 0; }
	bool IsHidden() { return TestFlag(Attributes, IFTP_HIDDEN); }
};

class IFileProtocol
{
public:
	virtual ~IFileProtocol() {}

	// Properties
	virtual bool IsForceActive() = 0;
	virtual void IsForceActive(bool i) = 0;
	virtual bool IsLongList() = 0;
	virtual void IsLongList(bool i) = 0;
	virtual bool IsShowHidden() = 0;
	virtual void IsShowHidden(bool i) = 0;
	virtual Progress *GetMeter() = 0;
	virtual void SetMeter(Progress *m) = 0;
	virtual bool GetAuthed() = 0;

	// Data
	virtual GSocketI *Handle() = 0;
	
	// Connection
	virtual FtpOpenStatus Open(GSocketI *S, char *RemoteHost, int Port, char *User, char *Password) = 0;
	virtual bool Close() = 0;
	virtual bool IsOpen() = 0;
	virtual void Noop() = 0;

	// Directory
	virtual bool GetDir(char *Dir) = 0;
	virtual bool SetDir(char *Dir) = 0;
	virtual bool CreateDir(char *Dir) = 0;
	virtual bool DeleteDir(char *Dir) = 0;
	virtual bool ListDir(List<IFtpEntry> *Dir) = 0;
	virtual bool UpDir() = 0;
	
	// File
	virtual bool DeleteFile(char *Remote) = 0;
	virtual bool DownloadFile(char *Local, IFtpEntry *Remote, bool Binary = true) = 0;
	virtual bool UploadFile(char *Local, char *Remote, bool Binary = true) = 0;
	virtual bool RenameFile(char *From, char *To) = 0;
	virtual bool SetPerms(char *File, int Perms) = 0;
	virtual bool ResumeAt(int64 Pos) = 0;
	virtual void Abort() = 0;
};

class IFtp : public IFileProtocol
{
	class IFtpPrivate *d;

	// Sockets...
	GSocketI *Socket;	// commands

	int WriteLine(char *Msg = 0);
	int ReadLine(char *Msg = 0, bool Debug = false);
	bool TransferFile(char *Local, char *Remote, int64 RemoteSize, bool Upload, bool Binary);

	// Data connections
	char Ip[64];
	int Port;
	bool PassiveMode;
	bool ForceActive;
	bool LongList;
	bool ShowHidden;
	IFtpDataPipe *Pipe;
	Progress *Meter;
	bool Authenticated;

	// State
	int RestorePos;
	bool AbortTransfer;
	
	bool SetupData(bool Binary);
	bool ConnectData();

public:

	IFtp(IFtpDataPipe *OptionalPipe = 0);
	virtual ~IFtp();

	// Properties
	bool IsForceActive() { return ForceActive; }
	void IsForceActive(bool i) { ForceActive = i; }
	bool IsLongList() { return LongList; }
	void IsLongList(bool i) { LongList = i; }
	bool IsShowHidden() { return ShowHidden; }
	void IsShowHidden(bool i) { ShowHidden = i; }
	Progress *GetMeter() { return Meter; }
	void SetMeter(Progress *m) { Meter = m; }
	bool GetAuthed() { return Authenticated; }

	// Data
	GSocketI *Handle() { return Socket; }
	
	// Connection
	FtpOpenStatus Open(GSocketI *S, char *RemoteHost, int Port, char *User, char *Password);
	bool Close();
	bool IsOpen();
	void Noop();

	// Directory
	bool GetDir(char *Dir);
	bool SetDir(char *Dir);
	bool CreateDir(char *Dir);
	bool DeleteDir(char *Dir);
	bool ListDir(List<IFtpEntry> *Dir);
	bool UpDir();
	
	// File
	bool DeleteFile(char *Remote);
	bool DownloadFile(char *Local, IFtpEntry *Remote, bool Binary = true);
	bool UploadFile(char *Local, char *Remote, bool Binary = true);
	bool RenameFile(char *From, char *To);
	bool SetPerms(char *File, int Perms);
	bool ResumeAt(int64 Pos);
	void Abort() { AbortTransfer = true; }
};

#endif
