/**		\file
		\author Matthew Allen
*/

#ifndef __MAIL_H
#define __MAIL_H

#include "INet.h"
#include "Base64.h"
#include "Progress.h"
#include "GVariant.h"

// Defines
#define MAX_LINE_SIZE						1024
#define MAX_NAME_SIZE						64
#define EMAIL_LINE_SIZE						76

#define IsDigit(c) ((c) >= '0' AND (c) <= '9')

// MIME content types
#define CONTENT_NONE						0
#define CONTENT_BASE64						1
#define CONTENT_QUOTED_PRINTABLE			2
#define CONTENT_OCTET_STREAM				3

// Mail logging defines
#define MAIL_SEND_COLOUR					Rgb24(0, 0, 0xff)
#define MAIL_RECEIVE_COLOUR					Rgb24(0, 0x8f, 0)
#define MAIL_ERROR_COLOUR					Rgb24(0xff, 0, 0)
#define MAIL_WARNING_COLOUR					Rgb24(0xff, 0x7f, 0)
#define MAIL_INFO_COLOUR					Rgb24(0, 0, 0)

// Helper functions
extern void TokeniseStrList(char *Str, List<char> &Output, char *Delim);
extern char ConvHexToBin(char c);
#define ConvBinToHex(i) (((i)<10)?'0'+(i):'A'+(i)-10)
extern void DecodeAddrName(char *Start, char *&Name, char *&Addr, char *DefaultDomain);
extern char *DecodeRfc2047(char *Str);
extern char *EncodeRfc2047(char *Str, char *CodePage, List<char> *CharsetPrefs);
extern char *DecodeBase64Str(char *Str, int Len = -1);
extern char *DecodeQuotedPrintableStr(char *Str, int Len = -1);
extern bool Is8Bit(char *Text);
extern int MaxLineLen(char *Text);
extern char *EncodeImapString(char *s);
extern char *DecodeImapString(char *s);

// Classes
class MailProtocol;

class MailProtocolProgress
{
public:
	int Start;
	int Value;
	int Range;

	MailProtocolProgress()
	{
		Empty();
	}

	void Empty()
	{
		Start = 0;
		Value = 0;
		Range = 0;
	}
};

class LogEntry
{
public:
	char *Text;
	COLOUR c;

	LogEntry(char *t, COLOUR col = 0)
	{
		c = col;
		Text = 0;
		
		if (t)
		{
			char *n = strchr(t, '\r');
			if (n)
			{
				Text = NewStr(t, (int)n-(int)t);
			}
			else
			{
				Text = NewStr(t);
			}
		}
	}

	~LogEntry()
	{
		DeleteArray(Text);
	}
};

/// Attachment descriptor
class FileDescriptor : public GObject
{
protected:
	// Global
	int64 Size;
	char *MimeType;

	// Read from file
	GFile File;
	GFile *Embeded;
	bool OwnEmbeded;
	int64 Offset;
	GSemaphore *Lock;

	// Write to memory
	uchar *Data;

public:
	FileDescriptor(GFile *embed, int64 Offset, int64 Size, char *Name);
	FileDescriptor(char *name);
	FileDescriptor(char *data, int64 len);
	FileDescriptor();
	~FileDescriptor();

	void SetLock(GSemaphore *l);
	GSemaphore *GetLock();
	void SetOwnEmbeded(bool i);

	// Access functions
	GFile *GotoObject();		// Get data to read
	uchar *GetData();			// Get data from write
	int Sizeof();
	char *GetMimeType() { return MimeType; }
	void SetMimeType(char *s) { DeleteArray(MimeType); MimeType = NewStr(s); }

	// Decode MIME data to memory
	bool Decode(char *ContentType,
				char *ContentTransferEncoding,
				char *MimeData,
				int MimeDataLength);
};

#define MAIL_ADDR_TO				0
#define MAIL_ADDR_CC				1
#define MAIL_ADDR_BCC				2
#define MAIL_ADDR_FROM				3

/// Address dscriptor
class AddressDescriptor : public GObject, public GDom
{
public:
	uint8 Status;
	uchar CC;	// MAIL_ADDR_??
	char *Name;
	char *Addr;
	
	void *Data; // application defined

	AddressDescriptor(AddressDescriptor *Copy = 0);
	~AddressDescriptor();
	void _Delete();

	void Print(char *Str);
	int Sizeof()
	{
		return	SizeofStr(Name) +
				SizeofStr(Addr);
	}

	bool Serialize(GFile &f, bool Write)
	{
		bool Status = true;
		if (Write)
		{
			WriteStr(f, Name);
			WriteStr(f, Addr);
		}
		else
		{
			DeleteArray(Name);
			Name = ReadStr(f PassDebugArgs);
			DeleteArray(Addr);
			Addr = ReadStr(f PassDebugArgs);
		}
		return Status;
	}
};

// FIELD_PRIORITY is equivilent to the header field: X-Priority
//	1 - High
//	...
//	5 - Low
#define MAIL_PRIORITY_HIGH				1
#define MAIL_PRIORITY_NORMAL			3 // ???
#define MAIL_PRIORITY_LOW				5

class MailMessage
{
	char*						Text;
	char*						TextCharset;
	
	char*						Html;
	char*						HtmlCharset;

public:
	List<AddressDescriptor>		To;
	AddressDescriptor			*From;
	AddressDescriptor			*Reply;
	char*						Subject;
	char*						MessageID;
	List<FileDescriptor>		FileDesc;
	char*						InternetHeader;
	char						Priority;
	int							MarkColour;
	uint8						DispositionNotificationTo; // read receipt
	char*						References;
	
	// Protocol specific
	GObject						*Private;

	// Class
	MailMessage();
	virtual ~MailMessage();
	void Empty();

	virtual char *GetBody();
	virtual bool SetBody(char *Txt, int Bytes = -1, bool Copy = true, char *Cs = 0);
	virtual char *GetBodyCharset();
	virtual bool SetBodyCharset(char *Cs, bool Copy = true);

	virtual char *GetHtml();
	virtual bool SetHtml(char *Txt, int Bytes = -1, bool Copy = true, char *Cs = 0);
	virtual char *GetHtmlCharset();
	virtual bool SetHtmlCharset(char *Cs, bool Copy = true);
	
	// Conversion to/from MIME
	GStringPipe					*Raw;

	bool Encode(GStream &Out, GStream *HeadersSink, MailProtocol *Protocol, bool Mime = true);
	bool EncodeHeaders(GStream &Out, MailProtocol *Protocol, bool Mime = true);
	bool EncodeBody(GStream &Out, MailProtocol *Protocol, bool Mime = true);
	bool EncodeQuotedPrintable(GStream &Out, char *In);
	bool EncodeText(GStream &Out, char *In);
};

/// Base class for mail protocol implementations
class MailProtocol
{
protected:
	char Buffer[4<<10];
	GSocketI *Socket;

	bool Read();
	bool Write(char *Buf = NULL, bool Log = false);
	
	virtual void OnUserMessage(char *Str) {}

public:
	// Logging
	GWindow *LogNotify;
	int LogMsg;
	int LogData; // Opaque MsgA() parameter passed back to app
	void Log(char *Str, COLOUR c);
	void Log(LogEntry *e);

	// Task Progress
	MailProtocolProgress *Items;
	MailProtocolProgress *Transfer;

	// Settings
	char *ServerMsg;
	char *ProgramName;
	char *DefaultDomain;
	char *ExtraOutgoingHeaders;
	List<char> CharsetPrefs;

	// Object
	MailProtocol();
	virtual ~MailProtocol();

	// Methods
	GSocketI *GetSocket() { return Socket; }
};

/////////////////////////////////////////////////////////////////////
// Mail IO parent classes

/// Enable STARTTLS support (requires an SSL capable socket)
#define MAIL_USE_STARTTLS			0x01
/// Use authentication
#define MAIL_USE_AUTH				0x02
/// Force the use of PLAIN type authentication
#define MAIL_USE_PLAIN				0x04
/// Force the use of LOGIN type authentication
#define MAIL_USE_LOGIN				0x08
/// Force the use of NTLM type authentication
#define MAIL_USE_NTLM				0x10

/// Mail sending protocol
class MailSink : public MailProtocol
{
public:
	/// Connection setup/shutdown
	virtual bool Open
	(
		/// The transport layer to use
		GSocketI *S,
		/// The host to connect to
		char *RemoteHost,
		/// The local domain
		char *LocalDomain,
		/// The sink username (or NULL)
		char *UserName,
		/// The sink password (or NULL)
		char *Password,
		/// The port to connect with or 0 for default (25)
		int Port,
		/// Options: Use any of #MAIL_SINK_STARTTLS, #MAIL_SINK_AUTH, #MAIL_SINK_USE_PLAIN, #MAIL_SINK_USE_LOGIN or'd together.
		int Flags
	) = 0;
	/// Close the connection
	virtual bool Close() = 0;

	// Commands available while connected
	
	/// Write the email's contents into the GStringPipe returned from
	/// SendStart and then call SendEnd to finish the transaction
	virtual GStringPipe *SendStart(List<AddressDescriptor> &To, AddressDescriptor *From) = 0;
	
	/// Finishes the mail send
	virtual bool SendEnd(GStringPipe *Sink) = 0;

	// Deprecated
	virtual bool Send(MailMessage *Msg, bool Mime) { return false; }
};

class MailTransaction
{
public:
	int Index;
	int Flags;
	bool Delete;
	bool Status;
	bool Oversize;
	GStream *Stream;

	MailTransaction();
	~MailTransaction();
};

/// Return code from MailSrcCallback
enum MailSrcStatus
{
	/// Download the whole email
	DownloadAll,
	/// Download just the top part
	DownloadTop,
	/// Skip this email
	DownloadNone,
	/// About the whole receive
	DownloadAbort
};

/// The callback function used by MailSource::Receive
typedef MailSrcStatus (*MailSrcCallback)
(
	/// The currently executing transaction
	MailTransaction *Trans,
	/// The size of the email about to be downloaded
	int64 Size,
	/// If DownloadTop is returned, you can set the number of lines to retreive here
	int *LinesToDownload,
	/// The data cookie passed into MailSource::Receive
	int Data
);

/// The callback function used by MailSource::Receive
typedef bool (*MailReceivedCallback)
(
	/// The currently executing transaction
	MailTransaction *Trans,
	/// The data cookie passed into MailSource::Receive
	int Data
);

/// Collection of callbacks called during mail receive. You should zero this
/// entire object before using it. Because if someone adds new callbacks after
/// you write the calling code you wouldn't want to leave some callbacks un-
/// initialized. A NULL callback is ignored.
struct MailCallbacks
{
	/// The callback data
	int CallbackData;
	/// Called before receiving mail
	MailSrcCallback OnSrc;
	/// Called after mail received
	MailReceivedCallback OnReceive;
};

/// Enable STARTTLS support (requires an SSL capable socket)
#define MAIL_SOURCE_STARTTLS			0x01
/// Use authentication
#define MAIL_SOURCE_AUTH				0x02
/// Force the use of PLAIN type authentication
#define MAIL_SOURCE_USE_PLAIN			0x04
/// Force the use of LOGIN type authentication
#define MAIL_SOURCE_USE_LOGIN			0x08

/// A generic mail source object
class MailSource : public MailProtocol
{
public:
	/// Opens a connection to the server
	virtual bool Open
	(
		/// The transport socket
		GSocketI *S,
		/// The hostname or IP of the server
		char *RemoteHost,
		/// The port on the host to connect to
		int Port,
		/// The username for authentication
		char *User,
		/// The password for authentication
		char *Password,
		/// A cookie that the implementation can store things in, which persists across sessions. (Dynamically allocated string)
		char *&Cookie,
		/// Any optional flags: #MAIL_SOURCE_STARTTLS, #MAIL_SOURCE_AUTH, #MAIL_SOURCE_USE_PLAIN, #MAIL_SOURCE_USE_LOGIN
		int Flags = 0) = 0;
	/// Closes the connection
	virtual bool Close() = 0;

	/// Returns the number of messages available on the server
	virtual int GetMessages() = 0;
	/// Receives a list of messages from the server.
	virtual bool Receive
	(
		/// An array of messages to receive. The MailTransaction objects contains the index of the message to receive
		/// and various status values returned after the operation.
		GArray<MailTransaction*> &Trans,
		/// An optional set of callback functions.
		MailCallbacks *Callbacks = 0
	) = 0;
	/// Deletes a message on the server
	virtual bool Delete(int Message) = 0;
	/// Gets the size of the message on the server
	virtual int Sizeof(int Message) = 0;
	/// Gets the size of all the messages on the server
	virtual bool GetSizes(GArray<int> &Sizes) { return false; }
	/// Gets the unique identifier of the message
	virtual bool GetUid(int Message, char *Id) = 0;
	/// Gets the unique identifiers of a list of messages
	virtual bool GetUidList(List<char> &Id) = 0;
	/// Gets the headers associated with a given message
	virtual char *GetHeaders(int Message) = 0;
	/// Sets the proxy server. e.g. HTTP mail.
	virtual void SetProxy(char *Server, int Port) {}
};

/////////////////////////////////////////////////////////////////////
// Mail IO implementations

/// SMTP implementation
class MailSmtp : public MailSink
{
protected:
	bool ReadReply(char *Str, GStringPipe *Pipe = 0);
	bool WriteText(char *Str);

public:
	MailSmtp();
	~MailSmtp();

	bool Open(GSocketI *S, char *RemoteHost, char *LocalDomain, char *UserName, char *Password, int Port = SMTP_PORT, int Flags = 0);
	bool Close();

	GStringPipe *SendStart(List<AddressDescriptor> &To, AddressDescriptor *From);
	bool SendEnd(GStringPipe *Sink);

	bool Send(MailMessage *Msg, bool Mime = false);
};

class MailSendFolder : public MailSink
{
	class MailPostFolderPrivate *d;

public:
	MailSendFolder(char *Path);
	~MailSendFolder();

	bool Open(GSocketI *S, char *RemoteHost, char *LocalDomain, char *UserName, char *Password, int Port = 0, int Flags = 0);
	bool Close();

	GStringPipe *SendStart(List<AddressDescriptor> &To, AddressDescriptor *From);
	bool SendEnd(GStringPipe *Sink);
};

class MailPop3 : public MailSource
{
protected:
	bool ReadReply(int Timeout = -1);
	bool ReadMultiLineReply(char *&Str);
	int GetInt();
	bool MailIsEnd(char *Ptr, int Len);

	char *End;
	char *Marker;
	int Messages;

public:
	MailPop3();
	~MailPop3();

	// Connection
	bool Open(GSocketI *S, char *RemoteHost, int Port, char *User, char *Password, char *&Cookie, int Flags = 0);
	bool Close();

	// Commands available while connected
	int GetMessages() { return Messages; }
	bool Receive(GArray<MailTransaction*> &Trans, MailCallbacks *Callbacks = 0);
	bool Delete(int Message);
	int Sizeof(int Message);
	bool GetSizes(GArray<int> &Sizes);
	bool GetUid(int Message, char *Id);
	bool GetUidList(List<char> &Id);
	char *GetHeaders(int Message);
};

class MailReceiveFolder : public MailSource
{
protected:
	class MailReceiveFolderPrivate *d;

public:
	MailReceiveFolder(char *Path);
	~MailReceiveFolder();

	// Connection
	bool Open(GSocketI *S, char *RemoteHost, int Port, char *User, char *Password, char *&Cookie, int Flags = 0);
	bool Close();

	// Commands available while connected
	int GetMessages();
	bool Receive(GArray<MailTransaction*> &Trans, MailCallbacks *Callbacks = 0);
	bool Delete(int Message);
	int Sizeof(int Message);
	bool GetUid(int Message, char *Id);
	bool GetUidList(List<char> &Id);
	char *GetHeaders(int Message);
};

class MailPhp : public MailSource
{
protected:
	class MailPhpPrivate *d;

	bool Get(GSocketI *S, char *Uri, GStream &Out, bool ChopDot);

public:
	MailPhp();
	~MailPhp();

	// Connection
	bool Open(GSocketI *S, char *RemoteHost, int Port, char *User, char *Password, char *&Cookie, int Flags = 0);
	bool Close();

	// Commands available while connected
	int GetMessages();
	bool Receive(GArray<MailTransaction*> &Trans, MailCallbacks *Callbacks = 0);
	bool Delete(int Message);
	int Sizeof(int Message);
	bool GetSizes(GArray<int> &Sizes);
	bool GetUid(int Message, char *Id);
	bool GetUidList(List<char> &Id);
	char *GetHeaders(int Message);
	void SetProxy(char *Server, int Port);
};

class MailImapMsg : public GObject
{
public:
	uint ImapSeen:1;
	uint ImapRecent:1;
	uint ImapDeleted:1;
	uint ImapExpunged:1;
	int Index;

	MailImapMsg()
	{
		ImapSeen = 0;
		ImapRecent = 0;
		ImapDeleted = 0;
		ImapExpunged = 0;
		Index = -1;
	}
};

class MailImapFolder
{
	friend class MailIMap;

	char Sep;
	char *Path;

public:
	bool NoSelect;
	bool NoInferiors;
	bool Marked;
	int Exists;
	int Recent;
	int Deleted;
	int UnseenIndex;

	MailImapFolder();
	virtual ~MailImapFolder();

	char *GetPath();
	void SetPath(char *s);
	char *GetName();
	void SetName(char *s);
};

class MailIMap : public MailSource, public GSemaphore
{
protected:
	class MailIMapPrivate *d;

	GSocketI *S;
	char Buf[1024];
	List<char> Dialog;
	List<char> Uid;

	void ClearDialog();
	void ClearUid();
	bool FillUidList();
	bool WriteBuf(bool ObsurePass = false);
	bool ReadResponse(int Cmd = -1, GStringPipe *Out = 0);
	bool ReadLine(char *Buf, int Size);
	bool GetParts(int Message, GStream &Out, char *Parts, char **Flags = 0);

public:
	MailIMap();
	~MailIMap();

	// Properties
	char GetFolderSep();
	char *GetCurrentPath();
	bool GetExpungeOnExit();
	void SetExpungeOnExit(bool b);
	bool ServerOption(char *Opt);

	// Connection
	bool Open(GSocketI *S, char *RemoteHost, int Port, char *User, char *Password, char *&Cookie, int Flags = 0);
	bool Close();

	// Commands available while connected
	bool Receive(GArray<MailTransaction*> &Trans, MailCallbacks *Callbacks = 0);
	bool ReceiveFlags(MailMessage *Msg);
	bool Send(MailImapFolder *Folder, char *Msg, int Index = -1, bool Seen = true);

	int GetMessages();
	bool Delete(int Message);
	int Sizeof(int Message);
	bool GetUid(int Message, char *Id);
	bool GetUidList(List<char> &Id);
	char *GetHeaders(int Message);

	// Imap specific commands
	bool GetFolders(List<MailImapFolder> &Folders, char *Path = 0);
	bool SelectFolder(char *Path, MailImapFolder *f = 0);
	char *GetSelectedFolder();
	int GetMessages(char *Path);
	bool CreateFolder(MailImapFolder *f);
	bool DeleteFolder(char *Path);
	bool RenameFolder(char *From, char *To);
	bool SetFolderFlags(MailImapFolder *f);
	bool ExpungeFolder(MailImapFolder *f);

	/// Idle processing...
	/// \returns true if something happened
	bool OnIdle(int *Exists, int *Recent, GArray<MailImapMsg> *Msg);
	bool Poll(int *Recent = 0);
};

#endif
