/// \file
/// \author Matthew Allen
///
#include <stdio.h>
#include <math.h>

#include "Lgi.h"
#include "GScripting.h"
#include "GLexCpp.h"

extern GHostFunc SystemLibrary[];

#define ThisToken()		(Cur < Tokens.Length() ? Tokens[Cur] : 0)
#define NextToken()		(Cur < Tokens.Length() - 1 ? Tokens[Cur + 1] : 0)
#define IsDigit(c)		( ((c)>='0') AND ((c)<='9')  )

char16 sChar[]		= { 'c','h','a','r', 0 };
char16 sInt[]		= { 'i','n','t', 0 };
char16 sUInt[]		= { 'u','i','n','t', 0 };
char16 sInt32[]		= { 'i','n','t','3','2', 0 };
char16 sUInt32[]	= { 'u','i','n','t','3','2', 0 };
char16 sInt64[]		= { 'i','n','t','6','4', 0 };
char16 sHWND[]		= { 'H','W','N','D', 0 };
char16 sDWORD[]		= { 'D','W','O','R','D', 0 };
char16 sLPTSTR[]	= { 's','L','P','T','S','T','R', 0 };
char16 sLPCTSTR[]	= { 's','L','P','C','T','S','T','R', 0 };
char16 sElse[]		= { 'e','l','s','e', 0 };

char16 sPeriod[] = { '.', 0 };
char16 sComma[] = { ',', 0 };
char16 sSemiColon[] = { ';',0 };
char16 sStartRdBracket[] = { '(', 0 };
char16 sEndRdBracket[] = { ')', 0 };
char16 sStartSqBracket[] = { '[', 0 };
char16 sEndSqBracket[] = { ']', 0 };
char16 sStartCurlyBracket[] = { '{', 0 };
char16 sEndCurlyBracket[] = { '}', 0 };


enum Operator
{
	OpNull,
	OpAssign,
	OpPlus,
	OpMinus,
	OpMul,
	OpDiv,
	OpMod,
	OpLessThan,
	OpLessThanEqual,
	OpGreaterThan,
	OpGreaterThanEqual,
	OpEquals,
	OpNotEquals,
	OpPlusEquals,
	OpMinusEquals,
	OpMulEquals,
	OpDivEquals,
	OpInc,
	OpDec,
	OpAnd,
	OpOr,
};

enum GStatementType
{
	GS_UNKNOWN,
	GS_IF,
	GS_FOR,
	GS_FUNCTION_DEFN,
	GS_FUNCTION_CALL,
	GS_LVALUE,
};

/// Executable statement
struct GCode
{
	/// The type of statement
	GStatementType	Type;
	/// The start token for the statement
	int				Start;
	/// Child code element
	GArray<GCode>	Code;

	union
	{
		struct
		{
			/// Index of the statement executed at the start of the loop
			GArray<GCode> *PreStatement;
			/// The index of the condition expression
			int Condition;
			/// Index of the statement to execute after the loop body is executed
			GArray<GCode> *PostStatement;

		}	For;

		struct
		{
			/// Index of the expression to evaluate
			int Condition;
			/// The code inside the 'else' part of the statement
			GArray<GCode> *Else;

		}	If;

		struct
		{
			/// The command to call
			GFunc *Method;
			/// Indexes to the start of each argument expression
			GArray<int> *Args;

		}	MethodCall;

		struct
		{
			/// The operator after the L-value
			Operator Op;
			/// The index of the expression
			int Expression;

		}	LValue;
	};

	~GCode()
	{
		switch (Type)
		{
			case GS_IF:
			{
				DeleteObj(If.Else);
				break;
			}
			case GS_FOR:
			{
				DeleteObj(For.PreStatement);
				DeleteObj(For.PostStatement);
				break;
			}
			case GS_FUNCTION_CALL:
			{
				DeleteObj(MethodCall.Args);
				break;
			}
			case GS_LVALUE:
			{
				break;
			}
		}
	}
};

class GExternType;
class GExternType
{
public:
	GVariantType Simple;
	GArray<GExternType*> Struct;

	int Sizeof()
	{
		switch (Simple)
		{
			case GV_INT32:
				return sizeof(int32);
			case GV_INT64:
				return sizeof(int64);
			case GV_STRING:
				return sizeof(char*);
			case GV_DOUBLE:
				return sizeof(double);
			case GV_NULL:
			{
				int s = 0;
				for (int i=0; i<Struct.Length(); i++)
				{
					s += Struct[i]->Sizeof();
				}
				return s;
			}
		}

		LgiAssert(0);
		return 0;
	}

	GExternType()
	{
		Simple = GV_NULL;
	}
};

struct GExternFunc : public GFunc
{
	char *Library;
	GExternType Return;
	GArray<GExternType*> Args;

	GExternFunc() : GFunc(0, ExternFunc)
	{
		Library = 0;
	}

	~GExternFunc()
	{
		DeleteArray(Library);
		Args.DeleteObjects();
	}

	bool Call(GScriptEnginePrivate *Priv, GVariant *Ret, ArgumentArray &Args);
};

struct GScriptFunc : public GFunc
{
	GCode *Body;
	GArray<char*> Args;

	GScriptFunc(char *meth = 0) : GFunc(meth, ScriptFunc)
	{
		Body = 0;
	}

	~GScriptFunc()
	{
		Args.DeleteArrays();
	}

	bool Call(GScriptEnginePrivate *Priv, GVariant *Ret, ArgumentArray &Args);
};

struct GDomPart
{
	char16 *Name;
	int Array;

	GDomPart()
	{
		Name = 0;
		Array = 0;
	}
};

struct GDomAddr
{
	GArray<GDomPart> Parts;

	char16 *operator[](int i)
	{
		if (i >= 0 AND i < Parts.Length())
		{
			return Parts[i].Name;
		}

		LgiAssert(0);
		return 0;
	}

	int Length()
	{
		return Parts.Length();
	}

	void Add(char16 *part, int array = 0)
	{
		GDomPart *p = &Parts[Parts.Length()];
		p->Name = part;
		p->Array = array;
	}
	
	void Print()
	{
		for (int i=0; i<Parts.Length(); i++)
		{
			if (i)
				printf(".");
				
			if (Parts[i].Array)
				printf("%S[%i]", Parts[i].Name, Parts[i].Array);
			else
				printf("%S", Parts[i].Name);
		}
		
		if (Parts.Length() == 0)
		{
			printf("<no parts>");
		}
	}
};

struct GDomRef
{
	GVariant *Var;
	char *Member;
	int Array; // start of expression to evaluate for array index
	bool FreeVar;

	GDomRef()
	{
		Var = 0;
		Member = 0;
		Array = 0;
		FreeVar = false;
	}

	~GDomRef()
	{
		DeleteArray(Member);
		if (FreeVar) DeleteObj(Var);
	}

	bool Get(GScriptEnginePrivate *Priv, GVariant &v);
	bool Set(GScriptEnginePrivate *Priv, GVariant &v);
};

bool InsertUtfCharacter(GVariant *Var, GVariant &v, int Idx)
{
	bool Status = false;
	uint8 *Start = (uint8*) Var->CastString();
	uint8 *Str = Start;
	if (Str)
	{
		int OldLen = strlen((char*) Str);
		int Len = OldLen;
		uint32 ch = *Str, i = 0;
		do
		{
			if (i++ == Idx)
				break;
		}
		while (ch = LgiUtf8To32(Str, Len));

		if (ch)
		{
			int NewC = 0, OldC = Len;
			uint8 *Temp = Str;
			int TempLen = Len;
			LgiUtf8To32(Temp, TempLen);
			OldC -= TempLen;

			char NewBytes[8];
			uint8 *s = (uint8*) NewBytes;
			if (v.Type == GV_STRING)
			{
				uint8 *f = (uint8*) v.Str(); // from
				if (IsUtf8_1Byte(*f))
				{
					*s++ = *f;
				}
				else if (IsUtf8_Lead(*f))
				{
					*s++ = *f++;
					while (IsUtf8_Trail(*f))
						*s++ = *f++;
				}
				NewC = s - (uint8*)NewBytes;
			}
			else
			{
				int l = sizeof(NewBytes);
				LgiUtf32To8(v.CastInt32(), s, l);
				NewC = sizeof(NewBytes) - l;
			}

			if (NewC > OldC)
			{
				// Re-alloc memory to make space for larger character
				char *Ins = NEW(char[OldLen + (NewC - OldC) + 1]);
				if (Ins)
				{
					char *i = Ins;
					memcpy(i, Start, Str - Start);
					i += Str - Start;
					memcpy(i, NewBytes, NewC);
					i += NewC;
					memcpy(i, Str + OldC, Len - OldC + 1);

					Var->Empty();
					Var->Type = GV_STRING;
					Var->Value.String = Ins;
					Status = true;
				}
			}
			else
			{
				// Just insert char in existing memory
				for (int n=0; n<NewC; n++)
					Str[n] = NewBytes[n];
				if (NewC < OldC)
					memmove(Str + NewC, Str + OldC, Len - OldC + 1);
				Status = true;
			}
		}
	}
	return Status;
}

class GScriptEnginePrivate : public GScriptUtils
{
public:
	GStringPipe Term;
	GViewI *Parent;
	char16 *Script;
	GHashTable Variables;
	GScriptContext *Context;
	GHashTable Methods;
	bool IsCompiling;

	GArray<GCode> Compiled;
	GArray<char16*> Tokens;
	GArray<int> Lines;

	GScriptEnginePrivate()
	{
		Script = 0;
		IsCompiling = false;
	}

	~GScriptEnginePrivate()
	{
		Empty();
	}

	char *ErrorTitle()
	{
		return (char*) (IsCompiling ? "CompileError" : "ExecuteError");
	}

	void Empty()
	{
		DeleteArray(Script);
		Tokens.DeleteArrays();
		Lines.Length(0);
		Compiled.Length(0);

		void *v;
		for (v = Variables.First(); v; v = Variables.Next())
		{
			GVariant *i = (GVariant*)v;
			DeleteObj(i);
		}
		Variables.Empty();

		GArray<GFunc*> Del;
		for (v = Methods.First(); v; v = Methods.Next())
		{
			GFunc *i = (GFunc*)v;
			if (i->Type != HostFunc)
			{
				Del.Add(i);
			}
		}
		for (int i=0; i<Del.Length(); i++)
		{
			Methods.Delete(Del[i]->Method);
			DeleteObj(Del[i]);
		}
	}

	int GetPrecedence(Operator o)
	{
		// http://www.cppreference.com/operator_precedence.html
		switch (o)
		{
			case OpAssign:
				return 15;

			case OpAnd:
				return 12;

			case OpOr:
				return 13;

			case OpEquals:
			case OpNotEquals:
				return 8;

			case OpLessThan:
			case OpLessThanEqual:
			case OpGreaterThan:
			case OpGreaterThanEqual:
				return 7;

			case OpPlus:
			case OpMinus:
				return 5;

			case OpMul:
			case OpDiv:
			case OpMod:
				return 4;
			
			case OpInc:
			case OpDec:
				return 1;
		}

		return -1;
	}

	Operator IsOp(char16 *s)
	{
		if (NOT s) return OpNull;

		if (s[0] != 0 AND NOT s[1])
		{
			// One character operator
			switch (*s)
			{
				case '=': return OpAssign;
				case '+': return OpPlus;
				case '-': return OpMinus;
				case '*': return OpMul;
				case '/': return OpDiv;
				case '<': return OpLessThan;
				case '>': return OpGreaterThan;
				case '%': return OpMod;
			}
		}
		else if (s[0] != 0 AND s[1] == '=' AND NOT s[2])
		{
			// 2 chars, "something" equals operator
			switch (*s)
			{
				case '!': return OpNotEquals;
				case '=': return OpEquals;
				case '<': return OpLessThanEqual;
				case '>': return OpGreaterThanEqual;

				case '+': return OpPlusEquals;
				case '-': return OpMinusEquals;
				case '*': return OpMulEquals;
				case '/': return OpDivEquals;
			}
		}
		else if (s[0] == '+' AND s[1] == '+' AND NOT s[2])
		{
			return OpInc;
		}
		else if (s[0] == '-' AND s[1] == '-' AND NOT s[2])
		{
			return OpDec;
		}
		else if (s[0] == '&' AND s[1] == '&' AND NOT s[2])
		{
			return OpAnd;
		}
		else if (s[0] == '|' AND s[1] == '|' AND NOT s[2])
		{
			return OpOr;
		}

		return OpNull;
	}

	void DumpCode(GArray<GCode> &c, int Depth = 0)
	{
		char Tabs[265];
		memset(Tabs, '\t', Depth);
		Tabs[Depth] = 0;

		for (int i=0; i<c.Length(); i++)
		{
			GCode &Code = c[i];
			switch (Code.Type)
			{
				case GS_IF:
				{
					LgiTrace("%sif (%i)\n%s{\n", Tabs, Code.If.Condition, Tabs);
					break;
				}
				case GS_FOR:
				{
					LgiTrace("%sfor (%p; %i; %p)\n%s{\n",
						Tabs,
						Code.For.PreStatement,
						Code.For.Condition,
						Code.For.PostStatement,
						Tabs);
					break;
				}
				case GS_FUNCTION_CALL:
				{
					LgiTrace("%sfunc:%s(",
						Tabs, Code.MethodCall.Method->Method);
					if (Code.MethodCall.Args)
					{
						int Len = Code.MethodCall.Args->Length();
						for (int n=0; n<Len; n++)
						{
							int Idx = (*Code.MethodCall.Args)[n];
							LgiTrace("%i%s", Idx, n<Len-1?", ":"");
						}
					}
					LgiTrace(")\n");
					break;
				}
				case GS_LVALUE:
				{
					LgiTrace("%slvalue\n", Tabs);
					break;
				}
				default:
				{
					LgiAssert(0);
					break;
				}
			}

			if (Code.Code.Length())
			{
				DumpCode(Code.Code, Depth + 1);
			}
		}
	}

	/// Adds tokens to the 'Tokens' array
	bool LexScript(char *script)
	{
		char16 *w = LgiNewUtf8To16(script);
		if (w)
		{
			int Line = 1;
			char16 *s = w, *t, *prev = w;
			while (t = LexCpp(s))
			{
				while (prev < s)
				{
					if (*prev++ == '\n')
						Line++;
				}

				Tokens.Add(t);
				Lines.Add(Line);
			}

			if (NOT Script)
			{
				Script = w;
			}
			else
			{
				DeleteArray(w);
			}

			return true;
		}

		return false;
	}

	bool CompileTo(GArray<GCode> &Compiled, char *script)
	{
		bool Status = false;

		// Lex script into tokens
		int Cur = Tokens.Length();
		if (LexScript(script))
		{
			// Do compile stage
			IsCompiling = true;
			while (Cur < Tokens.Length())
			{
				int Start = Cur;
				if (NOT Compile_Statement(Compiled, Cur))
				{
					if (Compiled.Length() > 0)
						Compiled.Length(Compiled.Length()-1);
					Status = false;
					break;
				}
				else Status = true;
			}
			IsCompiling = false;
		}

		return Status;
	}

	void ProcessArguments(ArgumentArray &Args, ArgumentArray &Mem, GArray<int> *Idx);
	int FindEndOfBlock(int Cur);
	GVariant *Var(char16 *name, bool create = true);
	bool ToVar(GVariant &v, char16 *s);
	bool Require(int &Cur, char *tok);
	bool Type(int &Cur, class GExternType &Type);
	bool CallExtern(int &Cur, uint32 &ReturnValue, class GExternFunc *Func);

	// Dom addressing
	bool CreateDomAddr(int &Cur, GDomAddr &Addr);
	bool ResolveDomAddr(class GDomRef &Ref, GDomAddr &Addr, bool Create);

	// Compiling of code	
	bool	Compile_MethodCall(int &Cur);
	bool	Compile_Expression(int &Cur, int Depth = 0);
	bool    Compile_Statement(GArray<GCode> &To, int &Cur);

	// Execution of code
	GVariant *Execute_Expression(int &Cur, GDom *Src = 0, int Depth = 0);
	bool    Execute_Statement(GArray<GCode> &To);
};

bool GDomRef::Get(GScriptEnginePrivate *Priv, GVariant &v)
{
	if (Var)
	{
		if (Member)
		{
			if (Var->Type == GV_DOM)
			{
				if (Var->Value.Dom)
				{
					GVariant *ArrVal = 0;
					if (Array)
					{
						int Cur = Array;
						ArrVal = Priv->Execute_Expression(Cur);
					}

					bool Ret = Var->Value.Dom->GetVariant(Member, v, ArrVal ? ArrVal->CastString() : 0);
					DeleteObj(ArrVal);
					return Ret;
				}
			}
			else if (Var->Type == GV_STRING)
			{
				if (stricmp(Member, "length") == 0)
				{
					v = (int) strlen(Var->Str());
					return true;
				}
			}
			else if (Var->Type == GV_DATETIME)
			{
				if (Var->Value.Date)
				{
					if (stricmp(Member, "day") == 0)
					{
						v = Var->Value.Date->Day();
					}
					else if (stricmp(Member, "month") == 0)
					{
						v = Var->Value.Date->Month();
					}
					else if (stricmp(Member, "year") == 0)
					{
						v = Var->Value.Date->Year();
					}
					else if (stricmp(Member, "hour") == 0)
					{
						v = Var->Value.Date->Hours();
					}
					else if (stricmp(Member, "minute") == 0)
					{
						v = Var->Value.Date->Minutes();
					}
					else if (stricmp(Member, "second") == 0)
					{
						v = Var->Value.Date->Seconds();
					}
					else return false;

					return true;
				}
			}
			else if (Var->Type == GV_LIST)
			{
				if (stricmp(Member, "length") == 0)
				{
					v = Var->Value.Lst->Length();
					return true;
				}
			}
			else if (Var->Type == GV_HASHTABLE)
			{
				if (stricmp(Member, "length") == 0)
				{
					v = Var->Value.Hash->Length();
					return true;
				}
			}
		}
		else
		{
			bool Status = false;

			if (Array)
			{
				int Cur = Array;
				GVariant *a = Priv->Execute_Expression(Cur);
				if (a)
				{
					if (Var->Type == GV_STRING)
					{
						int Idx = a->CastInt32();
						uint8 *Str = (uint8*) Var->CastString();
						if (Str)
						{
							int Len = strlen((char*) Str);
							int i = 0;
							uint32 ch;
							while (ch = LgiUtf8To32(Str, Len))
							{
								if (i++ == Idx)
								{
									v = (int)ch;
									Status = true;
									break;
								}
							}
						}
					}
					else if (Var->Type == GV_HASHTABLE)
					{
						char *Key = a->CastString();
						GVariant *h = (GVariant*) Var->Value.Hash->Find(Key);
						if (h)
						{
							v = *h;
						}
						else
						{
							v.Empty();
						}
						Status = true;
					}
					else if (Var->Type == GV_LIST)
					{
						int Idx = a->CastInt32();
						GVariant *h = (*Var->Value.Lst)[Idx];
						if (h)
						{
							v = *h;
							Status = true;
						}
					}
					else
					{
						Priv->Term.Print("%s:%i - Can't index into type '%i'.\n", __FILE__, __LINE__, Var->Type);
						LgiAssert(0);
					}

					DeleteObj(a);
				}
				else Priv->Term.Print("%s:%i - Can't evaluate array expression.\n", __FILE__, __LINE__);
			}
			else
			{
				v = *Var;
				Status = true;
			}

			return Status;
		}
	}

	return false;
}

bool GDomRef::Set(GScriptEnginePrivate *Priv, GVariant &v)
{
	if (Var)
	{
		if (Member)
		{
			if (Var->Type == GV_DOM)
			{
				if (Var->Value.Dom)
				{
					return Var->Value.Dom->SetValue(Member, v);
				}
			}
			else if (Var->Type == GV_STRING OR
					 Var->Type == GV_NULL)
			{
				if (stricmp(Member, "length") == 0)
				{
					Var->Empty();
					int Len = v.CastInt32();
					Var->Type = GV_STRING;
					if (Var->Value.String = NEW(char[Len+1]))
					{
						memset(Var->Value.String, ' ', Len);
						Var->Value.String[Len] = 0;
						return true;
					}
				}
			}
			else if (Var->Type == GV_DATETIME)
			{
				if (Var->Value.Date)
				{
					if (stricmp(Member, "day") == 0)
					{
						Var->Value.Date->Day(v.CastInt32());
					}
					else if (stricmp(Member, "month") == 0)
					{
						Var->Value.Date->Month(v.CastInt32());
					}
					else if (stricmp(Member, "year") == 0)
					{
						Var->Value.Date->Year(v.CastInt32());
					}
					else if (stricmp(Member, "hour") == 0)
					{
						Var->Value.Date->Hours(v.CastInt32());
					}
					else if (stricmp(Member, "minute") == 0)
					{
						Var->Value.Date->Minutes(v.CastInt32());
					}
					else if (stricmp(Member, "second") == 0)
					{
						Var->Value.Date->Seconds(v.CastInt32());
					}
					else return false;

					return true;
				}
			}
		}
		else
		{
			if (Array)
			{
				bool Status = false;

				int Cur = Array;
				GVariant *a = Priv->Execute_Expression(Cur);
				if (a)
				{
					if (Var->Type == GV_STRING)
					{
						Status = InsertUtfCharacter(Var, v, a->CastInt32());
					}
					else if (Var->Type == GV_HASHTABLE)
					{
						char *Key = a->CastString();
						GVariant *n = (GVariant*) Var->Value.Hash->Find(Key);
						if (NOT n)
						{
							Var->Value.Hash->Add(Key, n = NEW(GVariant));
						}
						if (n)
						{
							*n = v;
							Status = true;
						}
					}
					else if (Var->Type == GV_LIST)
					{
						int Idx = a->CastInt32();
						if (Idx < 0 OR Idx >= Var->Value.Lst->Length())
						{
							// Add a new element
							GVariant *h = NEW(GVariant(v));
							if (h)
							{
								Var->Value.Lst->Insert(h);
								Status = true;
							}
						}
						else
						{
							// Change an existing element
							GVariant *h = (*Var->Value.Lst)[Idx];
							if (h) *h = v;
							Status = true;
						}
					}
					else
					{
						Priv->Term.Print("%s:%i - Can't index into type '%i'.\n", __FILE__, __LINE__, Var->Type);
						LgiAssert(0);
					}

					DeleteObj(a);
				}
				else Priv->Term.Print("%s:%i - Can't evaluate array expression.\n", __FILE__, __LINE__);
			}
			else
			{
				*Var = v;
				return true;
			}
		}
	}

	return false;
}

int GScriptUtils::atoi(char16 *s)
{
	int i = 0;
	if (s)
	{
		char b[64];
		int Len = StrlenW(s) * sizeof(*s);
		int Bytes = LgiBufConvertCp(b, "utf-8", sizeof(b), (void*&)s, LGI_WideCharset, Len);
		b[Bytes/sizeof(*b)] = 0;
		i = ::atoi(b);
	}
	return i;
}

int64 GScriptUtils::atoi64(char16 *s)
{
	int64 i = 0;
	if (s)
	{
		#ifdef _MSC_VER

		i = _wtoi64(s);

		#else

		char b[64];
		int Len = StrlenW(s) * sizeof(*s);
		int Bytes = LgiBufConvertCp(b, "utf-8", sizeof(b), (void*&)s, LGI_WideCharset, Len);
		b[Bytes/sizeof(*b)] = 0;

		i = strtoll(b, 0, 10);
		
		#endif
	}
	return i;
}

double GScriptUtils::atof(char16 *s)
{
	double i = 0;
	if (s)
	{
		char b[64];
		int Len = StrlenW(s) * sizeof(*s);
		int Bytes = LgiBufConvertCp(b, "utf-8", sizeof(b), (void*&)s, LGI_WideCharset, Len);
		b[Bytes/sizeof(*b)] = 0;
		i = ::atof(b);
	}
	return i;
}

int GScriptUtils::htoi(char16 *s)
{
	int i = 0;
	if (s)
	{
		char b[64];
		int Len = StrlenW(s) * sizeof(*s);
		int Bytes = LgiBufConvertCp(b, "utf-8", sizeof(b), (void*&)s, LGI_WideCharset, Len);
		b[Bytes/sizeof(*b)] = 0;
		i = ::htoi(b);
	}
	return i;
}


#define CompToInt(v) \
	(v[0] == '0' AND v[1] == 'x') ? htoi(v + 2) : atoi(v);

GScriptEngine::GScriptEngine(GViewI *parent, GScriptContext *context)
{
	d = NEW(GScriptEnginePrivate);
	d->Parent = parent;
	d->Context = context;
	d->Script = 0;

	int i = 0;
	for (GHostFunc *cmd = SystemLibrary + i; cmd->Method; cmd++)
	{
		d->Methods.Add(cmd->Method, (GFunc*)cmd);
	}

	if (d->Context)
	{
		d->Context->SetEngine(this);
		GHostFunc *CmdTable = d->Context->GetCommands();
		if (CmdTable)
		{
			for (GHostFunc *cmd = CmdTable; cmd->Method; cmd++)
			{
				d->Methods.Add(cmd->Method, (GFunc*)cmd);
			}
		}
	}
}

GScriptEngine::~GScriptEngine()
{
	if (d->Context)
	{
		d->Context->SetEngine(0);
	}

	d->Tokens.DeleteArrays();
	DeleteObj(d);
}

GVariant *GScriptEngine::Var(char16 *name, bool create)
{
	return d->Var(name, create);
}

GStringPipe *GScriptEngine::GetTerm()
{
	return &d->Term;
}

bool GScriptEngine::CallMethod(char *Method, GVariant *Ret, ArgumentArray &Args)
{
	bool Status = false;

	if (Method)
	{
		GFunc *Func = (GFunc*) d->Methods.Find(Method);
		if (Func)
		{
			Status = Func->Call(d, Ret, Args);
		}		
	}

	return Status;
}

bool GScriptEnginePrivate::CreateDomAddr(int &Cur, GDomAddr &Addr)
{
	char16 *t = ThisToken();
	if (t AND isalpha(*t))
	{
		Addr.Add(t);
		Cur++;

		// Build DOM address
		while (t = ThisToken())
		{
			if (StricmpW(t, sPeriod) == 0)
			{
				Cur++;
				t = ThisToken();
				if (isalpha(*t))
				{
					Cur++;
					Addr.Add(t);
				}
				else
				{
					Term.Print("%s:%i - Expecting name after '.'.\n", ErrorTitle(), Lines[Cur]);
					break;
				}
			}
			else if (StricmpW(t, sStartSqBracket) == 0)
			{
				Cur++;

				int Start = Cur;
				Addr.Parts[Addr.Length()-1].Array = Start;

				if (Compile_Expression(Cur))
				{
					t = ThisToken();
					if (t AND StricmpW(t, sEndSqBracket) == 0)
					{
						Cur++;
					}
					else
					{
						Term.Print("%s:%i - Expecting ']'.\n", ErrorTitle(), Lines[Cur]);
						break;
					}
				}
				else
				{
					Term.Print("%s:%i - Expression error.\n", ErrorTitle(), Lines[Start]);
					break;
				}
			}
			else break;
		}

		return true;
	}

	return false;
}

#define DEBUG_TRACE_RESOLVEDOMADDR	0

bool GScriptEnginePrivate::ResolveDomAddr(GDomRef &Ref, GDomAddr &Addr, bool Create)
{
#if DEBUG_TRACE_RESOLVEDOMADDR
LgiTrace("Resolving '%S", Addr.Parts[0].Name);
for (int k=1; k<Addr.Parts.Length(); k++) LgiTrace(".%S", Addr.Parts[k].Name);
LgiTrace("'=");
#endif

	if (Addr.Length() == 1)
	{
		if (Ref.Var = Var(Addr[0], Create))
		{
#if DEBUG_TRACE_RESOLVEDOMADDR
LgiTrace("type=%i int=%i str='%s'\n", Ref.Var->Type, Ref.Var->CastInt32(), Ref.Var->Str());
#endif

			Ref.Array = Addr.Parts[0].Array;
		}
#if DEBUG_TRACE_RESOLVEDOMADDR
else LgiTrace("null\n");
#endif

		return Ref.Var != 0;
	}
	else if (Addr.Length() > 1)
	{
		Ref.Var = Var(Addr[0], Create);
		if (Ref.Var)
		{
			for (int i=1; i<Addr.Length() - 1; i++)
			{
				if (Ref.Var->Type == GV_DOM AND
					Ref.Var->Value.Dom)
				{
					GVariant *v = NEW(GVariant);
					if (v)
					{
						char *c = LgiNewUtf16To8(Addr[i]);
						if (c)
						{
							char *ArrayVal = 0;
							int ArrayPos = Addr.Parts[i].Array;
							if (ArrayPos)
							{
								GVariant *a = Execute_Expression(ArrayPos);
								if (a)
								{
									ArrayVal = NewStr(a->Str());
									DeleteObj(a);
								}
							}

							bool Has = Ref.Var->Value.Dom->GetVariant(c, *v, ArrayVal);
							DeleteArray(c);
							if (Has)
							{
								if (Ref.FreeVar)
									DeleteObj(Ref.Var);
								Ref.Var = v;
								Ref.FreeVar = true;
								v = 0;
							}
							else
							{
								DeleteObj(v);
								return false;
							}
						}

						DeleteObj(v);
					}
					else return false;
				}
				else return false;
			}

			GDomPart &Part = Addr.Parts[Addr.Length()-1];
			Ref.Member = LgiNewUtf16To8(Part.Name);
			Ref.Array = Part.Array;

#if DEBUG_TRACE_RESOLVEDOMADDR
if (Ref.Var)
	LgiTrace("type=%i int=%i str='%s'\n", Ref.Var->Type, Ref.Var->CastInt32(), Ref.Var->Str());
else
	LgiTrace("null\n");
#endif

			return true;
		}
	}

	return false;
}

void GScriptEngine::Empty()
{
	d->Empty();
}

bool GScriptEngine::Compile(char *Script, bool Add)
{
	bool Status = false;

	if (NOT Add)
		d->Empty();

	if (Script)
	{
		Status = d->CompileTo(d->Compiled, Script);

		#if 0
		d->DumpCode(d->Compiled);
		#endif
	}

	return Status;
}

bool GScriptEngine::Run()
{
	if (d->Compiled.Length() > 0)
	{
		// Run it
		return d->Execute_Statement(d->Compiled);
	}

	return false;
}

class ScriptTokenState
{
	GScriptEnginePrivate *d;
	int TokLen;
	int LinesLen;

public:
	ScriptTokenState(GScriptEnginePrivate *priv)
	{
		d = priv;
		TokLen = d->Tokens.Length();
		LinesLen = d->Lines.Length();
	}

	~ScriptTokenState()
	{
		for (int i=TokLen; i<d->Tokens.Length(); i++)
		{
			DeleteArray(d->Tokens[i]);
		}
		d->Tokens.Length(TokLen);
		d->Lines.Length(LinesLen);
	}
};

bool GScriptEngine::RunTemporary(char *Script)
{
	bool Status = false;

	if (Script)
	{
		ScriptTokenState State(d);

		GArray<GCode> Compiled;
		if (d->CompileTo(Compiled, Script))
		{
			if (Compiled.Length() > 0)
			{
				Status = d->Execute_Statement(Compiled);
			}
		}
	}

	return Status;
}

bool GScriptEngine::EvaluateExpression(GVariant *Result, GDom *VariableSource, char *Expression)
{
	bool Status = false;

	if (Result AND Expression)
	{
		ScriptTokenState State(d);

		int Cur = d->Tokens.Length();
		if (d->LexScript(Expression))
		{
			GVariant *v = d->Execute_Expression(Cur, VariableSource);
			if (v)
			{
				*Result = *v;
				DeleteObj(v);
				Status = true;
			}
		}
	}

	return Status;
}

void GScriptEngine::DumpVariables()
{
	d->Term.Print("Variables:\n");
	char *VarName;
	for (GVariant *v = (GVariant*)d->Variables.First(&VarName); v; v = (GVariant*)d->Variables.Next(&VarName))
	{
		switch (v->Type)
		{
			case GV_NULL:
				d->Term.Print("    %s = null\n", VarName);
				break;
			case GV_INT32:
				d->Term.Print("    %s = (int) %i\n", VarName, v->CastInt32());
				break;
			case GV_DOUBLE:
				d->Term.Print("    %s = (double) %f\n", VarName, v->CastDouble());
				break;
			case GV_STRING:
				d->Term.Print("    %s = (string) \"%s\"\n", VarName, v->Str());
				break;
			case GV_DOM:
				d->Term.Print("    %s = (GDom*) 0x%p\n", VarName, v->Value.Dom);
				break;
			case GV_VOID_PTR:
				d->Term.Print("    %s = (void*) 0x%p\n", VarName, v->Value.Ptr);
				break;
			default:
				d->Term.Print("    %s = (type %i) ?????\n", VarName, v->Type);
				break;
		}
	}
}

GVariant *GScriptEnginePrivate::Var(char16 *name, bool create)
{
	if (name)
	{
		char b[256];
		int Len = StrlenW(name) * sizeof(*name);
		int Bytes = LgiBufConvertCp(b, "utf-8", sizeof(b), (void*&)name, LGI_WideCharset, Len);
		b[Bytes/sizeof(*b)] = 0;

		GVariant *v = (GVariant*) Variables.Find(b);
		if (NOT v AND create)
		{
			v = NEW(GVariant);
			if (v)
			{
				Variables.Add(b, v);
			}
		}
		return v;
	}

	return 0;
}

bool GScriptEnginePrivate::ToVar(GVariant &v, char16 *s)
{
	if (s)
	{
		char16 *f = *s == '\"' ? s + 1 : s;
		bool IsNumber = false;
		if (IsDigit(*f))
		{
			IsNumber = true;
		}
		else if (*f == '-' OR *f == '.')
		{
			for (char16 *s = f; *s; s++)
			{
				if (IsDigit(*f))
				{
					IsNumber = true;
					break;
				}
				else if (NOT strchr("-e.", *s))
				{
					break;
				}
			}
		}

		if (IsNumber)
		{
			if (StrchrW(f, '.'))
			{
				v = atof(f);
				return true;
			}
			else if (f[0] == '0' AND f[1] == 'x')
			{
				v = htoi(f + 2);
				return true;
			}
			else
			{
				v = atoi(f);
				return true;
			}
		}
		else
		{
			v.Empty();
			v.Type = GV_STRING;
			if (*s == '\"')
			{
				int Len = StrlenW(s);
				v.Value.String = LgiNewUtf16To8(s + 1, (Len - 2) * sizeof(*s));
			}
			else
			{
				v.Value.String = LgiNewUtf16To8(s);
			}
			return true;
		}
	}

	return false;
}

bool GScriptEnginePrivate::Require(int &Cur, char *tok)
{
	if (tok AND Cur >= 0 AND Cur < Tokens.Length())
	{
		char16 *t = Tokens[Cur];
		if (t)
		{
			char b[64];
			int Len = StrlenW(t) * sizeof(*t);
			if (Len < 32)
			{
				int Bytes = LgiBufConvertCp(b, "utf-8", sizeof(b), (void*&)t, LGI_WideCharset, Len);
				b[Bytes/sizeof(*b)] = 0;
				if (stricmp(b, tok) == 0)
				{
					Cur++;
					return true;
				}
			}
		}
	}

	return false;
}

bool GScriptEnginePrivate::Compile_MethodCall(int &Cur)
{
	char16 *t = ThisToken();
	if (t)
	{
		if (NOT Require(Cur, "("))
		{
			Term.Print("%s:%i - Expecting '(' after method name\n", ErrorTitle(), Lines[Cur]);
			return false;
		}

		while (t = ThisToken())
		{
			if (*t == ')')
			{
				break;
			}
			else if (*t == ',')
			{
				Cur++;
			}
			else
			{
				if (NOT Compile_Expression(Cur))
				{
					break;
				}
			}
		}

		if (NOT Require(Cur, ")"))
		{
			Term.Print("%s:%i - Expecting '(' after method name\n", ErrorTitle(), Lines[Cur]);
			return false;
		}
		else return true;
	}

	return false;
}

bool GScriptEnginePrivate::Compile_Expression(int &Cur, int Depth)
{
	if (Cur >= 0 AND Cur < Tokens.Length())
	{
		int Args = 0;
		int Ops = 0;
		int StartToken = Cur;

		char16 *t;
		while (t = Tokens[Cur])
		{
			if (StricmpW(t, (char16*) L"(") == 0)
			{
				Cur++;
				if (Compile_Expression(Cur, Depth + 1))
				{
					Args++;
				}
				else
				{
					break;
				}
			}
			else if (StricmpW(t, (char16*) L")") == 0)
			{
				if (Depth > 0)
					Cur++;
				break;
			}
			else if (StricmpW(t, (char16*) L",") == 0 OR
					 StricmpW(t, (char16*) L";") == 0)
			{
				break;
			}
			else if (Depth == 0 AND StricmpW(t, (char16*) L"]") == 0)
			{
				break;
			}
			else
			{
				if (Args == Ops + 1)
				{
					Operator Op = IsOp(t);
					if (Op == OpNull)
					{
						GObject u;
						u.NameW(t);
						Term.Print("%s:%i - '%s' is not an operator.\n", ErrorTitle(), Lines[StartToken], u.Name());
						break;
					}
					else
					{
						Ops++;
					}
				}
				else
				{
					char m[64], *mp = m;
					for (char16 *c = t; *c AND mp < m + sizeof(m) - 1; c++)
					{
						*mp++ = *c;
					}
					*mp++ = 0;

					if (Methods.Find(m))
					{
						Cur++;

						if (NOT Compile_MethodCall(Cur))
						{
							return false;
						}

						Args++;
						Cur--;
					}
					else
					{
						GDomAddr Addr;
						if (CreateDomAddr(Cur, Addr))
						{
							Cur--;
							Args++;
						}
						else
						{
							Args++;
						}
					}
				}

				Cur++;
			}
		}

		if (Args > 1 AND Ops == Args - 1)
		{
			return true;
		}
		else if (Args == 1)
		{
			return true;
		}
	}
	else
	{
		Term.Print("%s:%i - Unexpected end of file.\n", __FILE__, __LINE__);
	}

	return false;
}

GVariant *GScriptEnginePrivate::Execute_Expression(int &Cur, GDom *Src, int Depth)
{
	if (Cur >= 0 AND Cur < Tokens.Length())
	{
		GArray<GVariant*> Args;
		GArray<Operator> Ops;
		int StartToken = Cur;

		char16 *t;
		while (t = Tokens[Cur])
		{
			if (StricmpW(t, (char16*) L"(") == 0)
			{
				Cur++;
				GVariant *r = Execute_Expression(Cur, Src, Depth + 1);
				if (r)
				{
					Args.Add(r);
				}
				else
				{
					Term.Print("%s:%i - Expression returned NULL (Line %i)\n",
						__FILE__, __LINE__, Lines[StartToken]);
					break;
				}
			}
			else if (StricmpW(t, (char16*) L")") == 0)
			{
				if (Depth > 0)
					Cur++;
				break;
			}
			else if (StricmpW(t, (char16*) L",") == 0 OR
					 StricmpW(t, (char16*) L";") == 0)
			{
				break;
			}
			else if (Depth == 0 AND StricmpW(t, (char16*) L"]") == 0)
			{
				break;
			}
			else
			{
				if (Args.Length() == Ops.Length() + 1)
				{
					Operator Op = IsOp(t);
					if (Op == OpNull)
					{
						GObject u;
						u.NameW(t);
						Term.Print("%s:%i - '%s' is not an operator.\n", ErrorTitle(), Lines[StartToken], u.Name());
						break;
					}
					else
					{
						Ops.Add(Op);
					}
				}
				else
				{
					GFunc *Method = 0;
					char16 *Next = NextToken();
					if (Next AND Next[0] == '(' AND Next[1] == 0)
					{
						char Name[256];
						void *In = t;
						int InLen = StrlenW(t) * sizeof(*t);
						int Converted = LgiBufConvertCp(Name, "utf-8", sizeof(Name)-1, In, LGI_WideCharset, InLen);
						Name[Converted] = 0;
						Method = (GFunc*) Methods.Find(Name);
						if (Method)
						{
							GArray<int> Idx;
							ArgumentArray Arg, Mem;
							GVariant *Ret = NEW(GVariant);

							// Collect arguments
							Cur += 2;
							while (t = ThisToken())
							{
								if (*t == ')')
								{
									Cur++;
									break;
								}
								else if (*t == ',')
								{
									Cur++;
								}
								else
								{
									Idx.Add(Cur);
									Compile_Expression(Cur);
								}
							}

							// Call method
							ProcessArguments(Arg, Mem, &Idx);
							if (Method->Call(this, Ret, Arg))
							{
								Args.Add(Ret);
							}
							else DeleteObj(Ret);

							Mem.DeleteObjects();
							Cur--;
						}
					}

					if (NOT Method)
					{
						GVariant *v = NEW(GVariant);

						int Start = Cur;
						GDomAddr Addr;
						if (CreateDomAddr(Cur, Addr))
						{
							if (Src)
							{
								GArray<char16> a;
								for (int i=Start; i<Cur; i++)
								{
									char16 *k = Tokens[i];
									for (int n=0; k[n]; n++)
										a.Add(k[n]);
								}
								a.Add(0);
								char *utf = LgiNewUtf16To8(&a[0]);
								Src->GetValue(utf, *v);
								DeleteArray(utf);
								Cur--;
							}
							else
							{
								Cur--;
								GDomRef Ref;
								if (ResolveDomAddr(Ref, Addr, false))
								{
									Ref.Get(this, *v);
								}
								else
								{
									LgiAssert(Ref.Array == 0);
									ToVar(*v, t);
								}
							}
						}
						else
						{
							ToVar(*v, t);
						}

						Args.Add(v);
					}
				}

				Cur++;
			}
		}

		if (Args.Length() > 1 AND Ops.Length() == Args.Length() - 1)
		{
			/*
			// Convert all the arguments to typed variants
			GArray<GVariant*> Val;
			for (int a=0; a<Args.Length(); a++)
			{
				GVariant *Value = NEW(GVariant);
				if (Value)
				{
					ToVar(*Value, Args[a]);

					if (Value->Type == GV_STRING)
					{
						GVariant *v = (GVariant*)Variables.Find(Value->Str());
						if (v)
						{
							*Value = *v;
						}
					}

					Val.Add(Value);
				}
			}
			Args.DeleteArrays();
			*/

			// Loop through and combine values and operators according to precedence
			while (Args.Length() > 1)
			{
				// Find which operator to handle first
				int LowestOp = -1;
				int LowestPre = -1;
				for (int i=0; i<Ops.Length(); i++)
				{
					int p = GetPrecedence(Ops[i]);
					if (LowestOp < 0 OR p < LowestPre)
					{
						LowestPre = p;
						LowestOp = i;
					}
				}

				// Combine Arg[LowestOp] Ops[LowestOp] Arg[LowestOp+1] and store the result in 'r'
				GVariant *r = NEW(GVariant);
				if (NOT r)
					break;

				GVariant &a = *Args[LowestOp];
				GVariant &b = *Args[LowestOp + 1];
				switch (Ops[LowestOp])
				{
					case OpPlus:
					{
						if (a.Type == GV_DOUBLE OR b.Type == GV_DOUBLE)
						{
							*r = a.CastDouble() + b.CastDouble();
						}
						else if (a.Type == GV_STRING OR b.Type == GV_STRING)
						{
							GStringPipe p;
							p.Print("%s%s", a.CastString(), b.CastString());
							r->Type = GV_STRING;
							r->Value.String = p.NewStr();
						}
						else
						{
							*r = a.CastInt32() + b.CastInt32();
						}
						break;
					}
					case OpMinus:
					{
						if (a.Type == GV_DOUBLE OR b.Type == GV_DOUBLE)
						{
							*r = a.CastDouble() - b.CastDouble();
						}
						else
						{
							*r = a.CastInt32() - b.CastInt32();
						}
						break;
					}
					case OpMul:
					{
						if (a.Type == GV_DOUBLE OR b.Type == GV_DOUBLE)
						{
							*r = a.CastDouble() * b.CastDouble();
						}
						else
						{
							*r = a.CastInt32() * b.CastInt32();
						}
						break;
					}
					case OpMod:
					{
						if (a.Type == GV_DOUBLE OR b.Type == GV_DOUBLE)
						{
							*r = fmod(a.CastDouble(), b.CastDouble());
						}
						else
						{
							*r = a.CastInt32() % b.CastInt32();
						}
						break;
					}
					case OpDiv:
					{
						if (a.Type == GV_DOUBLE OR b.Type == GV_DOUBLE)
						{
							double Denom = b.CastDouble();
							if (Denom != 0.0)
							{
								*r = a.CastDouble() / Denom;
							}
							else break;
						}
						else
						{
							int Denom = b.CastInt32();
							if (Denom)
							{
								*r = a.CastInt32() / Denom;
							}
							else break;
						}
						break;
					}
					case OpLessThan:
					{
						if (a.Type == GV_DOUBLE OR b.Type == GV_DOUBLE)
						{
							*r = a.CastDouble() < b.CastDouble();
						}
						else
						{
							*r = a.CastInt32() < b.CastInt32();
						}
						break;
					}
					case OpGreaterThan:
					{
						if (a.Type == GV_DOUBLE OR b.Type == GV_DOUBLE)
						{
							*r = a.CastDouble() > b.CastDouble();
						}
						else
						{
							*r = a.CastInt32() > b.CastInt32();
						}
						break;
					}
					case OpLessThanEqual:
					{
						if (a.Type == GV_DOUBLE OR b.Type == GV_DOUBLE)
						{
							*r = a.CastDouble() <= b.CastDouble();
						}
						else
						{
							*r = a.CastInt32() <= b.CastInt32();
						}
						break;
					}
					case OpGreaterThanEqual:
					{
						if (a.Type == GV_DOUBLE OR b.Type == GV_DOUBLE)
						{
							*r = a.CastDouble() >= b.CastDouble();
						}
						else
						{
							*r = a.CastInt32() >= b.CastInt32();
						}
						break;
					}
					case OpEquals:
					{
						if (a.Type == GV_DOUBLE OR b.Type == GV_DOUBLE)
						{
							*r = a.CastDouble() == b.CastDouble();
						}
						else
						{
							*r = a.CastInt32() == b.CastInt32();
						}
						break;
					}
					case OpNotEquals:
					{
						if (a.Type == GV_DOUBLE OR b.Type == GV_DOUBLE)
						{
							*r = a.CastDouble() != b.CastDouble();
						}
						else
						{
							*r = a.CastInt32() != b.CastInt32();
						}
						break;
					}
					case OpAnd:
					{
						*r = a.CastInt32() && b.CastInt32();
						break;
					}
					case OpOr:
					{
						*r = a.CastInt32() || b.CastInt32();
						break;
					}
					default:
					{
						Term.Print("%s:%i - Unhandled operator '%i'\n", ErrorTitle(), Lines[StartToken], Ops[LowestOp]);
						break;
					}
				}

				DeleteObj(Args[LowestOp]);
				DeleteObj(Args[LowestOp + 1]);
				Args[LowestOp] = r;
				Args.DeleteAt(LowestOp + 1, true);
				Ops.DeleteAt(LowestOp, true);
			}

			if (Args.Length() == 1 AND
				Ops.Length() == 0 AND
				Args[0])
			{
				return Args[0];
			}
			else
			{
				LgiAssert(0);
			}

			Args.DeleteObjects();
			return 0;
		}
		else if (Args.Length() == 1)
		{
			/*
			GVariant *v = Var(Args[0], false);
			if (v)
			{
				Args.DeleteArrays();
				if (v->Type == GV_STRING)
				{
					return LgiNewUtf8To16(v->Str());
				}

				switch (v->Type)
				{
					case GV_DOM:
					{
						char s[32];
						sprintf(s, "%i", v->Value.Dom);
						break;
					}
					case GV_VOID_PTR:
					{
						char s[32];
						sprintf(s, "%i", v->Value.Ptr);
						break;
					}
				}

				return LgiNewUtf8To16(v->CastString());
			}
			else if ((t = Args[0]) AND *t == '\"')
			{
				int Len = StrlenW(t);
				char16 *Str = NewStrW(t + 1, Len - 2);
				Args.DeleteArrays();
				return Str;
			}
			*/

			return Args[0];
		}
	}
	else
	{
		Term.Print("%s:%i - Unexpected end of file.\n", __FILE__, __LINE__);
	}

	return 0;
}

int GScriptEnginePrivate::FindEndOfBlock(int Cur)
{
	int Depth = 1;
	while (Cur < Tokens.Length())
	{
		char16 *t = Tokens[Cur];
		if (StricmpW(t, (char16*) L"{") == 0)
		{
			Depth++;
		}
		else if (StricmpW(t, (char16*) L"}") == 0)
		{
			Depth--;
			if (Depth == 0)
			{
				return Cur;
				break;
			}
		}
		Cur++;
	}

	return -1;
}

bool GScriptEnginePrivate::Type(int &Cur, GExternType &Type)
{
	char16 *t = Tokens[Cur];
	if (t)
	{
		if (StricmpW(t, sInt) == 0 OR
			StricmpW(t, sUInt) == 0 OR
			StricmpW(t, sUInt) == 0 OR
			StricmpW(t, sUInt32) == 0 OR
			StrcmpW(t, sHWND) == 0 OR
			StrcmpW(t, sDWORD) == 0)
		{
			Type.Simple = GV_INT32;
		}
		else if (StricmpW(t, sInt64) == 0)
		{
			Type.Simple = GV_INT64;
		}
		else if (StricmpW(t, sLPCTSTR) == 0 OR
				 StricmpW(t, sLPTSTR) == 0)
		{
			Type.Simple = GV_STRING;
		}
		else if (StricmpW(t, sChar) == 0)
		{
			t = Tokens[Cur + 1];
			if (t AND t[0] == '*' AND NOT t[1])
			{
				Cur++;
				Type.Simple = GV_STRING;
			}
		}
		else
		{
			return false;
		}

		Cur++;
		return true;
	}
	
	return false;
}

bool GFunc::Call(GScriptEnginePrivate *Priv, GVariant *Ret, GArray<int> *Idx)
{
	if (NOT Priv)
		return false;

	ArgumentArray Args, Mem;
	Priv->ProcessArguments(Args, Mem, Idx);
	bool Status = Call(Priv, Ret, Args);
	Mem.DeleteObjects();
	return Status;
}

bool GExternFunc::Call(GScriptEnginePrivate *Priv, GVariant *Ret, ArgumentArray &In)
{
	if (NOT Priv OR NOT Library OR NOT Method)
		return false;

	bool Status = false;
	GArray<uint32> ArgVal;
	GArray<char*> Mem;

	for (int i=0; i<In.Length(); i++)
	{
		/*
		int Cur = (*Idx)[i];
		char16 *t1 = Priv->Tokens[Cur];
		char16 *t2 = Priv->Tokens[Cur + 1];
		GVariant *v = t1 AND isalpha(*t1) AND t2 AND (*t2 == ',' OR *t2 == ')') ? Priv->Var(t1, false) : 0;
		*/

		GVariant *v = In[i];
		if (v)
		{
			switch (Args[i]->Simple)
			{
				case GV_INT32:
				{
					ArgVal[i] = v->CastInt32();
					break;
				}
				case GV_INT64:
				{
					ArgVal[i] = v->CastInt32();
					break;
				}
				case GV_STRING:
				{
					if (v->Str())
					{
						ArgVal[i] = (int) v->Str();
					}
					else
					{
						ArgVal[i] = (int) v->CastString();
					}
					break;
				}
				default:
				{
					LgiAssert(0);
					break;
				}
			}
		}
	}

	if (ArgVal.Length() == Args.Length())
	{
		GLibrary Lib(Library);
		if (Lib.IsLoaded())
		{
			void *c = Lib.GetAddress(Method);
			if (c)
			{
				uint32 a = ArgVal.Length();
				uint32 b = (uint32) &ArgVal[Args.Length()-1];
				uint32 r = 0;

				#if defined(WIN32)
				_asm
				{
					mov ecx, a;
					mov ebx, b;
				}
				label1:
				_asm
				{
					push [ebx];
					sub ebx, 4;
					loop label1;
					
					mov ebx, c;
					call ebx;
					mov r, eax;
				}
				#else
				// Not implemented
				LgiAssert(0);
				#endif

				if (Ret) (*Ret) = (int) r;
				Status = true;
			}
			else
			{
				Priv->Term.Print("%s - Extern library '%s' missing function '%s'.\n",
					Priv->ErrorTitle(),
					Library,
					Method);
			}
		}
		else
		{
			Priv->Term.Print("%s - Extern library '%s' missing.\n", Priv->ErrorTitle(), Library);
		}
	}
	else
	{
		Priv->Term.Print("%s - Extern '%s' expecting %i arguments, not %i given.\n",
			Priv->ErrorTitle(),
			Method,
			Args.Length(),
			ArgVal.Length());
	}

	Mem.DeleteArrays();

	return Status;
}

bool GScriptFunc::Call(GScriptEnginePrivate *Priv, GVariant *Ret, ArgumentArray &Values)
{
	if (NOT Body OR NOT Priv)
		return false;

	bool Status = false;

	// Process the arguments into values...
	if (Args.Length() == Values.Length())
	{
		// Values -> variables...
		GArray<GVariant*> Old;
		int i;
		for (i=0; i<Args.Length(); i++)
		{
			Old[i] = (GVariant*) Priv->Variables.Find(Args[i]);
			
			GVariant *n = NEW(GVariant);
			if (n)
			{
				*n = *(Values[i]);
				Priv->Variables.Add(Args[i], (void*)n);
			}
		}

		// Call the method
		Status = Priv->Execute_Statement(Body->Code);

		// Remove the variables, and replace with whatever was there before
		for (i=0; i<Args.Length(); i++)
		{
			GVariant *v = (GVariant*) Priv->Variables.Find(Args[i]);
			if (v)
			{
				Priv->Variables.Delete(Args[i]);
				DeleteObj(v);

				if (Old[i])
				{
					Priv->Variables.Add(Args[i], (void*)Old[i]);
				}
			}
		}
	}
	else
	{
		Priv->Term.Print("%s - Script function '%s' was expecting %i arguments, not the %i given.\n",
			Priv->ErrorTitle(),
			Method,
			Args.Length(),
			Values.Length());
	}

	return Status;
}

bool GHostFunc::Call(GScriptEnginePrivate *Priv, GVariant *Ret, ArgumentArray &Args)
{
	return (Priv->Context->*(Func))(Ret, Args);
}

void GScriptEnginePrivate::ProcessArguments(ArgumentArray &Args, ArgumentArray &Mem, GArray<int> *Idx)
{
	if (Idx)
	{
		for (int i=0; i<Idx->Length(); i++)
		{
			int Cur = (*Idx)[i];
			char16 *Tok = ThisToken();
			char16 *Next = NextToken();
			GVariant *v = 0;
			if (Tok AND Next AND isalpha(*Tok) AND (*Next == ')' OR *Next == ','))
			{
				v = Var(Tok, false);
			}
			if (v)
			{
				Args[i] = v;
			}
			else
			{
				GVariant *v = Execute_Expression(Cur);
				if (v)
				{
					Args[i] = v;
					Mem.Add(Args[i]);
				}
			}
		}
	}
}

bool GScriptEnginePrivate::Execute_Statement(GArray<GCode> &To)
{
	bool Status = false;

	for (int k=0; k<To.Length(); k++)
	{
		GCode &Code = To[k];
		switch (Code.Type)
		{
			case GS_IF:
			{
				int Cur = Code.If.Condition;
				GVariant *e = Execute_Expression(Cur);
				if (e)
				{
					bool IsTrue;
					switch (e->Type)
					{
						case GV_VOID_PTR:
							IsTrue = e->Value.Ptr != 0;
							break;
						case GV_DOM:
							IsTrue = e->Value.Dom != 0;
							break;
						case GV_STRING:
							IsTrue = e->Value.String != 0;
							break;
						default:
							IsTrue = e->CastInt32();
							break;
					}

					if (IsTrue)
					{
						Status = Execute_Statement(Code.Code);
					}
					else if (Code.If.Else)
					{
						Status = Execute_Statement(*Code.If.Else);
					}
					else
					{
						Status = true;
					}

					DeleteObj(e);
				}
				else
				{
					Term.Print("%s:%i - Couldn't evaluate if statement's condition.\n", ErrorTitle(), Lines[Code.Start]);
				}
				break;
			}
			case GS_FUNCTION_DEFN:
			{
				// Do nothing here...
				break;
			}
			case GS_FUNCTION_CALL:
			{
				if (NOT Code.MethodCall.Method)
					break;

				GVariant Ret;
				Status = Code.MethodCall.Method->Call(this, &Ret, Code.MethodCall.Args);
				if (NOT Status)
				{
					Term.Print("%s:%i - Command '%s' failed.\n", ErrorTitle(), Lines[Code.Start], Code.MethodCall.Method->Method);
				}
				break;
			}
			case GS_FOR:
			{
				if (Code.For.PreStatement)
				{
					if (NOT Execute_Statement(*Code.For.PreStatement))
					{
						Term.Print("%s:%i - Failed to execute for loop's PreStatement.\n", ErrorTitle(), Lines[Code.Start]);
						break;
					}
				}

				while (true)
				{
					// Evaluate the condition
					int Cur = Code.For.Condition;
					GVariant *v = Execute_Expression(Cur);
					if (NOT v)
					{
						break;
					}

					bool Cond = v->CastInt32() != 0;
					DeleteObj(v);
					if (NOT Cond)
					{
						Status = true;
						break;
					}

					// Execute the body
					Execute_Statement(Code.Code);

					// Execute the post statement
					if (Code.For.PostStatement)
					{
						if (NOT Execute_Statement(*Code.For.PostStatement))
						{
							Term.Print("%s:%i - Failed to execute for loop's PostStatement.\n", ErrorTitle(), Lines[Code.Start]);
							break;
						}
					}
				}
				break;
			}
			case GS_LVALUE:
			{
				GDomAddr Dom;
				int Cur = Code.Start;
				if (CreateDomAddr(Cur, Dom))
				{
					switch (Code.LValue.Op)
					{
						case OpAssign:
						{
							Cur = Code.LValue.Expression;
							GVariant *v = Execute_Expression(Cur);
							if (v)
							{
								GDomRef Ref;

								/*
								printf("Assigning to ");
								Dom.Print();
								printf("\n");
								*/

								if (ResolveDomAddr(Ref, Dom, true))
								{
									if (Ref.Member)
									{
										if (Ref.Var->Type == GV_DOM AND
											Ref.Var->Value.Dom)
										{
											Ref.Var->Value.Dom->SetValue(Ref.Member, *v);
										}
										else if (Ref.Var->Type == GV_STRING OR
												Ref.Var->Type == GV_NULL)
										{
											if (stricmp(Ref.Member, "length") == 0)
											{
												Ref.Var->Empty();

												int Len = v->CastInt32();
												char *Mem = NEW(char[Len+1]);
												if (Mem)
												{
													memset(Mem, ' ', Len);
													Mem[Len] = 0;
													Ref.Var->Value.String = Mem;
													Ref.Var->Type = GV_STRING;
												}
											}
										}
									}
									else
									{
										if (Ref.Array)
										{
											Ref.Set(this, *v);
										}
										else
										{
											// Pure assignment..
											*Ref.Var = *v;
										}
									}

									Status = true;
								}
								DeleteObj(v);
							}
							else
							{
								Term.Print("%s:%i - Failed to handle expression after assignment.\n", ErrorTitle(), Lines[Code.Start]);
							}
							break;
						}
						case OpInc:
						{
							GVariant *v = Var(Dom[0]);
							if (v)
							{
								*v = v->CastInt32() + 1;
								Status = true;
							}
							break;
						}
						case OpDec:
						{
							GVariant *v = Var(Dom[0]);
							if (v)
							{
								*v = v->CastInt32() - 1;
								Status = true;
							}
							break;
						}
						case OpPlusEquals:
						{
							Cur = Code.LValue.Expression;
							GVariant *v = Execute_Expression(Cur);
							if (v)
							{
								GDomRef Ref;
								if (ResolveDomAddr(Ref, Dom, true))
								{
									if (Ref.Member)
									{
										if (Ref.Var->Type == GV_DOM AND
											Ref.Var->Value.Dom)
										{
											GVariant a;
											if (Ref.Var->Value.Dom->GetValue(Ref.Member, a))
											{
												GVariant b = a.CastInt32() + v->CastInt32();
												Ref.Var->Value.Dom->SetValue(Ref.Member, b);
											}
										}
										else if (Ref.Var->Type == GV_STRING OR
												Ref.Var->Type == GV_NULL)
										{
											LgiAssert(0);

											/*
											if (stricmp(Ref.Member, "length") == 0)
											{
												Ref.Var->Empty();

												int Len = v->CastInt32();
												char *Mem = NEW(char[Len+1]);
												if (Mem)
												{
													memset(Mem, ' ', Len);
													Mem[Len] = 0;
													Ref.Var->Value.String = Mem;
													Ref.Var->Type = GV_STRING;
												}
											}
											*/
										}
									}
									else
									{
										*Ref.Var = Ref.Var->CastInt32() + v->CastInt32();
									}

									Status = true;
								}
								DeleteObj(v);
							}
							else
							{
								Term.Print("%s:%i - Failed to handle expression after assignment.\n", ErrorTitle(), Lines[Code.Start]);
							}
							break;
						}
						default:
						{
							LgiAssert(0);
							break;
						}
					}
				}
				break;
			}
			default:
			{
				LgiAssert(0);
				return false;
				break;
			}
		}
	}

	return true;
}

bool GScriptEnginePrivate::Compile_Statement(GArray<GCode> &To, int &Cur)
{
	bool Status = false;

	int StartToken = Cur;
	char16 *t = Cur < Tokens.Length() ? Tokens[Cur++] : 0;
	if (t)
	{
		if (StricmpW(t, (char16*) L"if") == 0)
		{
			GCode &Code = To[To.Length()];
			Code.Type = GS_IF;
			Code.Start = Cur - 1;
			Code.If.Else = 0;

			if (NOT Require(Cur, "("))
			{
				Term.Print("%s:%i - Missing '(' after 'if'.\n", ErrorTitle(), Lines[StartToken]);
			}
			else
			{
				Code.If.Condition = Cur;
				if (NOT Compile_Expression(Cur))
				{
					Term.Print("%s:%i - Couldn't compile if statement's condition expression.\n", ErrorTitle(), Lines[StartToken]);
				}
				else
				{
					if (NOT Require(Cur, ")"))
					{
						Term.Print("%s:%i - Missing ')' after if statement's expression.\n", ErrorTitle(), Lines[Cur]);
					}
					else
					{
						if (NOT Require(Cur, "{"))
						{
							Term.Print("%s:%i - Missing start block '{' in if statement.\n", ErrorTitle(), Lines[Cur]);
						}
						else
						{
							while (t = ThisToken())
							{
								if (StricmpW(t, (char16*) L"}") == 0)
								{
									Status = true;
									Cur++;
									break;
								}
								else
								{
									if (NOT Compile_Statement(Code.Code, Cur))
									{
										break;
									}
								}
							}

							if (Status AND
								(t = ThisToken()) != 0 AND
								StricmpW(t, sElse) == 0)
							{
								if (Code.If.Else = NEW(GArray<GCode>))
								{
									Cur++;
									if (NOT Require(Cur, "{"))
									{
										Term.Print("%s:%i - Missing start block '{' after else statement.\n", ErrorTitle(), Lines[Cur]);
									}
									else
									{
										while (t = ThisToken())
										{
											if (StricmpW(t, (char16*) L"}") == 0)
											{
												Status = true;
												Cur++;
												break;
											}
											else
											{
												if (NOT Compile_Statement(*Code.If.Else, Cur))
												{
													break;
												}
											}
										}
									}
								}
							}
						}
					}
				}
			}
		}
		else if (StricmpW(t, (char16*) L"function") == 0)
		{
			GCode &Code = To[To.Length()];
			Code.Type = GS_FUNCTION_DEFN;
			Code.Start = Cur - 1;
			if (t = ThisToken())
			{
				GScriptFunc *Func = 0;
				if (NOT isalpha(*t))
				{
					Term.Print("%s:%i - Function name must start with a letter.\n", ErrorTitle(), Lines[Cur]);
					goto EndFuncDefn;
				}

				Func = NEW(GScriptFunc);
				if (Func)
				{
					Func->Method = LgiNewUtf16To8(t);
					Func->Body = &Code;
					Methods.Add(Func->Method, (GFunc*)Func);

					if (NOT Require(++Cur, "("))
					{
						Term.Print("%s:%i - Missing '(' in funcion defn argument list.\n", ErrorTitle(), Lines[Cur]);
						goto EndFuncDefn;
					}

					// Read any arguments
					while (t = ThisToken())
					{
						if (isalpha(*t))
						{
							Func->Args.Add(LgiNewUtf16To8(t));
						}
						else if (StricmpW(t, sComma) == 0)
						{
							// Skip
						}
						else if (StricmpW(t, sEndRdBracket) == 0)
						{
							break;
						}
						else
						{
							char *t8 = LgiNewUtf16To8(t);
							Term.Print("%s:%i - Unexpected token '%s' in argument defn.\n", ErrorTitle(), Lines[Cur], t8);
							DeleteArray(t8);
						}
						Cur++;
					}

					if (NOT Require(Cur, ")"))
					{
						Term.Print("%s:%i - Missing ')' in funcion defn argument list.\n", ErrorTitle(), Lines[Cur]);
						goto EndFuncDefn;
					}

					if (NOT Require(Cur, "{"))
					{
						Term.Print("%s:%i - Expecting '{' after function name.\n", ErrorTitle(), Lines[Cur]);
						goto EndFuncDefn;
					}

					Status = true;
					do
					{
						t = ThisToken();
						if (*t != '}')
						{
							if (NOT Compile_Statement(Code.Code, Cur))
							{
								Status = false;
								break;
							}
						}
					}
					while (StricmpW(t, sEndCurlyBracket) != 0);

					if (Require(Cur, "}"))
					{
						Func = 0;
					}
					else
					{
						Status = false;
						Term.Print("%s:%i - Expecting '}' after function body.\n", ErrorTitle(), Lines[Cur]);
					}
				}

				EndFuncDefn:
				if (Func)
				{
					Methods.Delete(Func->Method);
					DeleteObj(Func);
				}
			}
			else
			{
				Term.Print("%s:%i - Expecting function name.\n", ErrorTitle(), Lines[Cur-1]);
			}						
		}
		else if (StricmpW(t, (char16*) L"extern") == 0)
		{
			GExternFunc *Func = NEW(GExternFunc);
			if (Func)
			{
				// Parse the return type
				if (Type(Cur, Func->Return))
				{
					char16 *t = Tokens[Cur];
					if (t AND isalpha(*t))
					{
						// Store the method name
						Func->Method = LgiNewUtf16To8(t);
						Cur++;

						if (NOT Require(Cur, "("))
						{
							Term.Print("%s:%i - Expecting '('.\n", ErrorTitle(), Lines[Cur]);
						}
						else
						{
							// Parse arguments to the func
							while (true)
							{
								GExternType *k = NEW(GExternType);
								if (k)
								{
									if (Type(Cur, *k))
									{
										Func->Args.Add(k);
									}
									else
									{
										Term.Print("%s:%i - Expecting type.\n", ErrorTitle(), Lines[Cur]);
										break;
									}

									t = Tokens[Cur];
									if (t AND *t == ',')
									{
										Cur++;
									}
									else break;										
								}
								else break;
							}

							if (NOT Require(Cur, ")"))
							{
								Term.Print("%s:%i - Expecting ')'.\n", ErrorTitle(), Lines[Cur]);
							}
							else if (NOT Require(Cur, "in"))
							{
								Term.Print("%s:%i - Expecting 'in'.\n", ErrorTitle(), Lines[Cur]);
							}
							else
							{
								t = Tokens[Cur];
								if (t)
								{
									Cur++;

									GVariant v;
									ToVar(v, t);
									Func->Library = NewStr(v.Str());

									if (NOT Require(Cur, ";"))
									{
										Term.Print("%s:%i - Expecting ';'.\n", ErrorTitle(), Lines[Cur]);
									}
									else
									{
										Methods.Add(Func->Method, (GFunc*)Func);
										Status = true;
										Func = 0;
									}
								}
							}
						}
					}
					else
					{
						Term.Print("%s:%i - Expecting method name.\n", ErrorTitle(), Lines[Cur]);
					}
				}

				DeleteObj(Func);
			}
		}
		else if (StricmpW(t, (char16*) L"for") == 0)
		{
			GCode &Code = To[To.Length()];
			Code.Type = GS_FOR;
			Code.Start = Cur - 1;

			// For loop
			if (NOT Require(Cur, "("))
			{
				Term.Print("%s:%i - Missing '(' after 'for'.\n", ErrorTitle(), Lines[StartToken]);
			}
			else
			{
				// Execute the initial statement
				if ((Code.For.PreStatement = NEW(GArray<GCode>)) != NULL AND
					Compile_Statement(*Code.For.PreStatement, Cur))
				{
					// Store the start position of the condition tokens
					Code.For.Condition = Cur;

					// Compile the condition
					Compile_Expression(Cur);

					if (Require(Cur, ";"))
					{
						// Skip over the post statement
						if ((Code.For.PostStatement = NEW(GArray<GCode>)) != NULL AND
							Compile_Statement(*Code.For.PostStatement, Cur))
						{
							// See the start of the block...
							if (Require(Cur, "{"))
							{
								// Execute content
								while (t = Tokens[Cur])
								{
									if (StricmpW(t, (char16*) L"}") == 0)
									{
										Status = true;
										Cur++;
										break;
									}
									else if (NOT Compile_Statement(Code.Code, Cur))
									{
										break;
									}
								}
							}
							else
							{
								Term.Print("%s:%i - Missing '{'.\n", ErrorTitle(), Lines[Cur]);
							}
						}
						else
						{
							Term.Print("%s:%i - Failed to compile post statement.\n", ErrorTitle(), Lines[StartToken]);
						}
					}
					else
					{
						Term.Print("%s:%i - Missing ';' after expression.\n", ErrorTitle(), Lines[StartToken]);
					}
				}
			}
		}
		else
		{
			// See if this token is a method
			bool Result = false;
			char *ErrMsg = 0;
			char *t8 = LgiNewUtf16To8(t);
			GFunc *Func = (GFunc*) Methods.Find(t8);
			if (Func)
			{
				GCode &Code = To[To.Length()];
				Code.Type = GS_FUNCTION_CALL;
				Code.Start = Cur - 1;
				Code.MethodCall.Args = 0;
				Code.MethodCall.Method = Func;

				if (NOT Require(Cur, "("))
				{
					Term.Print("%s:%i - Missing beginning '(' function call.\n", ErrorTitle(), Lines[Cur]);
				}
				else
				{
					// Parse through any arguments.
					while (true)
					{
						int Start = Cur;
						if (Compile_Expression(Cur))
						{
							if (NOT Code.MethodCall.Args)
							{
								Code.MethodCall.Args = NEW(GArray<int>);
							}
							if (Code.MethodCall.Args)
							{
								Code.MethodCall.Args->Add(Start);
							}
						}
						t = Tokens[Cur];
						if (t)
						{
							if (*t == ',')
							{
								Cur++;
							}
							else if (*t == ')')
							{
								Cur++;
								break;
							}
							else
							{
								GObject u;
								u.NameW(t);
								Term.Print("%s:%i - Not expecting token '%s' in argument list.\n",
									ErrorTitle(),
									Lines[Start], u.Name());
								break;
							}
						}
						else break;
					}

					if (Require(Cur, ";"))
					{
						Status = true;
					}
					else
					{
						Term.Print("%s:%i - Missing ';' after function call.\n", ErrorTitle(), Lines[StartToken]);
					}
				}
			}
			else
			{
				// Not an external method, is it an L-value?
				GDomAddr Dom;
				Cur--;
				if (CreateDomAddr(Cur, Dom))
				{
					// Look at operator
					char16 *n = ThisToken();
					if (n)
					{
						Operator Op = IsOp(n);
						switch (Op)
						{
							case OpPlusEquals:
							case OpAssign:
							case OpMinusEquals:
							case OpMulEquals:
							case OpDivEquals:
							{
								GCode &Code = To[To.Length()];
								Code.Type = GS_LVALUE;
								Code.Start = StartToken;
								Code.LValue.Op = Op;
								Code.LValue.Expression = ++Cur;

								if (Compile_Expression(Cur))
								{
									t = ThisToken();
									if (t)
									{
										if (StricmpW(t, sSemiColon) == 0)
										{
											Cur++;
											Status = true;
										}
										else if (StricmpW(t, sEndRdBracket) == 0)
										{
											Cur++;
											Status = true;
										}
									}
								}
								else
								{
									Term.Print("%s:%i - Failed to compile expression after assignment.\n", ErrorTitle(), Lines[StartToken]);
								}
								break;
							}
							case OpInc:
							case OpDec:
							{
								GCode &Code = To[To.Length()];
								Code.Type = GS_LVALUE;
								Code.Start = StartToken;
								Code.LValue.Op = Op;
								Code.LValue.Expression = ++Cur;
								
								t = ThisToken();
								if (t)
								{
									if (StricmpW(t, sSemiColon) == 0)
									{
										Cur++;
										Status = true;
									}
									else if (StricmpW(t, sEndRdBracket) == 0)
									{
										Cur++;
										Status = true;
									}
								}
								break;
							}
							default:
							{
								Term.Print("%s:%i - '%s' isn't an operator.\n", ErrorTitle(), Lines[StartToken], n);
								break;
							}
						}
					}
				}
			}

			DeleteArray(t8);
		}
	}
	else
	{
		Term.Print("%s:%i - Unexpected end of file in statement.\n", __FILE__, __LINE__);
	}

	return Status;
}


