
extern "C" {
#define _IMPEXP_KERNEL
#include "Os.h"
#include "fsproto.h"
}

#include "Bfs.h"
#include <Windows.h>

extern vnode_ops fs_entry;
#define DeleteObj(o) if (o) { delete o; o = 0; }

class BfsFile
{
public:
	void *Node;
	void *Cookie;
	int Pos;

	BfsFile()
	{
		Node = 0;
		Cookie = 0;
		Pos = 0;
	}
};

class Bfs
{
	char *FileName;
	void *Volume;
	void *Root; // root vnode

	int64 FileSize(char *f)
	{
		WIN32_FIND_DATA Data;
		HANDLE h = FindFirstFile(f, &Data);
		int64 Status = h != INVALID_HANDLE_VALUE ? ((int64)Data.nFileSizeHigh << 32) | Data.nFileSizeLow : 0;
		FindClose(h);
		return Status;
	}

	void *DirToVnode(void *Node, char *Path, bool DirOnly, int Depth = 0)
	{
		if (Node && Path)
		{
			char *d = strchr(Path, '/');
			if (d++)
			{
				char *n = strchr(d, '/');
				char Part[128];
				char *Seg;
				if (n)
				{
					int Len = (int)n-(int)d;
					memcpy(Part, d, Len);
					Part[Len] = 0;
					Seg = Part;
				}
				else
				{
					if (DirOnly)
					{
						return Depth == 0 ? Node : 0;
					}

					Seg = d;
				}
					
				char *NewPath;
				vnode_id Id;
				if (fs_entry.walk(Volume, Node, Seg, &NewPath, &Id) == B_OK)
				{
					void *NewNode = 0;
					if (fs_entry.read_vnode(Volume, Id, 0, &NewNode) == B_OK)
					{
						if (n)
						{
							return DirToVnode(NewNode, n + 1, DirOnly, Depth + 1);
						}
						else
						{
							return NewNode;
						}
					}
				}
			}
		}
		return 0;
	}

public:
	Bfs(char *f)
	{
		int Size = 16 << 20; // 16mb
		int Status = 0;

		Root = 0;
		Volume = 0;
		FileName = strdup(f);

		if (FileSize(f) < 1024)
		{
			// Create volume
 			Status = fs_entry.initialize(FileName, 0, Size);
		}

		// Mount volume
		vnode_id RootId;
		Status = fs_entry.mount(0, FileName, 0, 0, Size, &Volume, &RootId);
		if (Status == B_OK)
		{
			fs_entry.read_vnode(Volume, RootId, 0, &Root);
		}

		if (Status)
		{
			char s[256];
			sprintf(s, "BFS failed with %i.\n", Status);
			MessageBox(0, s, "Error", MB_OK);
		}
	}

	~Bfs()
	{
		fs_entry.unmount(0);
	}

	int OpenFile(char *FileName, int Mode)
	{
		BfsFile *f = new BfsFile;
		if (f && FileName)
		{
			f->Node = DirToVnode(Root, FileName, false);
			if (f->Node)
			{
				if (fs_entry.open(Volume, f->Node, Mode, &f->Cookie) != B_OK)
				{
					DeleteObj(f);
				}
			}
			else
			{
				void *Dir = DirToVnode(Root, FileName, true);
				if (Dir)
				{
					char *Name = strrchr(FileName, '/');
					vnode_id Id;
					if (Name &&
						fs_entry.create(Volume, Dir, Name+1, S_IFREG, Mode, &Id, &f->Cookie) == B_OK)
					{
						fs_entry.read_vnode(Volume, Id, 0, &f->Node);
					}
					else
					{
						DeleteObj(f);
					}
				}
			}
		}
		return (int) f;
	}

	void CloseFile(int File)
	{
	}

	int Read(int File, void *Ptr, int Size)
	{
		int r = -1;
		BfsFile *f = (BfsFile*) File;
		if (f)
		{
			r = fs_entry.read(Volume, f->Node, f->Cookie, f->Pos, Ptr, (size_t*) &Size);
			if (r > 0)
			{
				f->Pos += Size;
			}
		}
		return r;
	}

	int Write(int File, void *Ptr, int Size)
	{
		int w = -1;
		BfsFile *f = (BfsFile*) File;
		if (f)
		{
			w = fs_entry.write(Volume, f->Node, f->Cookie, f->Pos, Ptr, (size_t*) &Size);
			if (w > 0)
			{
				f->Pos += Size;
			}
		}
		return w;
	}

	int64 GetFilePos(int File)
	{
		return 0;
	}

	int SetFilePos(int File, int64 Pos)
	{
		return 0;
	}

	int64 GetFileSize(int File)
	{
		BfsFile *f = (BfsFile*) File;
		if (f)
		{
			stat s;
			if (fs_entry.rstat(Volume, f->Node, &s) == B_OK)
			{
				return s.st_size;
			}
		}

		return -1;
	}

	int SetFileSize(int File, int64 Size)
	{
		return 0;
	}
};


Bfs *BfsCreate(char *FileName)
{
	return new Bfs(FileName);
}

void BfsDestroy(Bfs *b)
{
	if (b)
	{
		delete b;
	}
}

int BfsOpenFile(Bfs *b, char *FileName, int Mode)
{
	return b->OpenFile(FileName, Mode);
}

void BfsCloseFile(Bfs *b, int File)
{
	b->CloseFile(File);
}

int BfsReadFile(Bfs *b, int File, void *Ptr, int Size)
{
	return b->Read(File, Ptr, Size);
}

int BfsWriteFile(Bfs *b, int File, void *Ptr, int Size)
{
	return b->Write(File, Ptr, Size);
}

int64 BfsGetFileSize(Bfs *b, int File)
{
	return b->GetFileSize(File);
}

int BfsSetFileSize(Bfs *b, int File, int64 Size)
{
	return b->SetFileSize(File, Size);
}

int64 BfsGetFilePos(Bfs *b, int File)
{
	return b->GetFilePos(File);
}

int BfsSetFilePos(Bfs *b, int File, int64 Pos)
{
	return b->SetFilePos(File, Pos);
}
