/* Volume - BFS super block, mounting, etc.
**
** Initial version by Axel Dörfler, axeld@pinc-software.de
** This file may be used under the terms of the OpenBeOS License.
*/


#include "Debug.h"
#include "cpp.h"
#include "Volume.h"
#include "Journal.h"
#include "Inode.h"
#include "Query.h"

#include "KernelExport.h"
#include "fs_volume.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>


Volume::Volume(nspace_id id)
	:
	fID(id),
	fBlockAllocator(this),
	fLock("bfs volume"),
	fDirtyCachedBlocks(0),
	fUniqueID(0),
	fFlags(0),
	fIndicesNode(0)
{
	fJournal = 0;
}


Volume::~Volume()
{
}


bool
Volume::IsValidSuperBlock()
{
	if (fSuperBlock.magic1 != (int32)SUPER_BLOCK_MAGIC1
		|| fSuperBlock.magic2 != (int32)SUPER_BLOCK_MAGIC2
		|| fSuperBlock.magic3 != (int32)SUPER_BLOCK_MAGIC3
		|| (int32)fSuperBlock.block_size != fSuperBlock.inode_size
		|| fSuperBlock.fs_byte_order != SUPER_BLOCK_FS_LENDIAN
		|| (1UL << fSuperBlock.block_shift) != fSuperBlock.block_size
		|| fSuperBlock.num_ags < 1
		|| fSuperBlock.ag_shift < 1
		|| fSuperBlock.blocks_per_ag < 1
		|| fSuperBlock.num_blocks < 10
		|| fSuperBlock.num_ags != divide_roundup((int64)fSuperBlock.num_blocks,1L << fSuperBlock.ag_shift))
		return false;

	return true;
}


void 
Volume::Panic()
{
	FATAL(("we have to panic... switch to read-only mode!\n"));
	fFlags |= VOLUME_READ_ONLY;
#ifdef USER
	debugger("BFS panics!");
#elif defined(DEBUG)
	kernel_debugger("BFS panics!");
#endif
}

status_t
Volume::Initialize(const char *deviceName, uint32 flags)
{
	status_t Status = B_ERROR;

	fRootNode = 0;
	fIndicesNode = 0;
	fJournal = 0;
	fDevice = open(deviceName, O_RDWR);
	if (fDevice < B_OK)
		RETURN_ERROR(fDevice);

	int SuperSize = sizeof(fSuperBlock);
	assert(SuperSize == 512);
	memset(&fSuperBlock, 0, SuperSize);
	strcpy(fSuperBlock.name, "libbfs Volume");

	// Fill out the super block with some default information
	fSuperBlock.magic1 = (int32)SUPER_BLOCK_MAGIC1;
	fSuperBlock.magic2 = (int32)SUPER_BLOCK_MAGIC2;
	fSuperBlock.magic3 = (int32)SUPER_BLOCK_MAGIC3;
	fSuperBlock.block_size = fSuperBlock.inode_size = 2 << 10;
	fSuperBlock.fs_byte_order = SUPER_BLOCK_FS_LENDIAN;

	for (fSuperBlock.block_shift = 0;
		fSuperBlock.block_shift < sizeof(fSuperBlock.block_size) * 8 &&
		(fSuperBlock.block_size & (1 << fSuperBlock.block_shift)) == 0;
		fSuperBlock.block_shift++);

	// These values are taken from the BeOS personal edition BFS image
	// that I thought would make a good starting point to get a working
	// image, obviously limiting the size to 512mb should be addressed
	// later.
	fSuperBlock.flags = SUPER_BLOCK_DISK_CLEAN;
	fSuperBlock.num_blocks = (64 << 20) / fSuperBlock.block_size; // 64mb
	fSuperBlock.ag_shift = 14;
	fSuperBlock.num_ags = fSuperBlock.num_blocks >> fSuperBlock.ag_shift;
	fSuperBlock.blocks_per_ag = 1;
	
	fSuperBlock.log_start = 749;
	fSuperBlock.log_end = 749;

	// Allocate area for logging...
	fSuperBlock.log_blocks.allocation_group = 0;
	fSuperBlock.log_blocks.start = 17;
	fSuperBlock.log_blocks.length = 2048;

	// Allocate area for indexes.. 
	fSuperBlock.indices.allocation_group = 0;
	fSuperBlock.indices.start = fSuperBlock.log_blocks.start + fSuperBlock.log_blocks.length;
	fSuperBlock.indices.length = 1;

	// Put the root node in the middle of the free space
	fSuperBlock.root_dir.allocation_group = fSuperBlock.num_ags / 2;
	fSuperBlock.root_dir.start = 0;
	fSuperBlock.root_dir.length = 1;

	// Start the Journal
	fJournal = new Journal(this);

	if (fBlockAllocator.Initialize(true) == B_OK)
	{
		if (IsValidSuperBlock() &&
			WriteSuperBlock() == B_OK)
		{
			fRootNode = new Inode(this, ToVnode(Root()));
			if (fRootNode)
			{
				Transaction transaction(this, fRootNode->BlockNumber());
				Status = fRootNode->Create(&transaction, 0, "/", S_DIRECTORY, S_IREAD, 0);
				delete fRootNode;
				fRootNode = 0;
			}

			fIndicesNode = new Inode(this, ToVnode(Indices()));
			if (fIndicesNode)
			{
				Transaction transaction(this, fIndicesNode->BlockNumber());
				Status = fIndicesNode->Create(&transaction, 0, "", S_INDEX_DIR, S_IREAD, 0);
				delete fIndicesNode;
				fIndicesNode = 0;
			}

			if (fRootNode)
			{
				/*
				Status = fRootNode->Create(	&t,				// transaction
											0,				// parent
											"Root",			// name
											S_DIRECTORY,	// mode
											S_IREAD,		// open mode
											0);				// type ??
				*/
			}
		}
	}

	close(fDevice);
	
	return Status;
}


status_t
Volume::Mount(const char *deviceName, uint32 flags)
{
	if (flags & B_MOUNT_READ_ONLY)
		fFlags |= VOLUME_READ_ONLY;

	fDevice = open(deviceName,flags & B_MOUNT_READ_ONLY ? O_RDONLY : O_RDWR);
	
	// if we couldn't open the device, try read-only (don't rely on a specific error code)
	if (fDevice < B_OK && (flags & B_MOUNT_READ_ONLY) == 0) {
		fDevice = open(deviceName, O_RDONLY);
		fFlags |= VOLUME_READ_ONLY;
	}

	if (fDevice < B_OK)
		RETURN_ERROR(fDevice);

	// check if it's a regular file, and if so, disable the cache for the
	// underlaying file system
	struct stat stat;
	if (fstat(fDevice, &stat) < 0)
		RETURN_ERROR(B_ERROR);

//#ifndef USER
	if (stat.st_mode & S_FILE && ioctl(fDevice, IOCTL_FILE_UNCACHED_IO, NULL) < 0) {
		// mount read-only if the cache couldn't be disabled
#	ifdef DEBUG
		FATAL(("couldn't disable cache for image file - system may dead-lock!\n"));
#	else
		FATAL(("couldn't disable cache for image file!\n"));
		Panic();
#	endif
	}
//#endif

	// read the super block
	uchar buffer[1024];
	if (read_pos(fDevice, 0, buffer, sizeof(buffer)) != sizeof(buffer))
		return B_IO_ERROR;

	status_t status = B_OK;

	// Note: that does work only for x86, for PowerPC, the super block
	// is located at offset 0!
	memcpy(&fSuperBlock, buffer + 512, sizeof(disk_super_block));

	if (IsValidSuperBlock()) {
		// set the current log pointers, so that journaling will work correctly
		fLogStart = fSuperBlock.log_start;
		fLogEnd = fSuperBlock.log_end;

		if (init_cache_for_device(fDevice, NumBlocks()) == B_OK)
		{
			fJournal = new Journal(this);
			
			// replaying the log is the first thing we will do on this disk
			if (fJournal &&
				fJournal->InitCheck() == B_OK &&
				fBlockAllocator.Initialize(true) == B_OK)
			{
				fRootNode = new Inode(this, ToVnode(Root()));

				if (fRootNode && fRootNode->InitCheck() == B_OK)
				{
					if (new_vnode(fID, ToVnode(Root()),(void *)fRootNode) == B_OK)
					{
						// try to get indices root dir

						// question: why doesn't get_vnode() work here??
						// answer: we have not yet backpropagated the pointer to the
						// volume in bfs_mount(), so bfs_read_vnode() can't get it.
						// But it's not needed to do that anyway.

						fIndicesNode = new Inode(this, ToVnode(Indices()));
						if (fIndicesNode == NULL
							|| fIndicesNode->InitCheck() < B_OK
							|| !fIndicesNode->IsContainer())
						{
							_asm int 3
							INFORM(("bfs: volume doesn't have indices!\n"));

							if (fIndicesNode) {
								// if this is the case, the index root node is gone bad, and
								// BFS switch to read-only mode
								fFlags |= VOLUME_READ_ONLY;
								fIndicesNode = NULL;
							}
						}

						// all went fine
						return B_OK;
					}
					else
						status = B_NO_MEMORY;
				}
				else
				{
					assert(0);
					status = B_BAD_VALUE;
				}

				FATAL(("could not create root node: new_vnode() failed!\n"));
			} else {
				// ToDo: improve error reporting for a bad journal
				status = B_NO_MEMORY;
				FATAL(("could not initialize journal/block bitmap allocator!\n"));
			}

			remove_cached_device_blocks(fDevice, NO_WRITES);
		} else {
			FATAL(("could not initialize cache!\n"));
			status = B_IO_ERROR;
		}
		FATAL(("invalid super block!\n"));
	}
	else
		status = B_BAD_VALUE;

	close(fDevice);

	return status;
}


status_t
Volume::Unmount()
{
	// This will also flush the log & all blocks to disk
	delete fJournal;
	fJournal = NULL;

	delete fIndicesNode;

	remove_cached_device_blocks(fDevice, IsReadOnly() ? NO_WRITES : ALLOW_WRITES);
	close(fDevice);

	return B_OK;
}


status_t 
Volume::Sync()
{
	return fJournal->FlushLogAndBlocks();
}


status_t
Volume::ValidateBlockRun(block_run run)
{
	if (run.allocation_group < 0 || run.allocation_group > (int32)AllocationGroups()
		|| run.start > (1UL << AllocationGroupShift())
		|| run.length == 0
		|| uint32(run.length + run.start) > (1UL << AllocationGroupShift())) {
		Panic();
		FATAL(("*** invalid run(%ld,%d,%d)\n", run.allocation_group, run.start, run.length));
		return B_BAD_DATA;
	}
	return B_OK;
}


block_run 
Volume::ToBlockRun(off_t block) const
{
	block_run run;
	run.allocation_group = block >> fSuperBlock.ag_shift;
	run.start = block & ~((1 << fSuperBlock.ag_shift) - 1);
	run.length = 1;
	return run;
}


status_t
Volume::CreateIndicesRoot(Transaction *transaction)
{
	off_t id;
	status_t status = Inode::Create(transaction, NULL, NULL,
		S_INDEX_DIR | S_STR_INDEX | S_DIRECTORY | 0700, 0, 0, &id, &fIndicesNode);
	if (status < B_OK)
		RETURN_ERROR(status);

	fSuperBlock.indices = ToBlockRun(id);
	return WriteSuperBlock();
}


status_t 
Volume::AllocateForInode(Transaction *transaction, const Inode *parent, mode_t type, block_run &run)
{
	return fBlockAllocator.AllocateForInode(transaction, &parent->BlockRun(), type, run);
}


status_t 
Volume::WriteSuperBlock()
{
	if (write_pos(fDevice, 512, (uchar*) &fSuperBlock, sizeof(disk_super_block)) != sizeof(disk_super_block))
		return B_IO_ERROR;

	return B_OK;
}


void
Volume::UpdateLiveQueries(Inode *inode, const char *attribute, int32 type, const uint8 *oldKey,
	size_t oldLength, const uint8 *newKey, size_t newLength)
{
	if (fQueryLock.Lock() < B_OK)
		return;

	Query *query = NULL;
	while ((query = fQueries.Next(query)) != NULL)
		query->LiveUpdate(inode, attribute, type, oldKey, oldLength, newKey, newLength);

	fQueryLock.Unlock();
}


void 
Volume::AddQuery(Query *query)
{
	if (fQueryLock.Lock() < B_OK)
		return;

	fQueries.Add(query);

	fQueryLock.Unlock();
}


void 
Volume::RemoveQuery(Query *query)
{
	if (fQueryLock.Lock() < B_OK)
		return;

	fQueries.Remove(query);

	fQueryLock.Unlock();
}

