/**
	\file
	\author Matthew Allen
	\date 25/3/97
	\brief Graphics file filters
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "Gdc2.h"
#include "GString.h"

#ifndef WIN32
#define BI_RLE8       1L
#define BI_RLE4       2L
#else
#include "Lgi.h"
#endif

int FindHeader(int Offset, char *Str, GFile &f)
{
	int i = 0;
	
	if (Offset >= 0)
	{
		f.Seek(Offset, SEEK_SET);
	}
	
	while (NOT f.Eof())
	{
		char c;
		
		f >> c;
		if (Str[i] == c || Str[i] == '?')
		{
			i++;
			if (!Str[i])
			{
				return true;
			}
		}
		else
		{
			i = 0;
		}
	}

	return false;
}

/// Windows and OS/2 BMP file filter
class GdcBmp : public GFilter
{
	int ActualBits;
	int ScanSize;
	bool SetSize(long *Info, GDeviceContext *pDC);

public:
	GdcBmp();

	int GetCapabilites() { return FILTER_CAP_READ | FILTER_CAP_WRITE; }

	/// Reads a BMP file
	bool ReadImage(GSurface *pDC);
	
	/// Writes a Windows BMP file
	bool WriteImage(GSurface *pDC);
};

class GdcBmpFactory : public GFilterFactory
{
	bool CheckFile(char *File, int Access, uchar *Hint)
	{
		return (File) ? stristr(File, ".bmp") != 0 : false;
	}

	GFilter *NewObject()
	{
		return NEW(GdcBmp);
	}

} BmpFactory;

#define BMP_ID			0x4D42

#ifdef WIN32
#pragma pack(push, before_pack)
#pragma pack(1)
#endif

class BMP_FILE {
public:
	uint16		Type;
	uint32		Size;
	uint16		Reserved1;
	uint16		Reserved2;
	uint32		OffsetToData;
};

class BMP_WININFO {
public:
	int32		Size;
	int32		Sx;
	int32		Sy;
	uint16		Planes;
	uint16		Bits;
	uint32		Compression;
	uint32		DataSize;
	int32		XPels;
	int32		YPels;
	uint32		ColoursUsed;
	uint32		ColourImportant;
};

class BMP_OS2INFO {
public:
	int32		Size;
	uint32		Sx;
	uint32		Sy;
	uint32		Planes;
	uint32		Bits;
};

#ifdef WIN32
#pragma pack(pop, before_pack)
#endif

bool GdcBmp::SetSize(long *Info, GDeviceContext *pDC)
{
	bool Status = false;

	switch (*Info)
	{
		case sizeof(BMP_WININFO):
		{
			BMP_WININFO *i = (BMP_WININFO*) Info;
			ActualBits = i->Bits;
			ScanSize = BMPWIDTH(i->Sx * i->Bits);
			Status = pDC->Create(i->Sx, i->Sy, max(i->Bits, 8), ScanSize);
			break;
		}

		case sizeof(BMP_OS2INFO):
		{
			BMP_OS2INFO *i = (BMP_OS2INFO*) Info;
			ActualBits = i->Bits;
			ScanSize = BMPWIDTH(i->Sx * i->Bits);
			Status = pDC->Create(i->Sx, i->Sy, max(i->Bits, 8), ScanSize);
			Seek(sizeof(BMP_OS2INFO)-sizeof(BMP_WININFO), SEEK_CUR);
			break;
		}
	}

	return Status;
}

bool GdcBmp::ReadImage(GSurface *pDC)
{
	bool Status = false;
	ActualBits = 0;
	ScanSize = 0;
	
	if (pDC)
	{
		BMP_FILE File;
		BMP_WININFO Info;
		GRect Cr;

		*this >> File.Type;
		*this >> File.Size;
		*this >> File.Reserved1;
		*this >> File.Reserved2;
		*this >> File.OffsetToData;

		*this >> Info.Size;
		*this >> Info.Sx;
		*this >> Info.Sy;
		*this >> Info.Planes;
		*this >> Info.Bits;
		*this >> Info.Compression;
		*this >> Info.DataSize;
		*this >> Info.XPels;
		*this >> Info.YPels;
		*this >> Info.ColoursUsed;
		*this >> Info.ColourImportant;

		if (File.Type == BMP_ID)
		{
			if (SetSize((long*) &Info, pDC))
			{
				if (pDC->GetBits() <= 8)
				{
					int Colours = 1 << ActualBits;
					GPalette *Palette = NEW(GPalette);
					if (Palette)
					{
						Palette->SetSize(Colours);
						GdcRGB *Start = (*Palette)[0];
						if (Start)
						{
							Read(Start, sizeof(GdcRGB)*Colours);
							Palette->SwapRAndB();
						}

						Palette->Update();
						pDC->Palette(Palette);
					}
				}

				GBmpMem *pMem = GetSurface(pDC);
				Seek(File.OffsetToData, SEEK_SET);

				Status = true;

				if (Info.Compression == BI_RLE8)
				{
					// 8 bit RLE compressed image
					int Remaining = GetSize() - GetPos();
					uchar *Data = NEW(uchar[Remaining]);
					if (Data)
					{
						if (Read(Data, Remaining) == Remaining)
						{
							int x=0, y=pDC->Y()-1;
							uchar *p = Data;
							bool Done = false;

							while (NOT Done AND p < Data + Remaining)
							{
								uchar Length = *p++;
								uchar Colour = *p++;

								if (Length == 0)
								{
									switch (Colour)
									{
										case 0:
										{
											x = 0;
											y--;
											break;
										}
										case 1:
										{
											Done = true;
											break;
										}
										case 2:
										{
											x += *p++;
											y -= *p++;
											break;
										}
										default:
										{
											// absolute mode
											uchar *Pixel = (*pDC)[y];
											if (Pixel AND y >= 0 AND y < pDC->Y())
											{
												int Len = min(Colour, pDC->X() - x);
												if (Len > 0)
												{
													memcpy(Pixel + x, p, Len);
													x += Colour;
													p += Colour;
												}
											}
											else
											{
												p += Colour;
											}
											if ((int) p & 1) p++;
											break;
										}
									}
								}
								else
								{
									// encoded mode
									uchar *Pixel = (*pDC)[y];
									if (Pixel AND y >= 0 AND y < pDC->Y())
									{
										int Len = min(Length, pDC->X() - x);
										if (Len > 0)
										{
											memset(Pixel + x, Colour, Len);
											x += Length;
										}
									}
								}
							}
						}
					}
				}
				else if (Info.Compression == BI_RLE4)
				{
					// 4 bit RLE compressed image

					// not implemented
					printf("%s:%i - BI_RLE4 not implemented.\n", __FILE__, __LINE__);
				}
				else
				{
					// Progress
					if (Meter)
					{
						Meter->SetDescription("scanlines");
						Meter->SetLimits(0, pMem->y-1);
					}

					// straight
					switch (ActualBits)
					{
						case 1:
						{
							uchar *Buffer = NEW(uchar[ScanSize]);
							if (Buffer)
							{
								for (int y=pMem->y-1; y>=0; y--)
								{
									Status &= Read(Buffer, ScanSize) == ScanSize;
									if (Status)
									{
										uchar Mask = 0x80;
										uchar *d = Buffer;
										for (int x=0; x<pDC->X(); x++)
										{
											pDC->Colour((*d & Mask) ? 1 : 0);
											pDC->Set(x, y);

											Mask >>= 1;
											if (NOT Mask)
											{
												Mask = 0x80;
												d++;
											}
										}
									}
									else
									{
										break;
									}

									if (Meter) Meter->Value(pMem->y-1-y);
								}

								DeleteArray(Buffer);
							}
							break;
						}
						case 4:
						{
							uchar *Buffer = NEW(uchar[ScanSize]);
							if (Buffer)
							{
								for (int y=pMem->y-1; y>=0; y--)
								{
									Status &= Read(Buffer, ScanSize) == ScanSize;

									uchar *d = Buffer;
									for (int x=0; x<pDC->X(); x++)
									{
										if (x & 1)
										{
											pDC->Colour(*d & 0xf);
											d++;
										}
										else
										{
											pDC->Colour(*d >> 4);
										}
										pDC->Set(x, y);
									}

									if (Meter) Meter->Value(pMem->y-1-y);
								}

								DeleteArray(Buffer);
							}
							break;
						}
						case 24:
						#ifdef XWIN
						{
							for (int i=pMem->y-1; i>=0; i--)
							{
								uchar *Ptr = pMem->Base + (pMem->Line * i);
								Status &= Read(Ptr, ScanSize) == ScanSize;

								ulong *Out = ((ulong*)Ptr) + (pDC->X()-1);
								uchar *In = Ptr + ((pDC->X()-1) * 3);
								for (int x=0; x<pDC->X(); x++)
								{
									*Out-- = Rgb32(In[2], In[1], In[0]);
									In -= 3;
								}
								
								if (Meter) Meter->Value(pMem->y-1-i);
							}
							break;
						}
						#endif
						case 8:
						case 16:
						case 32:
						{
							for (int i=pMem->y-1; i>=0; i--)
							{
								Status &= Read(pMem->Base + (pMem->Line * i), ScanSize) == ScanSize;
								if (Meter) Meter->Value(pMem->y-1-i);
							}
							
							break;
						}
						default:
						{
							Status = false;
							printf("%s:%i - Bmp had unsupported bit depth.\n", __FILE__, __LINE__);
							break;
						}
					}
				}

				Cr.ZOff(pDC->X()-1, pDC->Y()-1);
				pDC->Update(GDC_BITS_CHANGE|GDC_PAL_CHANGE);
			}
			else
			{
				Cr.ZOff(-1, -1);
				printf("%s:%i - SetSize failed.\n", __FILE__, __LINE__);
			}
		}
		else
		{
			Cr.ZOff(-1, -1);
			printf("%s:%i - Bmp type check failed '%.2s'.\n", __FILE__, __LINE__, &File.Type);
		}

		pDC->ClipRgn(&Cr);
	}
	
	return Status;
}

bool GdcBmp::WriteImage(GSurface *pDC)
{
	bool Status = false;
	
	if (pDC)
	{
		BMP_FILE File;
		BMP_WININFO Info;
		GBmpMem *pMem = GetSurface(pDC);
		int Colours = (pMem->Bits <= 8) ? 1 << pMem->Bits : 0;
		int UsedBits = pMem->Bits;

		GPalette *Palette = pDC->Palette();
		if (pMem->Bits <= 8 AND
			Palette)
		{
			int Size = Palette->GetSize();
			int Bits = 8;
			for (int c=256; c; c>>=1, Bits--)
			{
				if (c & Size)
				{
					Colours = c;
					UsedBits = Bits;
					break;
				}
			}
		}

		if (NOT pMem OR
			NOT pMem->x OR
			NOT pMem->y)
		{
			return false;
		}

		GFile::SetSize(0);

		File.Type = BMP_ID;
		File.OffsetToData = sizeof(BMP_FILE) + sizeof(BMP_WININFO) + (Colours * 4);
		File.Size = File.OffsetToData + (pMem->Line * pMem->y);
		File.Reserved1 = 0;
		File.Reserved2 = 0;

		Info.Size = sizeof(BMP_WININFO);
		Info.Sx = pMem->x;
		Info.Sy = pMem->y;
		Info.Planes = 1;
		Info.Bits = UsedBits;
		Info.Compression = 0;
		Info.DataSize = 0;
		Info.XPels = 3000;
		Info.YPels = 3000;
		Info.ColoursUsed = Colours;
		Info.ColourImportant = Colours;

		if (Write(&File, sizeof(File)) AND
			Write(&Info, sizeof(Info)))
		{
			if (pMem->Bits <= 8)
			{
				int Done = 0;

				if (Palette)
				{
					GdcRGB *Start = (*Palette)[0];
					if (Start)
					{
						Palette->SwapRAndB();
						Write(Start, sizeof(GdcRGB)*Palette->GetSize());
						Palette->SwapRAndB();
						Done += Palette->GetSize();
					}
				}

				char Grey[4];
				Grey[3] = 0;

				while (Done < Colours)
				{
					int C = Done * 256 / Colours;

					Grey[0] = C;
					Grey[1] = C;
					Grey[2] = C;

					Write(Grey, 4);
					Done++;
				}
			}

			// Progress
			if (Meter)
			{
				Meter->SetDescription("scanlines");
				Meter->SetLimits(0, pMem->y-1);
			}

			int Bytes = BMPWIDTH(pMem->x * UsedBits);
			Status = true;
			switch (UsedBits)
			{
				case 1:
				{
					uchar *Buffer = NEW(uchar[Bytes]);
					if (Buffer)
					{
						for (int i=pMem->y-1; i>=0; i--)
						{
							uchar *Src = pMem->Base + (i * pMem->Line);
							
							// assemble the bits
							memset(Buffer, 0, Bytes);
							for (int x=0; x<pMem->x; x++)
							{
								Buffer[x>>3] |= (Src[x] & 1) << (x & 3);
							}

							// write the bits
							Status &= Write(Buffer, Bytes) == Bytes;
							
							// update status
							if (Meter) Meter->Value(pMem->y-1-i);
						}

						DeleteArray(Buffer);
					}
					break;
				}
				case 4:
				{
					uchar *Buffer = NEW(uchar[Bytes]);
					if (Buffer)
					{
						for (int i=pMem->y-1; i>=0; i--)
						{
							uchar *Src = pMem->Base + (i * pMem->Line);

							// assemble the nibbles
							for (int x=0; x<pMem->x; x+=2)
							{
								Buffer[x>>1] = (Src[x]<<4) | (Src[x+1]&0x0f);
							}

							// write the line
							Status &= Write(Buffer, Bytes) == Bytes;
							
							// update status
							if (Meter) Meter->Value(pMem->y-1-i);
						}

						DeleteArray(Buffer);
					}
					break;
				}
				default:
				{
					for (int i=pMem->y-1; i>=0; i--)
					{
						Status &= Write(pMem->Base + (i * pMem->Line), Bytes) == Bytes;
						if (Meter) Meter->Value(pMem->y-1-i);
					}
					break;
				}
			}
		}
	}

	Close();
	
	return Status;
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
// ICO file format
class GdcIco : public GFilter
{
	int TruncSize(int i)
	{
		if (i >= 64) return 64;
		if (i >= 32) return 32;
		return 16;
	}

public:
	GdcIco();

	int GetCapabilites() { return FILTER_CAP_READ | FILTER_CAP_WRITE; }
	int GetImages() { return 1; }
	bool ReadImage(GSurface *pDC);
	bool WriteImage(GSurface *pDC);
};

class GdcIcoFactory : public GFilterFactory
{
	bool CheckFile(char *File, int Access, uchar *Hint)
	{
		return (File) ? stristr(File, ".ico") != 0 : false;
	}

	GFilter *NewObject()
	{
		return NEW(GdcIco);
	}

} IcoFactory;

GdcIco::GdcIco()
{
	Props.Set("Type", "Icon file");
	Props.Set("Extension", "ICO");
}

bool GdcIco::ReadImage(GSurface *pDC)
{
	bool Status = false;
	int MyBits = 0;

	int16 Reserved_1;
	int16 Type;
	int16 Count;

	*this >> Reserved_1;
	*this >> Type;
	*this >> Count;

	for (int Image = 0; Image < Count; Image++)
	{
		int8	Width;
		int8	Height;
		int8	ColorCount;
		int8	Reserved_2;
		int16	Planes;
		int16	BitCount;
		int32	BytesInRes;
		int32	ImageOffset;
		
		*this >> Width;
		*this >> Height;
		*this >> ColorCount;
		*this >> Reserved_2;
		*this >> Planes;
		*this >> BitCount;
		*this >> BytesInRes;
		*this >> ImageOffset;
		
		int BytesLeft = BytesInRes;
		
		int CurrentPos = GetPos();
		Seek(ImageOffset, SEEK_SET);

		BMP_WININFO	Header;
		GdcRGB *Colours;
		uchar *XorBytes = 0;
		uchar *AndBytes = 0;

		LgiAssert(sizeof(Header) == 40);
		Read(&Header, sizeof(Header));
		BytesLeft -= sizeof(Header);

		int Cols = 1 << (Header.Bits);
		Colours = NEW(GdcRGB[Cols]);
		if (Colours)
		{
			Read(Colours, sizeof(GdcRGB) * Cols);
			BytesLeft -= sizeof(GdcRGB) * Cols;
		}

		Header.Sy >>= 1;
		
		int XorSize = BMPWIDTH(Header.Sx * Header.Bits) * Height;
		int AndSize = BMPWIDTH(Header.Sx) * Height;
		
		if (BytesLeft >= XorSize)
		{
			XorBytes = NEW(uchar[XorSize]);
			if (XorBytes)
			{
				Read(XorBytes, XorSize);
				BytesLeft -= XorSize;
			}
		}
		
		if (BytesLeft >= AndSize)
		{
			AndBytes = NEW(uchar[AndSize]);
			if (AndBytes)
			{
				Read(AndBytes, AndSize);
				BytesLeft -= AndSize;
			}
		}

		/*
		printf("BytesInRes=%i, Xor=%i, And=%i, Pal=%i, BytesLeft=%i\n",
			BytesInRes,
			XorSize,
			AndSize,
			sizeof(GdcRGB) * Cols,
			BytesLeft);
		*/

		if (Colours AND
			XorBytes AND
			(Header.Bits > MyBits OR Width > pDC->X() OR Height > pDC->Y()) AND
			pDC->Create(Width, Height, max(8, Header.Bits) ))
		{
			MyBits = Header.Bits;
			pDC->Colour(0, 24);
			pDC->Rectangle();

			GPalette *Pal = NEW(GPalette);
			if (Pal)
			{
				Pal->SetSize(Cols);
				memcpy((*Pal)[0], Colours, sizeof(GdcRGB) * Cols);
				Pal->SwapRAndB();
				pDC->Palette(Pal, true);
			}
			
			if (AndBytes)
			{
				pDC->IsAlpha(true);
			}			

			GSurface *pAlpha = pDC->AlphaDC();
			int XorLine = XorSize / Height;
			int AndLine = AndSize / Height;
			for (int y=0; y<Height; y++)
			{
				uchar *Dest = (*pDC)[y];
				uchar *Alpha = pAlpha ? (*pAlpha)[y] : 0;
				uchar *Src = XorBytes + ((Height-y-1) * XorLine);
				uchar *Mask = AndBytes ? AndBytes + ((Height-y-1) * AndLine) : 0;

				switch (Header.Bits)
				{
					case 1:
					{
						uchar m = 0x80;
						for (int x=0; x<Width; x++)
						{
							if (Alpha)
							{
								Alpha[x] = (Mask[x>>3] & (0x80 >> (x&7))) ? 0 : 0xff;
							}

							if (Src[x/8] & m)
							{
								Dest[x] = 1;
							}
							else
							{
								Dest[x] = 0;
							}
							m >>= 1;
							if (NOT m) m = 0x80;
						}
						Status = true;
						break;
					}
					case 4:
					{
						for (int x=0; x<Width; x++)
						{
							if (Alpha)
							{
								Alpha[x] = (Mask[x>>3] & (0x80 >> (x&7))) ? 0 : 0xff;
							}

							if (x & 1)
							{
								Dest[x] = Src[x>>1] & 0xf;
							}
							else
							{
								Dest[x] = Src[x>>1] >> 4;
							}
						}
						Status = true;
						break;
					}
					case 8:
					{
						for (int x=0; x<Width; x++)
						{
							if (Alpha)
							{
								Alpha[x] = (Mask[x>>3] & (0x80 >> (x&7))) ? 0 : 0xff;
							}

							Dest[x] = Src[x];
						}

						Status = true;
						break;
					}
				}
			}
		}
		else
		{
			printf("%s:%i - Header size error: %i != %i + %i, Img: %ix%i @ %i bits\n",
				__FILE__, __LINE__,
				Header.DataSize, XorSize, AndSize,
				Header.Sx, Header.Sy, Header.Bits);
		}

		DeleteArray(Colours);
		DeleteArray(XorBytes);
		DeleteArray(AndBytes);
		Seek(CurrentPos, SEEK_SET);
	}

	return Status;
}

bool GdcIco::WriteImage(GSurface *pDC)
{
	bool Status = false;

	if (pDC AND
		pDC->GetBits() <= 8)
	{
		SetSize(0);

		int ActualBits = pDC->GetBits();
		GPalette *Pal = pDC->Palette();
		if (Pal)
		{
			if (Pal->GetSize() <= 2)
			{
				ActualBits = 1;
			}
			else if (Pal->GetSize() <= 16)
			{
				ActualBits = 4;
			}
		}

		// write file header
		int Colours = 1 << ActualBits;
		int16 Reserved_1 = 0;
		int16 Type = 1;
		int16 Count = 1;

		*this << Reserved_1;
		*this << Type;
		*this << Count;

		// write directory list
		int8	Width = TruncSize(pDC->X());
		int8	Height = TruncSize(pDC->Y());
		int8	ColorCount = Colours;
		int8	Reserved_2 = 0;
		int16	Planes = 0;
		int16	BitCount = 0;

		int Line = (ActualBits * Width) / 8;
		int MaskLine = BMPWIDTH(Width/8);
		int32	BytesInRes = sizeof(BMP_WININFO) +
							(sizeof(GdcRGB) * Colours) +
							(Line * Height) +
							(MaskLine * Height);
		
		*this << Width;
		*this << Height;
		*this << ColorCount;
		*this << Reserved_2;
		*this << Planes;
		*this << BitCount;
		*this << BytesInRes;

		int32	ImageOffset = GetPos() + sizeof(ImageOffset);
		*this << ImageOffset;

		// Write icon itself
		BMP_WININFO	Header;
		LgiAssert(sizeof(Header) == 40);
		Header.Size = sizeof(Header);
		Header.Sx = Width;
		Header.Sy = Height * 2;
		Header.Planes = 1;
		Header.Bits = ActualBits;
		Header.Compression = 0; // BI_RGB;
		Header.DataSize = (Line * Height) + (MaskLine * Height);
		Header.XPels = 0;
		Header.YPels = 0;
		Header.ColoursUsed = 0;
		Header.ColourImportant = 0;
		Write(&Header, sizeof(Header));

		// Write palette
		if (Pal)
		{
			Pal->SwapRAndB();
			Write((*Pal)[0], sizeof(GdcRGB) * (1 << ActualBits));
			Pal->SwapRAndB();
		}

		// Get background colour
		COLOUR Back = 0xffffffff;
		Props.Get("back", (int&)Back);

		// Write "Colour" bits
		int y;
		for (y=Height-1; y>=0; y--)
		{
			uchar *Src = (*pDC)[y];

			// xor bits (colour)
			switch (ActualBits)
			{
				case 4:
				{
					uchar Dest = 0;
					for (int x=0; x<Width; x++)
					{
						if (x&1)
						{
							Dest |=	(Back == Src[x]) ? 0 : Src[x] & 0xf;
							*this << Dest;
							Dest = 0;
						}
						else
						{
							Dest |=	(Back == Src[x]) ? 0 : (Src[x] & 0xf) << 4;
						}
					}

					Status = true;
					break;
				}
				case 8:
				{
					for (int x=0; x<Width; x++)
					{
						uchar Dest = (Back == Src[x]) ? 0 : Src[x];
						*this << Dest;
					}
					Status = true;
					break;
				}
			}
		}

		// Write "And" bits (mask)
		for (y=Height-1; y>=0; y--)
		{
			uchar *Src = (*pDC)[y];
			uchar Dest = 0;
			uchar Mask = 0x80;
			
			int x;
			for (x = 0; x < Width; x++)
			{
				if (Src[x] == Back) Dest |= Mask;
				Mask >>= 1;
				if (NOT Mask)
				{
					*this << Dest;
					Dest = 0;
					Mask = 0x80;
				}
			}

			x = Width / 8;
			while (x < MaskLine)
			{
				uchar c = 0;
				*this << c;
				x++;
			}

			Status = true;
		}
	}

	return Status;
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
// Run length encoded DC
GdcRleDC::GdcRleDC()
{
	Flags = GDC_RLE_COLOUR | GDC_RLE_READONLY;
	Length = 0;
	Alloc = 0;
	Data = 0;
	ScanLine = 0;
	Key = 0;

	pMem = NEW(GBmpMem);
	if (pMem)
	{
		pMem->Base = 0;
		pMem->x = 0;
		pMem->y = 0;
		pMem->Bits = 0;
		pMem->Line = 0;
		pMem->Flags = 0;
	}
}

GdcRleDC::~GdcRleDC()
{
	Empty();
}

void GdcRleDC::Empty()
{
	Length = 0;
	Alloc = 0;
	DeleteArray(Data);
	DeleteArray(ScanLine);
}

bool GdcRleDC::Create(int x, int y, int bits, int LineLen)
{
	bool Status = GMemDC::Create(x, y, bits, LineLen);
	if (Status)
	{
		Flags &= ~GDC_RLE_READONLY;
	}
	return Status;
}

bool GdcRleDC::CreateInfo(int x, int y, int bits)
{
	bool Status = false;

	DeleteObj(pMem);
	pMem = NEW(GBmpMem);
	if (pMem)
	{
		pMem->x = x;
		pMem->y = y;
		pMem->Bits = bits;
		pMem->Base = 0;
		pMem->Line = 0;
		pMem->Flags = 0;

		Flags |= GDC_RLE_READONLY;

		if (bits == 8)
		{
			Flags |= GDC_RLE_MONO;
		}
	}

	return Status;
}

void GdcRleDC::ReadOnly(bool Read)
{
	if (Read)
	{
		Flags |= GDC_RLE_READONLY;
	}
	else
	{
		if (pMem)
		{
			pMem->Base = NEW(uchar[pMem->Line * pMem->y]);
		}
		else
		{
			// just in case... shouldn't ever happen
			// am I paranoid?
			// 'course :)
			Create(1, 1, 24);
		}

		Flags &= ~GDC_RLE_READONLY;
	}
}

bool GdcRleDC::ReadOnly()
{
	return (Flags & GDC_RLE_READONLY) != 0;
}

void GdcRleDC::Mono(bool Mono)
{
	if (Mono)
	{
		Flags |= GDC_RLE_MONO;
		Flags &= ~GDC_RLE_COLOUR;
	}
	else
	{
		Flags &= ~GDC_RLE_MONO;
		Flags |= GDC_RLE_COLOUR;
	}
}

bool GdcRleDC::Mono()
{
	return (Flags & GDC_RLE_MONO) != 0;
}

bool GdcRleDC::SetLength(int Len)
{
	bool Status = true;

	if (Len + 1 > Alloc)
	{
		int NewAlloc = (Len + 0x4000) & (~0x3FFF);
		uchar *Temp = NEW(uchar[NewAlloc]);
		if (Temp)
		{
			if (NewAlloc > Length)
			{
				memset(Temp + Length, 0, NewAlloc - Length);
			}

			memcpy(Temp, Data, Length);
			DeleteArray(Data);
			DeleteArray(ScanLine);
			Data = Temp;
			Length = Len;
			Alloc = NewAlloc;
		}
		else
		{
			Status = false;
		}
	}
	else
	{
		// TODO: resize Data to be smaller if theres a
		// considerable size change
		Length = Len;
	}

	return Status;
}

void GdcRleDC::Update(int UpdateFlags)
{
	bool Error = false;

	if (	(UpdateFlags & GDC_BITS_CHANGE) AND
		NOT (Flags & GDC_RLE_READONLY))
	{
		Empty();

		COLOUR Key = Get(0, 0);
		ulong Pos = 0;
		ulong PixelSize = GetBits() / 8;
		
		for (ulong y=0; NOT Error AND y<Y(); y++)
		{
			for (ulong x=0; x<X(); )
			{
				ulong Skip = 0;
				ulong Pixels = 0;
				uchar *Bits = 0;

				while (Get(x, y) == Key AND x < X())
				{
					Skip++;
					x++;
				}

				Bits = (*this)[y] + (x * PixelSize);
				while (Get(x, y) != Key AND x < X())
				{
					Pixels++;
					x++;
				}

				if (Flags & GDC_RLE_COLOUR)
				{
					if (SetLength(	Length +
							(2 * sizeof(ulong)) +
							(Pixels * PixelSize)))
					{
						*((ulong*)(Data+Pos)) = Skip;
						Pos += sizeof(Skip);
						*((ulong*)(Data+Pos)) = Pixels;
						Pos += sizeof(Pixels);
						
						int l = Pixels * PixelSize;
						memcpy(Data+Pos, Bits, l);
						Pos += l;
					}
					else
					{
						Error = true;
						break;
					}
				}
				else if (Flags & GDC_RLE_MONO)
				{
					if (SetLength(	Length + (2 * sizeof(ulong))))
					{
						*((ulong*)(Data+Pos)) = Skip;
						Pos += sizeof(Skip);
						*((ulong*)(Data+Pos)) = Pixels;
						Pos += sizeof(Pixels);
					}
					else
					{
						Error = true;
						break;
					}
				}
			}
		}
	}

	if (Error)
	{
		Empty();
	}
	else
	{
		FindScanLines();
	}
}

bool GdcRleDC::FindScanLines()
{
	bool Status = false;
	bool Error = false;
	DeleteArray(ScanLine);

	if (Data)
	{
		ScanLine = NEW(uchar*[Y()]);
		if (ScanLine)
		{
			ulong Pos = 0;
			ulong PixelSize = GetBits() / 8;
			
			for (ulong y=0; NOT Error AND y<Y(); y++)
			{
				ScanLine[y] = Data + Pos;
				ulong x;
				for (x=0; x<X();)
				{
					ulong Skip = 0;
					ulong Pixels = 0;

					Skip = *((ulong*)(Data + Pos));
					Pos += sizeof(Skip);
					Pixels = *((ulong*)(Data + Pos));
					Pos += sizeof(Pixels);

					if (Flags & GDC_RLE_COLOUR)
					{
						Pos += Pixels * PixelSize;
					}

					x += Pixels + Skip;
				}

				if (x != X() OR Pos > Length)
				{
					Error = true;
				}
			}
		}

		if (Error)
		{
			DeleteArray(ScanLine);
		}
		else
		{
			Status = true;
		}
	}

	return Status;
}

void GdcRleDC::Draw(GSurface *Dest, int Ox, int Oy)
{
	int OriX, OriY;
	Dest->GetOrigin(OriX, OriY);
	Ox += OriX;
	Oy += OriY;

	if (Dest AND Data)
	{
		bool ReBuildScanInfo = false;
		
		if (NOT ScanLine) ReBuildScanInfo = true;
		if (ScanLine AND Data AND ScanLine[0] != Data) ReBuildScanInfo = true;

		if (ReBuildScanInfo)
		{
			// need to rebuild scan line info
			if (NOT FindScanLines())
			{
				// couldn't create scan line info so quit
				return;
			}
		}

		GRect S;
		S.ZOff(X()-1, Y()-1);
		GRect D;
		D = S;
		D.Offset(Ox, Oy);
		GRect DClip = D;
		GRect Temp = Dest->ClipRgn();
		D.Bound(&Temp);

		if (DClip == D AND (*Dest)[0])
		{
			// no clipping needed
			int PixLen = Dest->GetBits() / 8;

			if (Flags & GDC_RLE_COLOUR)
			{
				if (Dest->GetBits() == GetBits())
				{
					for (int y=0; y<Y(); y++)
					{
						uchar *s = ScanLine[y];
						uchar *d = (*Dest)[Oy+y];
						for (int x=0; x<X(); )
						{
							x += *((ulong*)s); s+=4;
							int Pixels = *((ulong*)s); s+=4;
							int Len = Pixels*PixLen;
							memcpy(d + ((Ox+x)*PixLen), s, Len);
							x += Pixels;
							s += Len;
						}
					}
				}
			}
			else if (Flags & GDC_RLE_MONO)
			{
				GApplicator *pDApp = Dest->Applicator();

				for (int y=0; y<Y(); y++)
				{
					uchar *s = ScanLine[y];
					for (int x=0; x<X(); )
					{
						x += *((ulong*)s); s+=4;
						int Pixels = *((ulong*)s); s+=4;
						pDApp->SetPtr(Ox+x, Oy+y);
						pDApp->Rectangle(Pixels, 1);
						x += Pixels;
					}
				}
			}
		}
		else if (DClip.Valid())
		{
			// clip
			int PixLen = Dest->GetBits() / 8;

			if (Flags & GDC_RLE_COLOUR)
			{
				if (Dest->GetBits() == GetBits())
				{
					for (int y=0; y<Y(); y++)
					{
						int Fy = Oy+y;
						if (Fy >= Temp.y1 AND Fy < Temp.y2) // clip y
						{
							uchar *s = ScanLine[y];
							uchar *d = (*Dest)[Fy];
							for (int x=0; x<X(); )
							{
								x += *((ulong*)s); s+=4;
								int Pixels = *((ulong*)s); s+=4;

								int Fx = Ox+x;
								int PreClipPixels = max(0, Temp.x1 - Fx);
								int PostClipPixels = max(0, (Pixels + Fx) - Temp.x2 - 1);
								int PixelsLeft = Pixels - (PreClipPixels + PostClipPixels);
								if (PixelsLeft > 0) // clip x
								{
									memcpy(	d + ((Fx - PreClipPixels) * PixLen),
										s + (PreClipPixels * PixLen),
										PixelsLeft * PixLen);
								}

								x += Pixels;
								s += Pixels * PixLen;
							}
						}
					}
				}
			}
			/*
			else if (Flags & GDC_RLE_MONO)
			{
				GApplicator *pDApp = Dest->Applicator();

				for (int y=0; y<Y(); y++)
				{
					uchar *s = ScanLine[y];
					for (int x=0; x<X(); )
					{
						x += *((ulong*)s); s+=4;
						int Pixels = *((ulong*)s); s+=4;
						pDApp->SetPtr(Ox+x, Oy+y);
						pDApp->Rectangle(Pixels, 1);
						x += Pixels;
					}
				}
			}
			*/
		}
	}
}

int GdcRleDC::SizeOf()
{
	return (sizeof(int) * 5) + Length;
}

bool GdcRleDC::Read(GFile &F)
{
	Empty();
	
	int32 Len = 0, x = 0, y = 0, bits = 0;
	int8 Monochrome;
	F >> Len;
	if (SetLength(Len))
	{
		F >> Key;
		F >> x;
		F >> y;
		F >> bits;
		F >> Monochrome;
		Mono(Monochrome);
		CreateInfo(x, y, bits);
		F.Read(Data, Len);
		return NOT F.GetStatus();
	}

	return false;
}

bool GdcRleDC::Write(GFile &F)
{
	if (Data)
	{
		char Monochrome = Mono();
		
		F << Length;
		F << Key;
		F << X();
		F << Y();
		F << GetBits();
		F << Monochrome;

		F.Write(Data, Length);

		return NOT F.GetStatus();
	}

	return false;
}

void GdcRleDC::Move(GdcRleDC *pDC)
{
	if (pDC)
	{
		Key = pDC->Key;
		Flags = pDC->Flags;
		Length = pDC->Length;
		Alloc = pDC->Alloc;
		Data = pDC->Data;
		ScanLine = pDC->ScanLine;
		pMem = pDC->pMem;
		pPalette = pDC->pPalette;

		pDC->pPalette = 0;
		pDC->pMem = 0;
		pDC->Length = 0;
		pDC->Alloc = 0;
		pDC->Data = 0;
		pDC->ScanLine = 0;
	}
}

////////////////////////////////////////////////////////////////////
GFilterFactory *GFilterFactory::First = 0;
GFilterFactory::GFilterFactory()
{
	// add this filter to the global list
	Next = First;
	First = this;
}

GFilterFactory::~GFilterFactory()
{
	// delete from the global list
	if (First == this)
	{
		First = First->Next;
	}
	else
	{
		GFilterFactory *i = First;
		while (i->Next AND i->Next != this)
		{
			i = i->Next;
		}
		if (i->Next == this)
		{
			i->Next = Next;
		}
		else
		{
			// we aren't in the list
			// thats bad
			LgiAssert(0);
		}
	}
}

GFilter *GFilterFactory::New(char *File, int Access, uchar *Hint)
{
	GFilterFactory *i = First;
	while (i)
	{
		if (i->CheckFile(File, Access, Hint))
		{
			return i->NewObject();
		}
		i = i->Next;
	}
	return 0;
}

GFilter *GFilterFactory::NewAt(int n)
{
	int Status = 0;
	GFilterFactory *i = First;
	while (i)
	{
		if (Status++ == n)
		{
			return i->NewObject();
		}

		i = i->Next;
	}

	return 0;
}

int GFilterFactory::GetItems()
{
	int Status = 0;
	GFilterFactory *i = First;
	while (i)
	{
		Status++;
		i = i->Next;
	}
	return Status;
}

//////////////////////////////////////////////////////////////////////////
GSurface *LoadDC(char *Name)
{
	GSurface *pDC = 0;

	uchar *Hint = NEW(uchar[16]);
	if (Hint)
	{
		GFile f;
		if (f.Open(Name, O_READ))
		{
			memset(Hint, 0, 16);
			if (f.Read(Hint, 16) == 0)
			{
				DeleteArray(Hint);
			}
		}
		else DeleteArray(Hint);
	}

	GFilter *F = GFilterFactory::New(Name, FILTER_CAP_READ, Hint);
	if (F)
	{
		if (F->Open(Name, O_READ))
		{
			// a file...
			pDC = NEW(GMemDC);
			if (pDC)
			{
				if (NOT F->ReadImage(pDC))
				{
					DeleteObj(pDC);
					printf("%s:%i - Filter couldn't cope with '%s'.\n", __FILE__, __LINE__, Name);
				}
			}
		}
		else
		{
			printf("%s:%i - Couldn't open '%s' for reading.\n", __FILE__, __LINE__, Name);
		}
		
		DeleteObj(F);
	}
	else
	{
		printf("%s:%i - Couldn't find filter for '%s'\n", __FILE__, __LINE__, Name);
	}
	
	DeleteArray(Hint);
	
	#if defined WIN32
	if (NOT pDC)
	{
		// a resource... lock and load gentlemen
		HRSRC hRsrc = FindResource(NULL, Name, RT_BITMAP);
		if (hRsrc)
		{
			int Size = SizeofResource(NULL, hRsrc);
			HGLOBAL hMem = LoadResource(NULL, hRsrc);
			if (hMem)
			{
				uchar *Ptr = (uchar*) LockResource(hMem);
				if (Ptr)
				{
					GClipBoard Clip(NULL);
					pDC = Clip.ConvertFromPtr(Ptr);
					GlobalUnlock(hMem);
				}
			}
		}
	}
	#else
	if (pDC)
	{
		int PromoteTo = GdcD->GetOption(GDC_PROMOTE_ON_LOAD);

		if (PromoteTo > 0 AND
			PromoteTo != pDC->GetBits())
		{
			GSurface *pOld = pDC;
			GPalette *pPal = pDC->Palette();

			// printf("LoadDC, Promoting depth from %i to %i\n", pDC->GetBits(), PromoteTo);

			pDC = NEW(GMemDC);
			if (pOld AND
				pDC AND
				pDC->Create(pOld->X(), pOld->Y(), PromoteTo))
			{
				pDC->Blt(0, 0, pOld);
				DeleteObj(pOld);
			}
		}
		#ifdef BEOS
		else if (pDC->GetBits() == 8)
		{
			// Create remap table
			color_map *Map = (color_map*) system_colors();
			GPalette *Palette = pDC->Palette();
			if (Map AND Palette)
			{
				char ReMap[256];
				for (int i=0; i<256; i++)
				{
					GdcRGB *p = (*Palette)[i];
					if (p)
					{
						COLOUR c = Rgb15(p->R, p->G, p->B);
						ReMap[i] = Map->index_map[c];

						p->R = Map->color_list[i].red;
						p->G = Map->color_list[i].green;
						p->B = Map->color_list[i].blue;
					}
					else
					{
						ReMap[i] = 0;
					}
				}

				// Remap colours to BeOS palette
				for (int y=0; y<pDC->Y(); y++)
				{
					uchar *d = (*pDC)[y];
					if (d)
					{
						for (int x=0; x<pDC->X(); x++)
						{
							d[x] = ReMap[d[x]];
						}
					}
				}
			}
		}
		#endif
	}
	#endif

	return pDC;
}

bool WriteDC(char *Name, GSurface *pDC)
{
	bool Status = false;

	if (Name AND pDC)
	{
		GFilter *F = GFilterFactory::New(Name, FILTER_CAP_WRITE, 0);
		if (F AND F->Open(Name, O_WRITE))
		{
			// a file...
			Status = F->WriteImage(pDC);
		}
		DeleteObj(F);
	}

	return Status;
}

GdcBmp::GdcBmp()
{
	Props.Set("Type", "Windows or OS/2 Bitmap");
	Props.Set("Extension", "BMP");
}
