#include <stdio.h>

#include "Lgi.h"
#include "StoreCommon.h"

/// \brief This class can limit the reading/writing to a specific sub section of a file
/// Which should protect the application using the storage system from overwriting parts
/// of the file is shouldn't have access to. The default region is the whole file.

GSubFilePtr::GSubFilePtr(GSubFile *Parent, char *file, int line)
{
	File = Parent;
	SrcFile = NewStr(file);
	SrcLine = line;
	OurStatus = true;
	OurPos = 0;
	SetSwap(Parent->GetSwap());
	
	ClearSub();
}

GSubFilePtr::~GSubFilePtr()
{
	File->Detach(this);
	DeleteArray(SrcFile);
}

bool GSubFilePtr::SaveState()
{
	ActualPos = File->GetPos();
	ActualStatus = File->GetStatus();
	ActualSwap = File->GetSwap();
	
	File->SetSwap(GetSwap());
	File->SetStatus(OurStatus);
	
	int64 Pos = File->SetPos(OurPos);
	if (Pos == OurPos)
	{
		return true;
	}
	else
	{
		printf("%s:%i SaveState failed Pos=%Li OurPos=%Li\n",
			__FILE__, __LINE__, Pos, OurPos);
	}
	return false;
}

bool GSubFilePtr::RestoreState()
{
	SetSwap(File->GetSwap());
	OurStatus = File->GetStatus();
	OurPos = File->GetPos();

	File->SetStatus(ActualStatus);
	File->SetSwap(ActualSwap);
	return File->SetPos(ActualPos) == ActualPos;
}

void GSubFilePtr::SetSub(int start, int len)
{
	Sub = true;
	Start = start;
	Len = len;
}

void GSubFilePtr::ClearSub()
{
	Sub = false;
	Start = Len = 0;
}

int GSubFilePtr::Open(char *Str, int Int)
{
	LgiAssert(0);
	return 0;
}

bool GSubFilePtr::IsOpen()
{
	return true;
}

int GSubFilePtr::Close()
{
	LgiAssert(0);
	return 0;
}

int64 GSubFilePtr::GetSize()
{
	int64 s = -1;
	
	if (Sub)
	{
		s = Len;
	}
	else if (File->Lock())
	{
		s = File->GetSize();
		File->Unlock();
	}
	
	return s;
}

int64 GSubFilePtr::SetSize(int64 Size)
{
	LgiAssert(0);
	return -1;
}

int64 GSubFilePtr::GetPos()
{
	return OurPos;
}

int64 GSubFilePtr::SetPos(int64 pos)
{
	if (pos < 0)
	{
		printf("%s:%i - Pos < 0???\n", __FILE__, __LINE__);
	}
	
	return OurPos = pos;
}

int64 GSubFilePtr::Seek(int64 To, int Whence)
{
	switch (Whence)
	{
		case SEEK_SET:
			SetPos(To);
			break;
		case SEEK_CUR:
			SetPos(GetPos() + To);
			break;
		case SEEK_END:
			SetPos(GetSize() + To);
			break;
	}

	return GetPos();
}

GStreamI *GSubFilePtr::Clone()
{
	LgiAssert(0);
	return 0;
}

bool GSubFilePtr::Eof()
{
	bool Status = false;
	
	if (Sub)
	{
		Status = OurPos >= Start AND OurPos < Start + Len;
	}
	else if (File->Lock())
	{
		Status = File->Eof();
		File->Unlock();
	}
	
	return Status;		
}

bool GSubFilePtr::GetStatus()
{
	return OurStatus;
}

void GSubFilePtr::SetStatus(bool s)
{
	OurStatus = s;
}

int GSubFilePtr::Read(void *Buffer, int Size, int Flags)
{
	int Status = 0;
	
	if (File->Lock())
	{
		int End = OurPos + Size;
		if (NOT Sub OR
			(OurPos >= Start AND End <= Start + Len))
		{
			if (SaveState())
			{
				Status = File->Read(Buffer, Size, Flags);
				RestoreState();
			}
		}
		else
		{
			LgiAssert(0);
		}
		
		File->Unlock();
	}

	return Status;
}

int GSubFilePtr::Write(void *Buffer, int Size, int Flags)
{
	int Status = 0;
	
	if (File->Lock())
	{
		int End = OurPos + Size;
		if (NOT Sub OR
			(OurPos >= Start AND End <= Start + Len))
		{
			if (SaveState())
			{
				Status = File->Write(Buffer, Size, Flags);
				RestoreState();
			}
		}
		else
		{
			LgiAssert(0);
		}
		
		File->Unlock();
	}

	return Status;
}

int GSubFilePtr::ReadStr(char *Buf, int Size)
{
	LgiAssert(0);
	return 0;
}

int GSubFilePtr::WriteStr(char *Buf, int Size)
{
	LgiAssert(0);
	return 0;
}

int GSubFilePtr::Print(char *Format, ...)
{
	LgiAssert(0);
	return 0;
}

///////////////////////////////////////////////////////////////////////
GSubFile::GSubFile(GSemaphore *lock, bool Buffering)
{
	Lck = lock;

	#if GSUBFILE_NOBUFFERING
	Buffer = Buffering;
	Block = 0;
	Buf = 0;
	Shift = 0;
	Cur = -1;
	Pos = -1;
	Dirty = false;
	#endif
}

GSubFile::~GSubFile()
{
	while (Ptrs.Length())
	{
		GSubFilePtr *p = Ptrs[0];
		if (p)
		{
			printf("%s:%i - GSubFilePtr not released.\n", p->SrcFile, p->SrcLine);
			DeleteObj(p);
		}
		else break;
	}

	#if GSUBFILE_NOBUFFERING
	DeleteArray(Buf);
	#endif
}

GSubFilePtr *GSubFile::Create(char *file, int line)
{
	GSubFilePtr *p = 0;
	
	p = NEW(GSubFilePtr(this, file, line));
	if (p)
	{
		if (Lock())
		{
			Ptrs[Ptrs.Length()] = p;
			Unlock();
		}
	}
	
	return p;
}

void GSubFile::Detach(GSubFilePtr *Ptr)
{
	if (Lock())
	{
		LgiAssert(Ptrs.HasItem(Ptr));
		Ptrs.Delete(Ptr);
		Unlock();
	}
}

bool GSubFile::Lock()
{
	return Lck->Lock();
}

bool GSubFile::LockWithTimeout(int Timeout)
{
	return Lck->LockWithTimeout(Timeout);
}

void GSubFile::Unlock()
{
	Lck->Unlock();
}

#if GSUBFILE_NOBUFFERING
int GSubFile::Open(char *Str, int Int)
{
	int s = GFile::Open(Str, Int | (NOT Buffer ? O_NO_CACHE : 0));
	if (s AND NOT Buffer)
	{
		Block = GetBlockSize();
		Shift = 1;
		while ((1 << Shift) < Block)
			Shift++;
		Buf = NEW(uint8[Block]);
		Pos = 0;
	}
	return s;
}

int GSubFile::Read(void *OutBuf, int Size, int Flags)
{
	int Status = 0;

	if (NOT Buffer AND Buf)
	{
		uint8 *Out = (uint8*)OutBuf;
		int64 Blk = Pos >> Shift;
		int Mask = Block - 1;

		while (Size)
		{
			if (Blk != Cur)
			{
				SetBlock(Blk);
			}

			int Off = Pos & Mask;
			int Len = Block - Off;
			if (Len > Size)
				Len = Size;

			memcpy(Out, Buf + Off, Len);

			Pos += Len;
			Out += Len;
			Size -= Len;
			Status += Len;
			Blk++;
		}
	}
	else
	{
		Status = GFile::Read(OutBuf, Size, Flags);
		if (Status > 0)
		{
			Pos += Status;
		}
	}

	return Status;
}

int GSubFile::Write(void *InBuf, int Size, int Flags)
{
	int Status = 0;

	if (NOT Buffer AND Buf)
	{
		uint8 *In = (uint8*)InBuf;
		int64 Blk = Pos >> Shift;
		int Mask = Block - 1;

		while (Size)
		{
			if (Blk != Cur)
			{
				SetBlock(Blk);
			}

			int Off = Pos & Mask;
			int Len = Block - Off;
			if (Len > Size)
				Len = Size;

			memcpy(Buf + Off, In, Len);
			Dirty = true;

			Pos += Len;
			In += Len;
			Size -= Len;
			Status += Len;
			Blk++;
		}
	}
	else
	{
		Status = GFile::Write(InBuf, Size, Flags);
		if (Status > 0)
		{
			Pos += Status;
		}
	}

	return Status;
}

int64 GSubFile::GetPos()
{
	return Pos;
}

int64 GSubFile::SetPos(int64 pos)
{
	return Pos = pos;
}

bool GSubFile::SetBlock(int64 Blk)
{
	if (Blk != Cur)
	{
		int64 p, n;

		if (Dirty)
		{
			p = GFile::SetPos(Cur << Shift);
			n = GFile::Write(Buf, Block);
			if (n != Block)
				return false;
			
			Dirty = false;
		}

		Cur = Blk;

		p = GFile::SetPos(Cur << Shift);
		n = GFile::Read(Buf, Block);

		return n > 0;
	}

	return true;
}

#endif
