#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>

#include "Lgi.h"
#include "GXmlTree.h"
#include "GVariant.h"
#include "GToken.h"

//////////////////////////////////////////////////////////////////////////////
class GXmlTreePrivate
{
public:
	GXmlFactory *Factory;
	GStream *File;
	char *Error;
	int Flags;
	GHashTable Entities;
	GHashTable NoChildTags;

	GXmlTreePrivate()
	{
		Factory = 0;
		File = 0;
		Error = 0;
		Flags = 0;

		Entities.Add("lt", NewStr("<"));
		Entities.Add("gt", NewStr(">"));
		Entities.Add("amp", NewStr("&"));
	}
	
	~GXmlTreePrivate()
	{
		DeleteArray(Error);
		
		for (char *e = (char*)Entities.First(); e; e = (char*)Entities.Next())
		{
			DeleteArray(e);
		}
	}
};

//////////////////////////////////////////////////////////////////////////////
char *GXmlTree::EncodeEntities(char *s, int len)
{
	GStringPipe p;
	if (EncodeEntities(&p, s, len))
	{
		return p.NewStr();
	}

	return 0;
}

bool GXmlTree::EncodeEntities(GStream *to, char *start, int len)
{
	if (NOT start OR NOT to)
		return 0;

	int Amp = (d->Flags & GXT_NO_ENTITIES) ? 10000000 : '&';

	for (char *s = start; s AND *s;)
	{
		char *e = s;
		while (	*e AND
				*e != '\"' AND
				*e != Amp AND
				*e != '<' AND
				*e != '>' AND
				(len < 0 OR e < start + len))
			e++;
		
		if (*e)
		{
			to->Write(s, (int)e-(int)s);
			
			switch (*e)
			{
				case '<':
					to->Write((char*)"&lt;", 4);
					break;
				case '>':
					to->Write((char*)"&gt;", 4);
					break;
				case '&':
					to->Write((char*)"&amp;", 5);
					break;
				default:
				{
					char Ent[32];
					sprintf(Ent, "&#%i;", (uchar)*e);
					to->Write(Ent, strlen(Ent));
				}
			}
			
			s = e + 1;
		}
		else
		{
			to->Write(s, strlen(s));
			break;
		}
	}

	return true;
}

char *GXmlTree::DecodeEntities(char *s, int len)
{
	GStringPipe p;
	
	char *End = s + len;
	while (s AND *s AND s < End)
	{
		char *e = s;
		while (*e AND *e != '&' AND e < End) *e++;
		if (*e)
		{
			p.Push(s, (int)e-(int)s);
			if (e >= End OR
				*e != '&')
			{
				break;
			}
			else
			{
				e++;
				if (*e == '#')
				{
					e++;
					char16 c16 = atoi(e);
					char *c8 = LgiNewUtf16To8(&c16, sizeof(char16));
					if (c8)
					{
						p.Push(c8);
						DeleteArray(c8);
					}
					s = strchr(e, ';');
					if (s) s++;
				}
				else
				{
					char *Def = 0;
					char *Col = strchr(e, ';');
					if (Col AND (Col - e) < 16)
					{
						*Col = 0;
						Def = (char*)d->Entities.Find(e);
						*Col = ';';
						if (Def)
						{
							p.Push(Def);
							s = Col + 1;
						}
						else
						{
							p.Write((char*)"&", 1);
							s = e;
						}
					}
					else
					{
						p.Write((char*)"&", 1);
						s = e;
					}
				}
			}
		}
		else
		{
			p.Push(s);
			break;
		}
	}

	return p.NewStr();
}

//////////////////////////////////////////////////////////////////////////////
GXmlTag::GXmlTag(char *tag)
{
	Tag = NewStr(tag);
	Write = 0;
	Parent = 0;
	Content = 0;
}

GXmlTag::~GXmlTag()
{
	RemoveTag();
	Empty();
}

void GXmlTag::Empty()
{
	Attr.DeleteObjects();
	DeleteArray(Content);
	DeleteArray(Tag);
	
	GXmlTag *c;
	while (c = Children.First())
	{
		LgiAssert(c->Parent == this);
		DeleteObj(c);
	}
}

GXmlTag &GXmlTag::operator =(GXmlTag &t)
{
	Attr.DeleteObjects();
	DeleteArray(Content);
	DeleteArray(Tag);
	
	Content = NewStr(t.Content);
	Tag = NewStr(t.Tag);
	for (GXmlAttr *a=t.Attr.First(); a; a=t.Attr.Next())
	{
		Attr.Insert(NEW(GXmlAttr(a->Name, a->Value)));
	}
	return *this;
}

GXmlTag *GXmlTag::GetTag(char *Name, bool Create)
{
	GToken p(Name, ".");

	GXmlTag *t = this;
	for (int i=0; i<p.Length() AND t; i++)
	{
		GXmlTag *Child = 0;
		
		Iterator<GXmlTag> n(&t->Children);
		for (GXmlTag *c = n.First(); c; c = n.Next())
		{
			if (c->Tag AND stricmp(c->Tag, p[i]) == 0)
			{
				Child = c;
				break;
			}
		}
		if (NOT Child)
		{
			if (Create)
			{
				t->InsertTag( Child = NEW(GXmlTag(p[i])) );
				if (NOT Child)
				{
					return 0;
				}
			}
			else return 0;
		}
		t = Child;
	}
	
	return t;
}

bool GXmlTag::GetVariant(char *Name, GVariant &Value, char *Array)
{
	if (Name)
	{
		Iterator<GXmlTag> n(&Children);
		for (GXmlTag *c = n.First(); c; c = n.Next())
		{
			if (c->Tag AND stricmp(c->Tag, Name) == 0)
			{
				Value = (GDom*) c;
				return true;
			}
		}

		char *v = GetAttr(Name);
		if (v)
		{
			if (v[0] == 'b' AND
				v[1] == 'i' AND
				v[2] == 'n' AND
				v[3] == 'a' AND
				v[4] == 'r' AND
				v[5] == 'y' AND
				v[6] == '(')
			{
				Value.Empty();
				Value.Type = GV_BINARY;
				int Len = 0;
				char *Start = v + 7;
				for (char *s = Start; *s AND *s != ')'; s++)
					Len++;
				Len >>= 1;

				Value.Value.Binary.Length = Len;
				Value.Value.Binary.Data = NEW(uchar[Len]);
				if (Value.Value.Binary.Data)
				{
					char *s = Start;
					for (int i=0; i<Len; i++)
					{
						#define HexToBin()	\
							(*s >= '0' && *s <= '9') ? *s - '0' : \
							((*s >= 'a' && *s <= 'f') ? *s - 'a' + 10  : 0)

						int High = HexToBin();
						s++;
						int Low = HexToBin();
						s++;
						((uint8*)Value.Value.Binary.Data)[i] = High << 4 | Low;						
					}
				}
			}
			else
			{
				Value = v;
			}
			return true;
		}

		if (stricmp(Name, "text") == 0)
		{
			Value = Content;
			return true;
		}
	}

	return false;	
}

bool GXmlTag::SetVariant(char *Name, GVariant &Value, char *Array)
{
	if (Name)
	{
		switch (Value.Type)
		{
			case GV_NULL:
			{
				DelAttr(Name);
				break;
			}
			case GV_BOOL:
			{
				char i[32];
				sprintf(i, "%i", Value.Value.Bool);
				SetAttr(Name, i);
				break;
			}
			case GV_INT32:
			{
				char i[32];
				sprintf(i, "%i", Value.Value.Int);
				SetAttr(Name, i);
				break;
			}
			case GV_INT64:
			{
				char i[32];
				sprintf(i, LGI_PrintfInt64, Value.Value.Int64);
				SetAttr(Name, i);
				break;
			}
			case GV_DOUBLE:
			{
				char i[32];
				sprintf(i, "%f", Value.Value.Dbl);
				SetAttr(Name, i);
				break;
			}
			case GV_STRING:
			{
				SetAttr(Name, Value.Str());
				break;
			}
			case GV_BINARY:
			{
				GStringPipe p;
				p.Print("binary(");
				for (int i=0; i<Value.Value.Binary.Length; i++)
				{
					p.Print("%02.2x", ((uint8*)Value.Value.Binary.Data)[i]);
				}
				p.Print(")");
				char *s = p.NewStr();
				SetAttr(Name, s);
				DeleteArray(s);
				break;
			}
			default:
			{
				LgiAssert(0);
				return false;
			}
		}
		
		return true;
	}
	
	return false;
}

void GXmlTag::InsertTag(GXmlTag *t)
{
	if (t)
	{
		t->RemoveTag();
		t->Parent = this;		
		Children.Insert(t);
	}
}

void GXmlTag::RemoveTag()
{
	if (Parent)
	{
		Parent->Children.Delete(this);
	}
	Parent = 0;	
}

bool GXmlTag::Dump(int Depth)
{
	#define Tabs() { for (int i=0; i<Depth; i++) printf("    "); }
	
	Tabs();
	LgiTrace("Tag='%s'\n", Tag);
	Depth++;

	for (GXmlAttr *a=Attr.First(); a; a=Attr.Next())
	{
		Tabs();
		LgiTrace("Attr: %s='%s'\n", a->Name, a->Value);
	}	
	LgiTrace("\n");
	
	for (GXmlTag *t=Children.First(); t; t=Children.Next())
	{
		t->Dump(Depth + 1);
	}
	
	LgiTrace("\n");
	Depth--;
	
	#undef Tabs
	return true;
}

GXmlAttr *GXmlTag::_Attr(char *Name, bool Wr)
{
	if (Name)
	{
		for (GXmlAttr *a=Attr.First(); a; a=Attr.Next())
		{
			if (a->Name AND stricmp(a->Name, Name) == 0)
			{
				if (Wr)
				{
					DeleteArray(a->Value);
				}
				return a;
			}
		}

		if (Wr)
		{
			// Create
			GXmlAttr *n = NEW(GXmlAttr);
			if (n)
			{
				n->Name = NewStr(Name);
				Attr.Insert(n);
				return n;
			}
		}
	}

	return 0;
}

bool GXmlTag::DelAttr(char *Name)
{
	GXmlAttr *a = _Attr(Name, false);
	if (a)
	{
		Attr.Delete(a);
		DeleteObj(a);
		return true;
	}
	return false;
}

char *GXmlTag::GetAttr(char *n)
{
	GXmlAttr *a = _Attr(n, false);
	return a ? a->Value : 0;	
}

int GXmlTag::GetAsInt(char *n)
{
	GXmlAttr *a = _Attr(n, false);
	return a ? atoi(a->Value) : -1;	
}

bool GXmlTag::SetAttr(char *n, char *Value)
{
	GXmlAttr *a = _Attr(n, true);
	if (a)
	{
		a->OwnMem = true;
		a->Value = NewStr(Value);
		return true;
	}
	return false;
}

bool GXmlTag::SetAttr(char *n, int Value)
{
	GXmlAttr *a = _Attr(n, true);
	if (a)
	{
		char s[32];
		sprintf(s, "%i", Value);
		a->OwnMem = true;
		a->Value = NewStr(s);
		return true;
	}
	return false;
}

bool GXmlTag::SerializeAttr(char *Attr, int &Int)
{
	GXmlAttr *a = _Attr(Attr, Write);
	if (a)
	{
		if (Write)
		{
			char s[256];
			sprintf(s, "%i", Int);
			a->Value = NewStr(s);
			return a->Value != 0;
		}
		else if (a->Value)
		{
			Int = atoi(a->Value);
			return true;
		}
	}
	return false;
}

bool GXmlTag::SerializeAttr(char *Name, char *&Str)
{
	GXmlAttr *a = _Attr(Name, Write);
	if (a)
	{
		if (Write)
		{
			if (ValidStr(Str))
			{
				a->Value = NewStr(Str);
			}
			else
			{
				Attr.Delete(a);
				DeleteObj(a);
				return true;
			}
			
			return a->Value != 0;
		}
		else if (a->Value)
		{
			Str = NewStr(a->Value);
			return true;
		}
	}
	return false;
}

bool GXmlTag::SerializeAttr(char *Attr, double &Dbl)
{
	GXmlAttr *a = _Attr(Attr, Write);
	if (a)
	{
		if (Write)
		{
			char s[256];
			sprintf(s, "%f", Dbl);
			a->Value = NewStr(s);
			return a->Value != 0;
		}
		else if (a->Value)
		{
			Dbl = atof(a->Value);
			return true;
		}
	}
	return false;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////
GXmlTree::GXmlTree(int Flags)
{
	d = NEW(GXmlTreePrivate);
	d->Flags = Flags;
}

GXmlTree::~GXmlTree()
{
	DeleteObj(d);
}

GXmlTag *GXmlTree::Parse(GXmlTag *Tag, char *&t, bool &NoChildren, bool InTypeDef)
{
	static const char *White = " \t\r\n";
	#define SkipWhiteSpace(s) while (*s AND strchr(White, *s)) s++

	bool KeepWs = TestFlag(d->Flags, GXT_KEEP_WHITESPACE);
	char *Start = t;
	GStringPipe Before;
		
	// Skip white
	while (*t)
	{
		if (*t == '<')
		{
			if (t[1] == '!')
			{
				if (t[2] == '-' AND
					t[3] == '-')
				{
					// Start of comment
					if (KeepWs)
					{
						Before.Write(Start, t - Start);
					}

					char *End = strstr(t + 3, "-->");
					if (End)
					{
						t = End + 2;

						if (KeepWs)
						{
							Start = t;
						}
					}
					else
					{
						t = 0;
						break;
					}
				}
				else
				{
					char *End = strchr(t + 1, '>');
					if (End)
					{
						t = End;
					}
				}
			}
			else
			{
				break;
			}
		}

		t++;
	}

	if (KeepWs AND t > Start)
	{		
		Before.Write(Start, t - Start);

		if (Before.GetSize() > 0)
		{
			GXmlTag *PreContent = d->Factory ? d->Factory->Create(0) : NEW(GXmlTag);
			if (PreContent)
			{
				NoChildren = true;
				PreContent->Content = Before.NewStr();
				return PreContent;
			}
		}
	}

	if (*t AND *t == '<')
	{
		t++;

		SkipWhiteSpace(t);
		
		// Store tagname start
		char *TagName = t;
		bool TypeDef = *t == '!';

		if (*t == '/')
			t++;

		while
		(
			*t AND
			(
				isalpha(*t) OR
				*t == '!' OR
				*t == '?' OR
				*t == '-' OR
				*t == '_' OR
				*t == '.' OR
				*t == '[' OR
				*t == ']' OR
				(t > TagName AND isdigit(*t))
			)
		)
		{
			t++;
		}

		if (t > TagName)
		{
			// Create the tag object
			char Old = *t;
			*t = 0;
			if (NOT Tag)
			{
				Tag = d->Factory ? d->Factory->Create(TagName) : 0;
				if (NOT Tag)
					Tag = NEW(GXmlTag);
			}
			*t = Old;
			
			// Parse attributes into tag
			if (Tag)
			{	
				DeleteArray(Tag->Tag);
				Tag->Tag = NewStr(TagName, (int)t-(int)TagName);
				Tag->Attr.DeleteObjects();
				NoChildren = false;
				
				while (*t AND *t != '>')
				{
					// Skip white
					SkipWhiteSpace(t);
					
					if (*t == '>')
						break;

					if (*t == '/')
					{
						t++;
						NoChildren = true;
						break;
					}
					
					char *AttrName = t;

					if (TypeDef)
					{
						if (*t == '\"')
						{
							char *Start = ++t;
							t = strchr(t, '\"');
							if (t)
							{
								if (t > Start)
								{
									GXmlAttr *Attr = NEW(GXmlAttr);
									if (Attr)
									{
										Attr->Name = NewStr(Start, t - Start);
										Tag->Attr.Insert(Attr);
									}
								}

								t++;
								continue;
							}
							else
							{
								break;
							}
						}
						else if (*t == '[')
						{
							// Read all defs into children of this tag
							t++;
							while (*t AND *t != ']')
							{
								bool NoChildren = false;
								GXmlTag *c = Parse(0, t, NoChildren, true);
								if (c)
								{
									Tag->InsertTag(c);
								}
								else break;							
							}

							if (*t == ']')
								t++;
							continue;
						}
					}

					// Goto the end of the attr name
					while
					(
						*t AND
						(
							isalpha(*t) OR
							isdigit(*t) OR
							*t == '-' OR
							*t == '.' OR
							*t == '_' OR
							*t == ':'
						)
					)
					{
						t++;
					}

					if (t > AttrName)
					{
						GXmlAttr *Attr = NEW(GXmlAttr);
						if (Attr)
						{
							Attr->OwnMem = true;
							Attr->Name = NewStr(AttrName, (int)t-(int)AttrName);
							
							// Skip white
							SkipWhiteSpace(t);
							
							// Look for equals
							if (*t == '=')
							{
								t++;
								
								// Skip white
								SkipWhiteSpace(t);
								
								if (strchr("\"\'", *t))
								{
									// Delimited string
									char Delim = *t++;
									char *End = strchr(t, Delim);
									if (End)
									{
										if (d->Flags & GXT_NO_ENTITIES)
										{
											Attr->Value = NewStr(t, (int)End-(int)t);
										}
										else
										{
											Attr->Value = DecodeEntities(t, (int)End-(int)t);
										}
										t = End + 1;
									}
									else
									{
										break;
									}
								}
								else
								{
									// Non delimited string
									char *End = t;
									while (*End AND NOT strchr(White, *End) AND *End != '>'  AND *End != '/') *End++;
									Attr->Value = DecodeEntities(t, (int)End-(int)t);
									t = End;
								}
							}

							Tag->Attr.Insert(Attr);
						}
						else
						{
							LgiAssert(0);
							break;
						}
					}
					else
					{
						LgiTrace("%s:%i - Error tagname='%s'\n", __FILE__, __LINE__, Tag->Tag);
						LgiAssert(0);
						break;
					}

					// Skip white
					SkipWhiteSpace(t);
					if (*t == '?') t++;
				}
				
				// Skip white
				SkipWhiteSpace(t);
				
				if (*t == '>')
					t++;
			}
			
			char *ContentStart = t;
			while
			(
				*t AND
				*t != '<' AND
				(NOT InTypeDef OR *t != ']')
			)
			{
				t++;
			}
			if (t > ContentStart)
			{
				if (d->Flags & GXT_NO_ENTITIES)
				{
					Tag->Content = NewStr(ContentStart, (int)t - (int)ContentStart);
				}
				else
				{
					Tag->Content = DecodeEntities(ContentStart, (int)t - (int)ContentStart);
				}

				if (NOT TestFlag(d->Flags, GXT_KEEP_WHITESPACE) AND NOT ValidStr(Tag->Content))
				{
					DeleteArray(Tag->Content);
				}
			}
		}
		else if (*t)
		{
			printf("Xml: stopped at '%-20s'\n", t);
			LgiAssert(0);
		}
	}
	
	return Tag;
}

bool GXmlTree::Read(GXmlTag *Root, GStream *File, GXmlFactory *Factory)
{
	bool Status = false;
	
	if (Root AND File)
	{
		int Len = File->GetSize();
		if (Len > 0)
		{
			char *Str = NEW(char[Len+1]);
			if (Str)
			{
				int r = File->Read(Str, Len);
				if (r >= 0)
				{
					Str[r] = 0;
					
					char *Ptr = Str;
					d->Factory = Factory;
					
					GXmlTag *Current = Root;
					Status = true;
					bool First = true;
					while (Current AND Ptr AND *Ptr)
					{
						bool NoChildren;
						GXmlTag *t = Parse(First AND NOT TestFlag(d->Flags, GXT_NO_DOM) ? Root : 0, Ptr, NoChildren, false);
						First = false;
						if (t)
						{
							if (t->Tag AND
								t->Tag[0] == '!' AND
								strcmp(t->Tag, "!DOCTYPE") == 0)
							{
								for (GXmlTag *c=t->Children.First(); c; c=t->Children.Next())
								{
									if (c->Tag AND
										strcmp(c->Tag, "!ENTITY") == 0)
									{
										if (c->Attr.Length() == 2)
										{
											GXmlAttr *Ent = c->Attr[0];
											GXmlAttr *Value = c->Attr[1];
											if (Ent->Name AND
												Value->Name AND
												NOT d->Entities.Find(Ent->Name))
											{
												d->Entities.Add(Ent->Name, NewStr(Value->Name));
											}
										}
									}
								}
							}

							if (t->Tag AND t->Tag[0] == '/' AND Current->Tag)
							{
								// End tag
								if (stricmp(t->Tag + 1, Current->Tag) == 0)
								{
									Current = Current->Parent;
								}
								else
								{
									int Lines = 1;
									for (char *k = Ptr; k >= Str; k--)
									{
										if (*k == '\n') Lines++;
									}

									char s[256];
									sprintf(s, "Mismatched '%s' tag, got '%s' instead (Line %i).\n", t->Tag, Current->Tag, Lines);
									printf("%s:%i - XmlTree error %s\n", __FILE__, __LINE__, s);
									d->Error = NewStr(s);
									Status = false;

									#ifdef _DEBUG
									GXmlTree Dbg;
									GFile Out;
									if (Out.Open("c:\\temp\\out.xml", O_WRITE))
									{
										Dbg.Write(Root, &Out);
									}
									#endif
									break;
								}

								DeleteObj(t);
							}
							else
							{
								if (t != Root)
								{
									Current->InsertTag(t);
								}
								
								t->Write = false;
								t->Serialize();
								
								if (NOT TestFlag(d->Flags, GXT_NO_DOM) AND
									NOT NoChildren AND
									NOT d->NoChildTags.Find(t->Tag))
								{
									Current = t;
								}
							}
						}
						else break;
					}
					
					d->Factory = 0;
				}
				else
				{
					printf("%s:%i - Error.\n", __FILE__, __LINE__);
				}
				
				DeleteArray(Str);
			}
			else
			{
				printf("%s:%i - Error.\n", __FILE__, __LINE__);
			}
			
			// Root->Dump();
		}
	}
	else
	{
		printf("%s:%i - Error.\n", __FILE__, __LINE__);
	}
	
	return Status;
}

void GXmlTree::Output(GXmlTag *t, int Depth)
{
	#define Tabs if (NOT TestFlag(d->Flags, GXT_NO_PRETTY_WHITESPACE)) \
		{ for (int i=0; i<Depth; i++) d->File->Write((void*)"\t", 1); }

	t->Write = true;
	t->Serialize();
	Tabs
	d->File->Print("<%s", t->Tag);
	for (GXmlAttr *a=t->Attr.First(); a; )
	{
		// Write the attribute name
		d->File->Print(" %s=\"", a->Name);
		
		// Encode the value
		EncodeEntities(d->File, a->Value, -1);

		// Write the delimiter
		d->File->Write((void*)"\"", 1);
		a = t->Attr.Next();
		if (a AND TestFlag(d->Flags, GXT_PRETTY_WHITESPACE))
		{
			d->File->Write((void*)"\n", 1);
			for (int i=0; i<=Depth; i++) d->File->Write((void*)"\t", 1);
		}
	}
	
	// Write the child tags
	bool HasContent = ValidStr(t->Content);
	GXmlTag *c = t->Children.First();
	if (c OR HasContent)
	{
		d->File->Print(">");
		
		if (HasContent)
		{
			EncodeEntities(d->File, t->Content, -1);
			if (c)
			{
				d->File->Print("\n");
				Tabs
			}
		}

		if (c)
		{
			if (NOT HasContent)
			{
				d->File->Print("\n");
			}

			for (; c; c=t->Children.Next())
			{
				Output(c, Depth + 1);
			}

			Tabs
		}
	
		d->File->Print("</%s>\n", t->Tag);
	}
	else
	{
		d->File->Print(" />\n");
	}
	
	#undef Tabs
}

bool GXmlTree::Write(GXmlTag *Root, GStream *File)
{
	bool Status = false;
	
	if (Root AND File)
	{
		File->SetSize(0);
		d->File = File;
		Output(Root, 0);
		d->File = 0;
		Status = true;
	}
	
	return Status;
}

char *GXmlTree::GetErrorMsg()
{
	return d->Error;
}

GHashTable *GXmlTree::NoChildTags()
{
	return &d->NoChildTags;
}

GHashTable *GXmlTree::GetEntityTable()
{
	return &d->Entities;
}

