/// \file
/// \author Matthew Allen

#include "Lgi.h"
#include "GXmlTreeUi.h"
#include "GCheckBox.h"
#include "GButton.h"
#include "GRadioGroup.h"
#include "GSlider.h"
#include "GCombo.h"
#include "GEdit.h"

#include <ctype.h>

struct Mapping
{
	int Id;
	int Hint;
};

class GXmlToUiPriv
{
public:
	GHashTable Maps;
};

GXmlToUi::GXmlToUi()
{
	d = NEW(GXmlToUiPriv);
}

GXmlToUi::~GXmlToUi()
{
	for (Mapping *m=(Mapping*)d->Maps.First(); m; m=(Mapping*)d->Maps.Next())
	{
		DeleteObj(m);
	}
	DeleteObj(d);
}

void GXmlToUi::EnableAll(GViewI *Ui, bool Enable)
{
	if (Ui)
	{
		for (Mapping *m=(Mapping*)d->Maps.First(); m; m=(Mapping*)d->Maps.Next())
		{
			Ui->SetCtrlEnabled(m->Id, Enable);
		}
	}
}

void GXmlToUi::Map(char *Attr, int UiIdent, int Type)
{
	if (Attr AND UiIdent > 0)
	{
		Mapping *m = NEW(Mapping);
		if (m)
		{
			m->Id = UiIdent;
			m->Hint = Type;
			d->Maps.Add(Attr, (void*)m);
		}
	}
	else
	{
		LgiAssert(0);
	}
}

void GXmlToUi::EmptyMaps()
{
	for (Mapping *m = (Mapping*)d->Maps.First(); m; m = (Mapping*)d->Maps.Next())
	{
		DeleteObj(m);
	}
	d->Maps.Empty();
}

int GetCtrlType(GViewI *v)
{
	if (v)
	{
		if (dynamic_cast<GCheckBox*>(v) OR
			dynamic_cast<GButton*>(v) OR
			dynamic_cast<GRadioButton*>(v))
		{
			return GV_BOOL;
		}
		else if (dynamic_cast<GSlider*>(v) OR
				dynamic_cast<GCombo*>(v) OR
				dynamic_cast<GRadioGroup*>(v))
		{
			return GV_INT32;
		}
	}

	return GV_STRING;
}

int GetDataType(char *str)
{
	if (str)
	{
		bool Float = false;

		char16 w;
		int Len = strlen(str);
		while (w = LgiUtf8To32((uint8*&)str, Len))
		{
			if (strchr("e \t\r\n", w))
			{
				// Doesn't really tell us anything, so ignore it.
				// The 'e' part is sometimes part of a number or
				// ignore that too.
			}
			else if (NOT isdigit(w) OR w > 255)
			{
				return GV_STRING;
			}
			else if (w == '.')
			{
				Float = true;
			}
		}

		if (Float)
		{
			return GV_DOUBLE;
		}

		return GV_INT32;
	}

	return GV_NULL;
}

bool GXmlToUi::Convert(GDom *Tag, GViewI *Ui, bool ToUI)
{
	bool Status = false;

	if (Ui AND Tag)
	{
		GVariant v;
		if (ToUI)
		{
// printf("ToUi=%i\n", ToUI);
			// Xml -> UI
			char *Attr;
			for (Mapping *Map = (Mapping*)d->Maps.First(&Attr); Map; Map = (Mapping*)d->Maps.Next(&Attr))
			{
				if (Tag->GetValue(Attr, v))
				{
					int Type = Map->Hint ? Map->Hint : GetDataType(v.Str());

// printf("\tType=%i Attr=%s Id=%i : v.Type=%i v.Str()=%s v.Int=%i\n", Type, Attr, Map->Id, v.Type, v.Str(), v.CastInt32());
					if (Type == GV_BOOL OR
						Type == GV_INT32 OR
						Type == GV_INT64)
					{
						Ui->SetCtrlValue(Map->Id, v.CastInt32());
					}
					else
					{
						Ui->SetCtrlName(Map->Id, v.Str());
					}
				}
				else
				{
					GViewI *c = Ui->FindControl(Map->Id);
					if (c AND dynamic_cast<GEdit*>(c))
						c->Name("");
				}
			}
		}
		else
		{
			// UI -> Xml
			char *Attr;
			for (Mapping *Map = (Mapping*)d->Maps.First(&Attr); Map; Map = (Mapping*)d->Maps.Next(&Attr))
			{
				GViewI *c = Ui->FindControl(Map->Id);
				if (c)
				{
					int Type = Map->Hint ? Map->Hint : GetCtrlType(c);

					switch (Type)
					{
						case GV_INT32:
						case GV_BOOL:
						{
							Tag->SetValue(Attr, v = c->Value());
							break;
						}
						default:
						{
							char *Str = c->Name();
							
							if (ValidStr(Str))
								v = Str;
							else
								v.Empty();

							Tag->SetValue(Attr, v);
							break;
						}
					}

// printf("\tType=%i Attr=%s Id=%i : v.Type=%i v.Str()=%s v.Int=%i\n", Type, Attr, Map->Id, v.Type, v.Str(), v.CastInt32());
				}
				else
				{
					v.Empty();
					Tag->SetValue(Attr, v);
				}
			}
		}
	}

	return Status;
}

