/*hdr
**      FILE:           GList.cpp
**      AUTHOR:         Matthew Allen
**      DATE:           14/2/2000
**      DESCRIPTION:    Lgi self-drawn listbox
**
**      Copyright (C) 2000 Matthew Allen
**              fret@memecode.com
*/

#include <math.h>
#include <stdio.h>
#include <ctype.h>

#include "Lgi.h"
#include "GSkinEngine.h"
#include "GList.h"
#include "GEdit.h"
#include "GScrollBar.h"

#define DRAG_NONE					0
#define SELECT_ITEMS				1
#define RESIZE_COLUMN				2
#define DRAG_COLUMN					3
#define CLICK_COLUMN				4
#define TOGGLE_ITEMS				5
#define CLICK_ITEM					6

// Number of pixels you have to move the mouse until a drag is initiated.
#define DRAG_THRESHOLD				4

// Switches for various profiling code..
#define GLIST_POUR_PROFILE			0
#define GLIST_ONPAINT_PROFILE		0

// Options
#define DOUBLE_BUFFER_PAINT			0

// Colours
#define DragColumnColour			LC_LOW

#if defined(WIN32)
#if !defined(WS_EX_LAYERED)
#define WS_EX_LAYERED				0x80000
#endif
#if !defined(LWA_ALPHA)
#define LWA_ALPHA					2
#endif
typedef BOOL (__stdcall *_SetLayeredWindowAttributes)(HWND hwnd, COLORREF crKey, BYTE bAlpha, DWORD dwFlags);
#endif

#define ForAllItems(Var)			Iterator<GListItem> ItemIter(&Items); for (GListItem *Var = ItemIter.First(); Var; Var = ItemIter.Next())
#define ForAllItemsReverse(Var)		Iterator<GListItem> ItemIter(&Items); for (GListItem *Var = ItemIter.Last(); Var; Var = ItemIter.Prev())
#define VisibleItems()				CompletelyVisible // (LastVisible - FirstVisible + 1)
#define MaxScroll()					max(Items.Length() - CompletelyVisible, 0)

////////////////////////////////////////////////////////////////////////////////////////////
#define DRAG_COL_ALPHA				0xc0
#define LINUX_TRANS_COL				0

class GDragColumn : public GWindow
{
	GList *List;
	GListColumn *Col;
	int Index;
	int Offset;
	GdcPt2 ListScrPos;
	
	#ifdef LINUX
	GSurface *Back;
	#endif

public:
	int GetOffset() { return Offset; }
	int GetIndex() { return Index; }
	GListColumn *GetColumn() { return Col; }

	GDragColumn(GList *list, int col);
	~GDragColumn();

	void OnPaint(GSurface *pScreen);
};

class GListPrivate
{
public:
	// Mode
	GListMode Mode;
	int Columns;
	int VisibleColumns;

	// This is a pointer to a flag, that gets set when the object
	// is deleted. Used to trap events deleting the window. If an
	// event handler deletes the current window we can't touch any
	// of the member variables anymore, so we need to know to quit/return
	// ASAP.
	bool *DeleteFlag;

	// If this is true the ctrl is selecting lots of things
	// and we only want to notify once.
	bool NoSelectEvent;

	// Drag'n'drop
	GDragColumn *DragCol;
	GdcPt2 DragStart;
	int DragMode;
	int DragData;
	
	// Kayboard search
	int KeyLast;
	char16 *KeyBuf;
	
	// Column click
	int ColClick;
	GMouse ColMouse;	

	// Editing label
	GView *Edit;
	
	// Class
	GListPrivate()
	{
		Edit = 0;
		DragCol = 0;
		DragMode = 0;
		DragData = 0;
		KeyBuf = 0;
		DeleteFlag = 0;
		Columns = 0;
		VisibleColumns = 0;
		ColClick = -1;
		Mode = GListDetails;
		NoSelectEvent = false;
	}
	
	~GListPrivate()
	{
		DeleteObj(Edit);
		if (DeleteFlag)
		{
			*DeleteFlag = true;
		}
		DeleteObj(DragCol);
		DeleteArray(KeyBuf);
	}
};

class GListColumnPrivate
{
public:
	GRect Pos;
	bool Down;
	bool Drag;

	GList *Parent;
	char *cName;
	GDisplayString *Txt;
	int cWidth;
	int cType;
	GSurface *cIcon;
	int cImage;
	int cMark;
	bool OwnIcon;
	bool CanResize;

	GListColumnPrivate(GList *parent)
	{
		Parent = parent;
		Txt = 0;
		cName = 0;
		cWidth = 0;
		cIcon = 0;
		cType = GIC_ASK_TEXT;
		cImage = -1;
		cMark = GLI_MARK_NONE;
		Down = false;
		OwnIcon = false;
		CanResize = true;
		Drag = false;
	}

	~GListColumnPrivate()
	{
		DeleteArray(cName);
		if (OwnIcon)
		{
			DeleteObj(cIcon);
		}
		DeleteObj(Txt);
	}
};

class GListItemPrivate
{
public:
	bool Selected;
	int ListItem_Image;
	List<GListItemColumn> Cols;
	GArray<char*> Str;
	GArray<GDisplayString*> Display;
	int SelectionStart, SelectionEnd;
	int16 LayoutColumn;

	GListItemPrivate()
	{
		SelectionStart = SelectionEnd = -1;
		Selected = 0;
		ListItem_Image = -1;
		LayoutColumn = -1;
	}

	~GListItemPrivate()
	{
		Cols.DeleteObjects();
		EmptyStrings();
		EmptyDisplay();
	}

	void EmptyStrings()
	{
		for (int i=0; i<Str.Length(); i++)
		{
			DeleteArray(Str[i]);
		}
		Str.Length(0);
	}

	void EmptyDisplay()
	{
		for (int i=0; i<Display.Length(); i++)
		{
			DeleteObj(Display[i]);
		}
		Display.Length(0);
	}
};

/////////////////////////////////////////////////////////////////////////////////////////
GDragColumn::GDragColumn(GList *list, int col)
{
	List = list;
	Index = col;
	Offset = 0;
	#ifdef LINUX
	Back = 0;
	#endif
	Col = List->ColumnAt(Index);
	if (Col)
	{
		Col->d->Down = false;
		Col->d->Drag = true;

		GRect r = Col->d->Pos;
		r.y1 = 0;
		r.y2 = List->Y()-1;
		List->Invalidate(&r, true);

		#ifdef WIN32
		
		int Ver;
		bool Layered = LgiGetOs(&Ver) == LGI_OS_WINNT AND Ver >= 5;
		SetStyle(WS_POPUP);
		SetExStyle(GetExStyle() | WS_EX_TOOLWINDOW);
		if (Layered)
		{
			SetExStyle(GetExStyle() | WS_EX_LAYERED | WS_EX_TRANSPARENT);
		}
		
		#elif defined XWIN
		
		Back = NEW(GMemDC);
		if (Back AND Back->Create(List->X(), List->Y(), GdcD->GetBits()))
		{
			List->OnPaint(Back);
		}

		XSetWindowAttributes a;
		a.override_redirect = true;
		a.save_under = false; // true;
		XChangeWindowAttributes(Handle()->XDisplay(), Handle()->handle(), CWSaveUnder | CWOverrideRedirect, &a);
		
		#endif

		Attach(0);

		#ifdef WIN32
		
		if (Layered)
		{
			SetWindowLong(Handle(), GWL_EXSTYLE, GetWindowLong(Handle(), GWL_EXSTYLE) | WS_EX_LAYERED);

			GLibrary User32("User32");
			_SetLayeredWindowAttributes SetLayeredWindowAttributes = (_SetLayeredWindowAttributes)User32.GetAddress("SetLayeredWindowAttributes");
			if (SetLayeredWindowAttributes)
			{
				if (NOT SetLayeredWindowAttributes(Handle(), 0, DRAG_COL_ALPHA, LWA_ALPHA))
				{
					DWORD Err = GetLastError();
				}
			}
		}
		
		#endif

		GMouse m;
		List->GetMouse(m);
		Offset = m.x - r.x1;

		List->PointToScreen(ListScrPos);
		r.Offset(ListScrPos.x, ListScrPos.y);

		SetPos(r);
		Visible(true);
	}
}

GDragColumn::~GDragColumn()
{
	Visible(false);

	if (Col)
	{
		Col->d->Drag = false;
	}

	List->Invalidate();
}

#if LINUX_TRANS_COL
void GDragColumn::OnPosChange()
{
	Invalidate();
}
#endif

void GDragColumn::OnPaint(GSurface *pScreen)
{
	#if LINUX_TRANS_COL
	GSurface *Buf = NEW(GMemDC(X(), Y(), GdcD->GetBits()));
	GSurface *pDC = NEW(GMemDC(X(), Y(), GdcD->GetBits()));
	#else
	GSurface *pDC = pScreen;
	#endif
	
	pDC->SetOrigin(Col->d->Pos.x1, 0);
	if (Col) Col->d->Drag = false;
	List->OnPaint(pDC);
	if (Col) Col->d->Drag = true;
	pDC->SetOrigin(0, 0);
	
	#if LINUX_TRANS_COL
	if (Buf AND pDC)
	{
		GRect p = GetPos();
		
		// Fill the buffer with the background
		Buf->Blt(ListScrPos.x - p.x1, 0, Back);

		// Draw painted column over the back with alpha
		Buf->Op(GDC_ALPHA);
		GApplicator *App = Buf->Applicator();
		if (App)
		{
			App->SetVar(GAPP_ALPHA_A, DRAG_COL_ALPHA);
		}
		Buf->Blt(0, 0, pDC);

		// Put result on the screen
		pScreen->Blt(0, 0, Buf);
	}
	DeleteObj(Buf);
	DeleteObj(pDC);
	#endif
}

////////////////////////////////////////////////////////////////////////////////////////////
#define M_END_POPUP (M_USER+0x656)

class GItemEditBox : public GEdit
{
	GItemEdit *ItemEdit;

public:
	GItemEditBox(GItemEdit *i, int x, int y, char *s) : GEdit(100, 1, 1, x-3, y-3, s)
	{
		ItemEdit = i;
		Sunken(false);
		
		SetPos(GetPos());
	}

	void OnCreate()
	{
		GEdit::OnCreate();
		Focus(true);
	}

	void OnFocus(bool f)
	{
		if (NOT f)
		{
			#ifdef WIN32
			HWND pop = GetLastActivePopup(Handle());
			HWND par = GetParent()->Handle();
			// LgiTrace("pop=%p\n", pop);
			if (pop AND
				pop != Handle() AND
				par != pop AND
				IsWindow(pop) AND
				IsWindowVisible(pop))
			{
				// LgiTrace("pop=%p and isvisible\n", pop);
				return;
			}
			#endif

			GetParent()->Visible(false);
		}
	}

	bool OnKey(GKey &k)
	{
		if (k.c16 == VK_RETURN OR
			k.c16 == VK_ESCAPE)
		{
			if (k.Down())
			{			
				ItemEdit->OnNotify(this, k.c16);
			}
			
			return true;
		}
		
		return false;
	}
};

class GItemEditPrivate
{
public:
	GItem *Item;
	GEdit *Edit;
	int Index;
	bool Esc;

	GItemEditPrivate()
	{
		Esc = false;
		Item = 0;
		Index = 0;
	}
};

GItemEdit::GItemEdit(GView *parent, GItem *item, int index, int SelStart, int SelEnd)
	: GPopup(parent)
{
	d = NEW(GItemEditPrivate);
	d->Item = item;
	d->Index = index;

	_BorderSize = 0;
	Sunken(false);
	Raised(false);
	SetParent(parent);
	
	GdcPt2 p;
	GetParent()->PointToScreen(p);

	GRect r = d->Item->GetPos(d->Index);
	r.Offset(p.x-1, p.y);
	// r.x2 += 2;
	r.y2 += 2;
	SetPos(r);

	if (Attach(GetParent()))
	{
		d->Edit = NEW(GItemEditBox(this, r.X(), r.Y(), d->Item->GetText(d->Index)));
		if (d->Edit)
		{
			d->Edit->Attach(this);
			d->Edit->Focus(true);
			
			if (SelStart >= 0)
			{
				d->Edit->Select(SelStart, SelEnd-SelStart+1);
			}
		}

		Visible(true);
	}
}

GItemEdit::~GItemEdit()
{
	if (d->Item)
	{
		if (d->Edit AND NOT d->Esc)
		{
			d->Item->SetText(d->Edit->Name(), d->Index);
			d->Item->Update();
		}

		d->Item->OnEditLabelEnd();
	}
}

void GItemEdit::OnPaint(GSurface *pDC)
{
	#if 1

	// pDC->Colour(Rgb24(255, 0, 255), 24);
	pDC->Colour(LC_BLACK, 24);
	pDC->Rectangle();

	#else

	GRect r = GetClient();
	pDC->Colour(LC_BLACK, 24);
	pDC->Box(&r);
	r.Size(1, 1);

	#endif
}

int GItemEdit::OnNotify(GViewI *v, int f)
{
	switch (v->GetId())
	{
		case 100:
		{
			if (f == VK_ESCAPE)
			{
				d->Esc = true;
			}

			if (f == VK_ESCAPE OR f == VK_RETURN)
			{
				Visible(false);
			}

			break;
		}
	}

	return 0;
}

void GItemEdit::Visible(bool i)
{
	GPopup::Visible(i);
	if (NOT i)
	{
		PostEvent(M_END_POPUP);
	}
}

int GItemEdit::OnEvent(GMessage *Msg)
{
	switch (MsgCode(Msg))
	{
		case M_END_POPUP:
		{
			Quit();
			return 0;
		}
	}

	return GPopup::OnEvent(Msg);
}

////////////////////////////////////////////////////////////////////////////////////////////
// List column
GListColumn::GListColumn(GList *parent, char *name, int width)
	: ResObject(Res_Column)
{
	d = NEW(GListColumnPrivate(parent));
	d->cWidth = width;
	if (name)
		Name(name);
}

GListColumn::~GListColumn()
{
	DeleteObj(d);
}

GList *GListColumn::GetList()
{
	return d->Parent;
}

void GListColumn::Image(int i)
{
	d->cImage = i;
}

int GListColumn::Image()
{
	return d->cImage;
}

bool GListColumn::Resizable()
{
	return d->CanResize;
}

void GListColumn::Resizable(bool i)
{
	d->CanResize = i;
}

void GListColumn::Name(char *n)
{
	DeleteArray(d->cName);
	DeleteObj(d->Txt);
	d->cName = NewStr(n);
	d->Txt = NEW(GDisplayString(SysFont, n));
	if (d->Parent)
	{
		d->Parent->Invalidate(&d->Parent->ColumnHeader);
	}
}

char *GListColumn::Name()
{
	return d->cName;
}

int GListColumn::GetIndex()
{
	if (d->Parent)
	{
		return d->Parent->Columns.IndexOf(this);
	}

	return -1;
}

int GListColumn::GetContentSize()
{
	int Max = 0;

	if (d->Parent)
	{
		int Index = GetIndex();

		Iterator<GListItem> Items(&d->Parent->Items);
		for (GListItem *i=Items.First(); i; i=Items.Next())
		{
			GDisplayString *s = i->d->Display[Index];
			GDisplayString *Mem = 0;
			
			// If no cached string, create it for the list item
			if (NOT s OR s->IsTruncated())
			{
				GFont *f = i->GetFont();
				if (NOT f) f = d->Parent->GetFont();
				if (NOT f) f = SysFont;

				char *Text = i->d->Str[Index] ? i->d->Str[Index] : i->GetText(Index);
				if (s AND s->IsTruncated())
				{
					s = Mem = NEW(GDisplayString(f, Text?Text:(char*)""));
				}
				else
				{
					s = i->d->Display[Index] = NEW(GDisplayString(f, Text?Text:(char*)""));
				}
			}

			// Measure it
			if (s)
			{
				Max = max(Max, s->X());
			}
			
			DeleteObj(Mem);
		}
		
		// Measure the heading too
		GFont *f = d->Parent->GetFont();
		LgiAssert(f);
		if (f)
		{
			GDisplayString h(d->Parent->GetFont(), d->cName);
			int Hx = h.X() + (d->cMark ? 10 : 0);
			Max = max(Max, Hx);
		}
	}

	return Max;
}

void GListColumn::Width(int i)
{
	if (d->cWidth != i)
	{
		d->cWidth = i;
		
		// If we are attached to a list...
		if (d->Parent)
		{
			int MyIndex = GetIndex();

			// Clear all the cached strings for this column
			Iterator<GListItem> Items(&d->Parent->Items);
			for (GListItem *n=Items.First(); n; n=Items.Next())
			{
				DeleteObj(n->d->Display[MyIndex]);
			}

			if (d->Parent->IsAttached())
			{
				// Update the screen from this column across
				GRect Up = d->Parent->GetClient();
				Up.x1 = d->Pos.x1;
				d->Parent->Invalidate(&Up);
			}
		}

		// Notify listener
		GViewI *n = d->Parent->GetNotify() ? d->Parent->GetNotify() : d->Parent->GetParent();
		if (n)
		{
			n->OnNotify(d->Parent, GLIST_NOTIFY_COLS_SIZE);
		}
	}
}

int GListColumn::Width()
{
	return d->cWidth;
}

void GListColumn::Mark(int i)
{
	d->cMark = i;
	if (d->Parent)
	{
		d->Parent->Invalidate(&d->Parent->ColumnHeader);
	}
}

int GListColumn::Mark()
{
	return d->cMark;
}

void GListColumn::Type(int i)
{
	d->cType = i;
	if (d->Parent)
	{
		d->Parent->Invalidate(&d->Parent->ColumnHeader);
	}
}

int GListColumn::Type()
{
	return d->cType;
}

void GListColumn::Icon(GSurface *i, bool Own)
{
	if (d->OwnIcon)
	{
		DeleteObj(d->cIcon);
	}
	d->cIcon = i;
	d->OwnIcon = Own;

	if (d->Parent)
	{
		d->Parent->Invalidate(&d->Parent->ColumnHeader);
	}
}

GSurface *GListColumn::Icon()
{
	return d->cIcon;
}

int GListColumn::Value()
{
	return d->Down;
}

void GListColumn::OnPaint_Content(GSurface *pDC, GRect &r, bool FillBackground)
{
	if (NOT d->Drag)
	{
		int Off = d->Down ? 1 : 0;
		int Mx = r.x1 + 8, My = r.y1 + ((r.Y() - 8) / 2);
		if (d->cIcon)
		{
			if (FillBackground)
			{
				pDC->Colour(LC_MED, 24);
				pDC->Rectangle(&r);
			}

			int x = (r.X()-d->cIcon->X()) / 2;
			pDC->Blt(	r.x1 + x + Off,
						r.y1 + ((r.Y()-d->cIcon->Y())/2) + Off,
						d->cIcon);

			if (d->cMark)
			{
				Mx += x + d->cIcon->X() + 4;
			}
		}
		else if (d->cImage >= 0 AND d->Parent)
		{
			if (FillBackground)
			{
				pDC->Colour(LC_MED, 24);
				pDC->Rectangle(&r);
			}
			
			if (d->Parent->ImageList)
			{
				GRect *b = d->Parent->ImageList->GetBounds();
				int x = r.x1;
				int y = r.y1;
				if (b)
				{
					b += d->cImage;
					x = r.x1 + ((r.X()-b->X()) / 2) - b->x1;
					y = r.y1 + ((r.Y()-b->Y()) / 2) - b->y1;
				}				
				
				d->Parent->ImageList->Draw(pDC,
											x + Off,
											y + Off,
											d->cImage, 0);
			}

			if (d->cMark)
			{
				Mx += d->Parent->ImageList->TileX() + 4;
			}
		}
		else if (ValidStr(d->cName) AND d->Txt)
		{
			SysFont->Transparent(NOT FillBackground);
			
			SysFont->Colour(LC_TEXT, LC_MED);
			d->Txt->Draw(pDC, r.x1 + Off + 3, r.y1 + Off, &r);

			if (d->cMark)
			{
				Mx += d->Txt->X(); // SysFont->X(d->cName);
			}
		}
		else
		{
			if (FillBackground)
			{
				pDC->Colour(LC_MED, 24);
				pDC->Rectangle(&r);
			}
		}

		#define ARROW_SIZE	9
		pDC->Colour(LC_TEXT, 24);
		Mx += Off;
		My += Off - 1;

		switch (d->cMark)
		{
			case GLI_MARK_UP_ARROW:
			{
				pDC->Line(Mx + 2, My, Mx + 2, My + ARROW_SIZE - 1);
				pDC->Line(Mx, My + 2, Mx + 2, My);
				pDC->Line(Mx + 2, My, Mx + 4, My + 2);
				break;
			}
			case GLI_MARK_DOWN_ARROW:
			{
				pDC->Line(Mx + 2, My, Mx + 2, My + ARROW_SIZE - 1);
				pDC->Line(	Mx,
							My + ARROW_SIZE - 3,
							Mx + 2,
							My + ARROW_SIZE - 1);
				pDC->Line(	Mx + 2,
							My + ARROW_SIZE - 1,
							Mx + 4,
							My + ARROW_SIZE - 3);
				break;
			}
		}
	}
}

void ColumnPaint(void *UserData, GSurface *pDC, GRect &r, bool FillBackground)
{
	((GListColumn*)UserData)->OnPaint_Content(pDC, r, FillBackground);
}

void GListColumn::OnPaint(GSurface *pDC, GRect &Rgn)
{
	GRect r = Rgn;

	if (d->Drag)
	{
		pDC->Colour(DragColumnColour, 24);
		pDC->Rectangle(&r);
	}
	else
	{
		int Off = d->Down ? 1 : 0;

		if (GApp::SkinEngine)
		{
			GSkinState State;
			
			State.pScreen = pDC;			
			State.Text = &d->Txt;
			State.Rect = Rgn;
			State.Value = Value();
			State.Enabled = GetList()->Enabled();

			GApp::SkinEngine->OnPaint_ListColumn(ColumnPaint, this, &State);
		}
		else
		{
			if (d->Down)
			{
				LgiThinBorder(pDC, r, SUNKEN);
				LgiFlatBorder(pDC, r, 1);
			}
			else
			{
				LgiWideBorder(pDC, r, RAISED);
			}
			
			OnPaint_Content(pDC, r, true);
		}		
	}
}

////////////////////////////////////////////////////////////////////////////////////////////
GListItemColumn::GListItemColumn(GListItem *item, int col)
{
	_Column = col;
	_Item = item;
	_Value = 0;
	_Item->d->Cols.Insert(this);
}

GList *GListItemColumn::GetList()
{
	return _Item ? _Item->Parent : 0;
}

List<GListItem> *GListItemColumn::GetAllItems()
{
	return GetList() ? &GetList()->Items : 0;
}

void GListItemColumn::Value(int64 i)
{
	if (i != _Value)
	{
		_Value = i;
		_Item->Update();
		_Item->OnColumnNotify(_Column, _Value);
	}
}

GListItemColumn *GListItemColumn::GetItemCol(GListItem *i, int Col)
{
	if (i)
	{
		for (GListItemColumn *c=i->d->Cols.First(); c; c=i->d->Cols.Next())
		{
			if (c->_Column == Col)
			{
				return c;
			}
		}
	}

	return 0;
}

////////////////////////////////////////////////////////////////////////////////////////////
// List item
GListItem::GListItem()
{
	d = NEW(GListItemPrivate);
	Pos.ZOff(-1, -1);
	Parent = 0;
}

GListItem::~GListItem()
{
	if (Parent)
	{
		Parent->Remove(this);
	}

	DeleteObj(d);
}

void GListItem::OnEditLabelEnd()
{
	if (Parent)
	{
		Parent->d->Edit = 0;
	}
}

void GListItem::SetImage(int i)
{
	d->ListItem_Image = i;
}

int GListItem::GetImage(int Flags)
{
	return d->ListItem_Image;
}

List<GListItemColumn> *GListItem::GetItemCols()
{
	return &d->Cols;
}

/*
Calling this to store your data is optional. Just override the
"GetText" function to return your own data to avoid duplication
in memory.
*/
bool GListItem::SetText(char *s, int i)
{
	if (s AND i >= 0)
	{
		// Delete any existing column
		DeleteArray((char*&)d->Str[i]);
		DeleteObj(d->Display[i]);

		// Add new string in
		d->Str[i] = NewStr(s);

		return true;
	}

	return false;
}

// User can override this if they want to use their own data
char *GListItem::GetText(int i)
{
	return d->Str[i];
}

bool GListItem::Select()
{
	return d->Selected;
}

void GListItem::SetEditLabelSelection(int SelStart, int SelEnd)
{
	d->SelectionStart = SelStart;
	d->SelectionEnd = SelEnd;
}

GView *GListItem::EditLabel(int Col)
{
	GetList()->Capture(false);

	if (Parent)
	{
		if (NOT Parent->d->Edit)
		{
			Parent->d->Edit = NEW(GItemEdit(Parent, this, Col, d->SelectionStart, d->SelectionEnd));
			d->SelectionStart = d->SelectionEnd = -1;
		}

		return Parent->d->Edit;
	}

	return 0;	
}

GRect *GListItem::GetPos(int Col)
{
	static GRect r;

	r = Pos;

	if (Parent->GetMode() == GListDetails)
	{
		if (Col >= 0)
		{
			GListColumn *Column = 0;

			int Cx = Parent->GetImageList() ? 16 : 0;
			for (int c=0; c<Col; c++)
			{
				Column = Parent->ColumnAt(c);
				if (Column)
				{
					Cx += Column->Width();
				}
			}
			Column = Parent->ColumnAt(Col);

			if (Column)
			{
				r.x1 = Cx;
				r.x2 = Cx + Column->Width() - 1;
			}
		}
	}
	else
	{
		r.Offset(16, 0);
	}

	return &r;
}

void GListItem::Select(bool b)
{
	if (d->Selected != b)
	{
		d->Selected = b;
		Update();
		
		if (Parent AND
			d->Selected AND
			NOT Parent->d->NoSelectEvent)
		{
			Parent->OnItemSelect(this);
		}
	}
}

void GListItem::ScrollTo()
{
	if (Parent)
	{
		if (Parent->GetMode() == GListDetails AND Parent->VScroll)
		{
			int n = Parent->Items.IndexOf(this);
			if (n < Parent->FirstVisible)
			{
				Parent->VScroll->Value(n);
				Parent->Invalidate(&Parent->ItemsPos);
			}
			else if (n >= Parent->LastVisible)
			{
				Parent->VScroll->Value(n - (Parent->LastVisible - Parent->FirstVisible) + 1);
				Parent->Invalidate(&Parent->ItemsPos);
			}
		}
		else if (Parent->GetMode() == GListColumns AND Parent->HScroll)
		{
			int n = Parent->Items.IndexOf(this);
			if (n < Parent->FirstVisible)
			{
				printf("Down LayoutColumn=%i\n", d->LayoutColumn);

				Parent->HScroll->Value(d->LayoutColumn);
				Parent->Invalidate(&Parent->ItemsPos);
			}
			else if (n >= Parent->LastVisible)
			{
				int Range = Parent->HScroll->Page();
				
				printf("Up LayoutColumn=%i Range=%i\n", d->LayoutColumn, Range);
				
				Parent->HScroll->Value(d->LayoutColumn - Range);
				Parent->Invalidate(&Parent->ItemsPos);
			}
		}
	}
}

void GListItem::Update()
{
	d->EmptyDisplay();

	if (Parent)
	{
		GMeasureInfo Info;
		OnMeasure(&Info);

		GRect r = Pos;
		if (r.Valid())
		{
			if (Info.y != r.Y())
			{
				r.y2 = Parent->ItemsPos.y2;
			}

			Parent->Invalidate(&r);
		}
	}
}

void GListItem::OnMeasure(GMeasureInfo *Info)
{
	if (Info)
	{
		if (Parent->GetMode() == GListDetails)
		{
			Info->x = 1024;
		}
		else
		{
			GDisplayString *s = GetDs(0);
			Info->x = 22 + (s ? s->X() : 0);
		}
		
		Info->y = max(16, SysFont->GetHeight() + 2); // the default height
	}
}

bool GListItem::GridLines()
{
	return (Parent) ? Parent->GridLines : false;
}

void GListItem::OnMouseClick(GMouse &m)
{
	int Col = Parent->ColumnAtX(m.x);
	for (GListItemColumn *h = d->Cols.First(); h; h=d->Cols.Next())
	{
		if (Col == h->GetColumn())
		{
			h->OnMouseClick(m);
		}
	}
}

GDisplayString *GListItem::GetDs(int Col, int FitTo)
{
	if (NOT d->Display[Col])
	{
		GFont *f = GetFont();
		if (NOT f) f = Parent->GetFont();
		if (NOT f) f = SysFont;

		char *Text = d->Str[Col] ? d->Str[Col] : GetText(Col);
		d->Display[Col] = NEW(GDisplayString(f, Text?Text:(char*)""));
		if (d->Display[Col] AND FitTo > 0)
		{
			d->Display[Col]->TruncateWithDots(FitTo);
		}
	}
	return d->Display[Col];
}

void GListItem::OnPaintColumn(GSurface *pDC, GRect &r, int i, GListColumn *c)
{
	if (pDC AND c)
	{
		GRect ng = r; // non-grid area

		if (c->d->Drag)
		{
			pDC->Colour(DragColumnColour, 24);
			pDC->Rectangle(&ng);
		}
		else
		{
			COLOUR Background;
			if (Parent->GetMode() == GListDetails AND
				c->Mark() AND
				NOT d->Selected)
			{
				Background = GdcMixColour(0, Back, (double)1/32);
			}
			else
			{
				Background = Back;
			}
			if (GridLines())
			{
				ng.x2--;
				ng.y2--;
			}

			if (c->Type() == GIC_ASK_TEXT)
			{
				GDisplayString *Ds = GetDs(i, r.X());
				if (Ds)
				{
					Ds->GetFont()->TabSize(0);
					Ds->GetFont()->Transparent(false);
					Ds->GetFont()->Colour(Fore, Background);
					Ds->Draw(pDC, r.x1+1, r.y1+1, &ng);
				}
				else
				{
					pDC->Colour(Background, 24);
					pDC->Rectangle(&ng);
				}
			}
			else
			{
				pDC->Colour(Background, 24);
				pDC->Rectangle(&ng);

				if (c->Type() == GIC_ASK_IMAGE AND
					Parent->ImageList)
				{
					int Img = GetImage();
					if (Img >= 0)
					{
						Parent->ImageList->Draw(pDC, r.x1+1, r.y1, GetImage(), 0);
					}
				}
			}

			if (GridLines())
			{
				pDC->Colour(LC_LOW, 24);
				pDC->Line(r.x1, r.y2, r.x2, r.y2);
				pDC->Line(r.x2, r.y1, r.x2, r.y2);
			}
		}
	}
}

void GListItem::OnPaint(GSurface *pDC, GRect &r)
{
	if (Parent)
	{
		int i=0;
		int x = r.x1;
		
		// Icon?
		if (Parent->IconCol)
		{
			GRect cr(x, r.y1, x + Parent->IconCol->Width()-1, r.y2);

			// icon background
			pDC->Colour(d->Selected ? Parent->_Selection : LC_WORKSPACE, 24);
			pDC->Rectangle(&cr);

			// draw icon
			OnPaintColumn(pDC, cr, -1, Parent->IconCol);
			x = cr.x2 + 1;
		}

		// draw columns
		GListItemColumn *h = d->Cols.First();
		for (GListColumn *c = Parent->Columns.First(); c; c = Parent->Columns.Next(), i++)
		{
			GRect cr;
			if (Parent->GetMode() == GListColumns)
				cr.Set(x, r.y1, r.x2, r.y2);
			else
				cr.Set(x, r.y1, x + c->Width()-1, r.y2);
			
			OnPaintColumn(pDC, cr, i, c);
			if (h AND i == h->GetColumn())
			{
				h->OnPaintColumn(pDC, cr, i, c);
				h = d->Cols.Next();
			}
			x = cr.x2 + 1;
			
			if (Parent->GetMode() == GListColumns)
				break;
		}

		// after columns
		if (x < r.x2)
		{
			pDC->Colour(Back, 24);
			pDC->Rectangle(x, r.y1, r.x2, r.y2);
		}
	}
}

//////////////////////////////////////////////////////////////////////////////
// List control
GList::GList(int id, int x, int y, int cx, int cy, char *name)
	: ResObject(Res_ListView)
{
	d = NEW(GListPrivate);
	SetId(id);
	Name(name);
	
	Buf = 0;
	GridLines = false;
	FirstVisible = -1;
	LastVisible = -1;
	ColumnHeaders = true;
	ColumnHeader.ZOff(-1, -1);
	EditLabels = false;
	MultiItemSelect = true;
	IconCol = 0;
	CompletelyVisible = 0;
	Keyboard = -1;
	Sunken(true);
	Name("GList");

	#if defined WIN32
	SetStyle(GetStyle() | WS_TABSTOP);
	SetDlgCode(DLGC_WANTARROWS);
	Cursor = 0;
	#elif defined BEOS
	Handle()->SetViewColor(B_TRANSPARENT_COLOR);
	#elif defined LINUX
	Handle()->isTabStop(true);
	#endif

	GRect r(x, y, x+cx, y+cy);
	SetPos(r);
}

GList::~GList()
{
	DeleteObj(Buf);

	Empty();
	EmptyColumns();
	DeleteObj(d);
}

GListMode GList::GetMode()
{
	return d->Mode;
}

void GList::SetMode(GListMode m)
{
	if (d->Mode ^ m)
	{
		d->Mode = m;
		
		if (IsAttached())
		{
			Pour();
			Invalidate();
		}
	}
}

void GList::OnItemClick(GListItem *Item, GMouse &m)
{
	if (Item)
	{
		Item->OnMouseClick(m);
	}
}

void GList::OnItemBeginDrag(GListItem *Item, GMouse &m)
{
	if (Item)
	{
		Item->OnBeginDrag(m);
	}
}

void GList::OnItemSelect(GListItem *Item)
{
	if (Item)
	{
		Keyboard = Items.IndexOf(Item);
		Item->OnSelect();

		if (NOT MultiItemSelect AND
			Item->d->Selected)
		{
			// deselect all other items
			ForAllItems(i)
			{
				if (i != Item)
				{
					if (i->d->Selected)
					{
						i->d->Selected = false;
						i->Update();
					}
				}
			}
		}
	}

	// Notify selection change
	SendNotify(GLIST_NOTIFY_SELECT);
}

GListColumn *GList::AddColumn(char *Name, int Width, int Where)
{
	GListColumn *c = NEW(GListColumn(this, Name, Width));
	if (c)
	{
		Columns.Insert(c, Where);
		SendNotify(GLIST_NOTIFY_COLS_CHANGE);
	}
	return c;
}

bool GList::AddColumn(GListColumn *Col, int Where)
{
	bool Status = false;

	if (Col)
	{
		Status = Columns.Insert(Col, Where);
		if (Status)
		{
			SendNotify(GLIST_NOTIFY_COLS_CHANGE);
		}
	}

	return Status;
}

bool GList::DeleteColumn(GListColumn *Col)
{
	if (Col AND
		Columns.HasItem(Col))
	{
		Columns.Delete(Col);
		DeleteObj(Col);
		Invalidate(&ColumnHeader);

		SendNotify(GLIST_NOTIFY_COLS_CHANGE);
		return true;
	}

	return false;
}

void GList::DragColumn(int Index)
{
	DeleteObj(d->DragCol);
	d->DragCol = NEW(GDragColumn(this, Index));
	if (d->DragCol)
	{
		Capture(true);
		d->DragMode = DRAG_COLUMN;
	}
}

int GList::ColumnAtX(int x, GListColumn **Col, int *Offset)
{
	GListColumn *Column = 0;
	if (NOT Col) Col = &Column;

	int Cx = GetImageList() ? 16 : 0;
	int c;
	for (c=0; *Col = Columns.ItemAt(c); c++)
	{
		if (x >= Cx AND x < Cx + (*Col)->Width())
		{
			break;
		}
		Cx += (*Col)->Width();
	}

	if (*Col)
	{
		if (Offset)
		{
			*Offset = Cx;
		}

		return c;
	}

	return -1;
}

void GList::EmptyColumns()
{
	Columns.DeleteObjects();
	DeleteObj(IconCol);
	Invalidate(&ColumnHeader);
	SendNotify(GLIST_NOTIFY_COLS_CHANGE);
}

int GList::OnEvent(GMessage *Msg)
{
	return GLayout::OnEvent(Msg);
}

int GList::OnNotify(GViewI *Ctrl, int Flags)
{
	if
	(
		(Ctrl->GetId() == IDC_VSCROLL AND VScroll)
		OR
		(Ctrl->GetId() == IDC_HSCROLL AND HScroll)
	)
	{
		Invalidate(&ItemsPos);
	}

	return 0;
}

GRect &GList::GetClientRect()
{
	static GRect r;

	r = GetPos();
	r.Offset(-r.x1, -r.y1);

	return r;
}

GListItem *GList::HitItem(int x, int y, int *Index)
{
	int n=0;
	ForAllItems(i)
	{
		// if (y >= i->Pos.y1 AND y <= i->Pos.y2)
		if (i->Pos.Overlap(x, y))
		{
			if (Index) *Index = n;
			return i;
		}
		n++;
	}

	return NULL;
}

void GList::KeyScroll(int iTo, int iFrom, bool SelectItems)
{
	int Start = -1, End = -1, i = 0;
	{
		ForAllItems(n)
		{
			if (n->Select())
			{
				if (Start < 0)
				{
					Start = i;
				}
			}
			else if (Start >= 0 AND
					End < 0)
			{
				End = i - 1;
			}

			i++;
		}

		if (End < 0) End = i - 1;
	}

	iTo = limit(iTo, 0, Items.Length()-1);
	iFrom = limit(iFrom, 0, Items.Length()-1);
	GListItem *To = Items.ItemAt(iTo);
	GListItem *From = Items.ItemAt(iFrom);
	int Inc = (iTo < iFrom) ? -1 : 1;

	if (To AND From AND iTo != iFrom)
	{
		GListItem *Item = 0;
		if (SelectItems)
		{
			int OtherEnd = Keyboard == End ? Start : End;
			int Min = min(OtherEnd, iTo);
			int Max = max(OtherEnd, iTo);
			i = 0;

			d->NoSelectEvent = true;
			ForAllItems(n)
			{
				n->Select(i>=Min AND i<=Max);
				i++;
			}
			d->NoSelectEvent = false;
			OnItemSelect(First());
		}
		else
		{
			Select(To);
		}

		To->ScrollTo();

		Keyboard = iTo;
	}
}

void GList::OnMouseWheel(double Lines)
{
	if (VScroll)
	{
		int Old = VScroll->Value();
		VScroll->Value(Old + (int)Lines);
		if (Old != VScroll->Value())
		{
			Invalidate(&ItemsPos);
		}
	}
	
	if (HScroll)
	{
		int Old = HScroll->Value();
		HScroll->Value(Old + (int)(Lines / 3));
		if (Old != HScroll->Value())
		{
			Invalidate(&ItemsPos);
		}
	}
}

bool GList::OnKey(GKey &k)
{
	bool Status = false;

	GListItem *Item = GetSelection();
	if (Item)
	{
		Status = Item->OnKey(k);
	}

	if (k.Down())
	{
		if (k.Ctrl())
		{
			switch (k.c16)
			{
				case 'A':
				case 'a':
				{
					SelectAll();
					Status = true;
					break;
				}
			}
		}
		else
		{
			switch (k.c16)
			{
				case VK_RETURN:
				{
					#ifdef WIN32
					if (NOT k.IsChar)
					#endif
					{
						SendNotify(GLIST_NOTIFY_RETURN);
					}
					break;
				}
				case VK_ESCAPE:
				{
					SendNotify(GLIST_NOTIFY_ESC_KEY);
					break;
				}
				case VK_DELETE:
				{
					SendNotify(GLIST_NOTIFY_DEL_KEY);
					break;					
				}
				case VK_UP:
				{
					// int i = Value();
					KeyScroll(Keyboard-1, Keyboard, k.Shift());
					Status = true;
					break;
				}
				case VK_DOWN:
				{
					KeyScroll(Keyboard+1, Keyboard, k.Shift());
					Status = true;
					break;
				}
				case VK_LEFT:
				{
					if (GetMode() == GListColumns)
					{
						GListItem *Hit = GetSelection();
						if (Hit)
						{
							GListItem *To = 0;
							int ToDist = 0x7fffffff;
							
							for (GListItem *i=Items[FirstVisible]; i AND i->Pos.Valid(); i=Items.Next())
							{
								if (i->Pos.x2 < Hit->Pos.x1)
								{
									int Dx = i->Pos.x1 - Hit->Pos.x1;
									int Dy = i->Pos.y1 - Hit->Pos.y1;
									int IDist = Dx * Dx + Dy * Dy;
									if (NOT To OR IDist < ToDist)
									{
										To = i;
										ToDist = IDist;
									}
								}
							}
							
							if (NOT To AND HScroll)
							{
								if (Hit->d->LayoutColumn == HScroll->Value() + 1)
								{
									Iterator<GListItem> It(&Items);
									GListItem *i;
									
									// Seek back to the start of the column before the 
									// first visible column
									for (i=It[FirstVisible]; i; i=It.Prev())
									{
										if (i->d->LayoutColumn < HScroll->Value())
										{
											It.Next();
											break;
										}
									}
									
									// Now find the entry at the right height
								}
							}

							if (To)
							{
								Select(0);
								To->Select(true);
								To->ScrollTo();
							}
						}
					}
					Status = true;
					break;
				}
				case VK_RIGHT:
				{
					if (GetMode() == GListColumns)
					{
						GListItem *Hit = GetSelection();
						if (Hit)
						{
							GListItem *To = 0;
							int ToDist = 0x7fffffff;
							
							for (GListItem *i=Items[FirstVisible]; i AND i->Pos.Valid(); i=Items.Next())
							{
								if (i->Pos.x1 > Hit->Pos.x2)
								{
									int Dx = i->Pos.x1 - Hit->Pos.x1;
									int Dy = i->Pos.y1 - Hit->Pos.y1;
									int IDist = Dx * Dx + Dy * Dy;
									if (NOT To OR IDist < ToDist)
									{
										To = i;
										ToDist = IDist;
									}
								}
							}
							
							if (To)
							{
								Select(0);
								To->Select(true);
								To->ScrollTo();
							}
						}
					}
					Status = true;
					break;
				}
				case VK_PAGEUP:
				{
					int Vis = VisibleItems();
					Vis = max(Vis, 0);

					KeyScroll(Keyboard-Vis, Keyboard, k.Shift());
					Status = true;
					break;
				}
				case VK_PAGEDOWN:
				{
					int Vis = VisibleItems();
					Vis = max(Vis, 0);
					KeyScroll(Keyboard+Vis, Keyboard, k.Shift());
					Status = true;
					break;
				}
				case VK_END:
				{
					// int i = Value();
					KeyScroll(Items.Length()-1, Keyboard, k.Shift());
					Status = true;
					break;
				}
				case VK_HOME:
				{
					KeyScroll(0, Keyboard, k.Shift());
					Status = true;
					break;
				}
				#ifdef VK_APPS
				case VK_APPS:
				{
					GListItem *s = GetSelection();
					if (s)
					{
						GRect *r = &s->Pos;
						if (r)
						{
							GMouse m;
							GListItem *FirstVisible = ItemAt((VScroll) ? VScroll->Value() : 0);

							m.x = 32 + ItemsPos.x1;
							m.y = r->y1 + (r->Y() >> 1) - (FirstVisible ? FirstVisible->Pos.y1 : 0) + ItemsPos.y1;
							
							m.Target = this;
							m.ViewCoords = true;
							m.Down(true);
							m.Right(true);
							
							OnMouseClick(m);
						}
					}
					break;
				}
				#endif
				default:
				{
					if
					(
						NOT Status
						AND
						k.IsChar
						AND
						(
							isdigit(k.c16)
							OR
							isalpha(k.c16)
							OR
							strchr("_.-", k.c16)
						)
					)
					{
						int Now = LgiCurrentTime();
						GStringPipe p;
						if (d->KeyBuf AND Now < d->KeyLast + 1500)
						{
							p.Push(d->KeyBuf);
						}
						DeleteArray(d->KeyBuf);
						d->KeyLast = Now;
						p.Push(&k.c16, 1);
						d->KeyBuf = p.NewStrW();
						if (d->KeyBuf)
						{
							char *c8 = LgiNewUtf16To8(d->KeyBuf);
							if (c8)
							{
								int Col = 0;
								bool Ascend = true;
								for (GListColumn *c = Columns.First(); c; c = Columns.Next())
								{
									if (c->Mark())
									{
										Col = Columns.IndexOf(c);
										if (c->Mark() == GLI_MARK_UP_ARROW)
										{
											Ascend = false;
										}
									}
								}
								
								bool Selected = false;
								Iterator<GListItem> It(&Items);
								for (GListItem *i = Ascend ? It.First() : It.Last(); i; i = Ascend ? It.Next() : It.Prev())
								{
									if (NOT Selected)
									{
										char *t = i->GetText(Col);
										if (t AND stricmp(t, c8) >= 0)
										{
											i->Select(true);
											i->ScrollTo();
											Selected = true;
										}
										else
										{
											i->Select(false);
										}
									}
									else
									{
										i->Select(false);
									}
								}
								
								DeleteArray(c8);
							}
						}						
						Status = true;
					}
					break;
				}
			}
		}
	}

	return Status;
}

int GList::HitColumn(int x, int y, GListColumn *&Resize, GListColumn *&Over)
{
	int Index = -1;

	Resize = 0;
	Over = 0;

	if (ColumnHeaders AND
		ColumnHeader.Overlap(x, y))
	{
		// Clicked on a column heading
		int cx = ColumnHeader.x1 + ((IconCol) ? IconCol->Width() : 0);
		int n = 0;
		for (GListColumn *c = Columns.First(); c; c = Columns.Next(), n++)
		{
			cx += c->Width();
			if (abs(x-cx) < 5)
			{
				if (c->Resizable())
				{
					Resize = c;
					Index = n;
					break;
				}
			}
			else if (c->d->Pos.Overlap(x, y))
			{
				Over = c;
				Index = n;
				break;
			}
		}
	}

	return Index;
}

int GList::OnHitTest(int x, int y)
{
	#ifdef WIN32
	GListColumn *Resize, *Over;
	HitColumn(x, y, Resize, Over);
	if (Resize)
	{
		SetCursor(LCUR_SizeHor);
		return 1;
	}
	#endif

	return -1;
}

void GList::OnMouseClick(GMouse &m)
{
	if (Lock())
	{
		if (m.Down())
		{
			Focus(true);

			d->DragMode = DRAG_NONE;
			d->DragStart.x = m.x;
			d->DragStart.y = m.y;

			if (ColumnHeaders AND
				ColumnHeader.Overlap(m.x, m.y))
			{
				// Clicked on a column heading
				GListColumn *Resize, *Over;
				int Index = HitColumn(m.x, m.y, Resize, Over);

				if (Resize)
				{
					if (m.Double())
					{
						Resize->Width(Resize->GetContentSize() + DEFAULT_COLUMN_SPACING);
					}
					else
					{
						d->DragMode = RESIZE_COLUMN;
						d->DragData = Columns.IndexOf(Resize);
						Capture(true);
					}
				}
				else
				{
					d->DragMode = CLICK_COLUMN;
					d->DragData = Columns.IndexOf(Over);
					if (Over)
					{
						Over->d->Down = true;
						Invalidate(&Over->d->Pos);
						Capture(true);
					}
				}
			}
			else if (ItemsPos.Overlap(m.x, m.y))
			{
				// Clicked in the items area
				int ItemIndex = -1;
				GListItem *Item = HitItem(m.x, m.y, &ItemIndex);
				GViewI *Notify = Item ? (GetNotify()) ? GetNotify() : GetParent() : 0;
				d->DragData = ItemIndex;

				if (Item AND Item->Select())
				{
					// Click on selected item
					if (m.Ctrl())
					{
						Item->Select(false);
						OnItemClick(Item, m);
					}
					else
					{
						// Could be drag'n'drop operation
						// Or just a select
						int64 StartHandler = LgiCurrentTime();

						// this will get set if 'this' is deleted.
						bool DeleteFlag = false;

						// Setup the delete flag pointer
						d->DeleteFlag = &DeleteFlag;

						// Do the event... may delete 'this' object, or hang for a long time
						OnItemClick(Item, m);

						// If the object has been deleted... exit out of here NOW!
						if (DeleteFlag)
						{
							return;
						}

						// Shut down the delete flag pointer... it'll point to invalid stack soon.
						d->DeleteFlag = 0;

						// Check if the handler hung for a long time...
						bool HandlerHung = LgiCurrentTime() - StartHandler > 200;
						if (NOT HandlerHung)
						{
							// Start d'n'd watcher pulse...
							SetPulse(100);
							Capture(true);
							d->DragMode = CLICK_ITEM;
						}
						
						if (NOT IsCapturing())
						{
							// If capture failed then we reset the dragmode...
							d->DragMode = DRAG_NONE;
						}
					}
				}
				else
				{
					// Selection change
					if (m.Shift() AND MultiSelect())
					{
						int n = 0;
						int a = min(ItemIndex, Keyboard);
						int b = max(ItemIndex, Keyboard);

						ForAllItems(i)
						{
							i->Select(n >= a AND n <= b);
							n++;
						}
					}
					else
					{
						bool PostSelect = false;
						
						ForAllItems(i)
						{
							if (Item == i) // clicked item
							{
								if (m.Ctrl())
								{
									// Toggle selected state
									if (NOT i->Select())
									{
										Keyboard = Items.IndexOf(i);
									}

									i->Select(NOT i->Select());
								}
								else
								{
									// Select this after we have delselected everything else
									PostSelect = true;
								}
							}
							else if (NOT m.Ctrl() OR NOT MultiSelect())
							{
								i->Select(false);
							}
						}

						if (PostSelect)
						{
							Item->Select(true);
							Keyboard = Items.IndexOf(Item);
						}

						if (NOT m.Ctrl() AND Items.First())
						{
							d->DragMode = SELECT_ITEMS;
							SetPulse(100);
							Capture(true);
						}
					}

					OnItemClick(Item, m);
				}

				if (Notify)
				{
					Notify->OnNotify(this, m.Double() ? GLIST_NOTIFY_DBL_CLICK : GLIST_NOTIFY_CLICK);
				}
			}
		}
		else // Up Click
		{
			switch (d->DragMode)
			{
				case CLICK_COLUMN:
				{
					GListColumn *c = Columns.ItemAt(d->DragData);
					if (c)
					{
						c->d->Down = false;
						Invalidate(&c->d->Pos);

						if (c->d->Pos.Overlap(m.x, m.y))
						{
							OnColumnClick(Columns.IndexOf(c), m);
						}
					}
					else
					{
						OnColumnClick(-1, m);
					}
					break;
				}
				case CLICK_ITEM:
				{
					GListItem *Item = Items.ItemAt(d->DragData);
					if (Item)
					{
						bool Change = false;
						bool Notify = false;
						
						ForAllItems(i)
						{
							bool Sel = Item == i;
							if (Sel ^ i->Select())
							{
								Change = true;
								i->Select(Sel);
								if (Sel)
								{
									Notify = true;
									OnItemSelect(i);
								}
							}
						}
						
						if (NOT Notify)
						{
							OnItemSelect(0);
						}
					}
					break;
				}
				case DRAG_COLUMN:
				{
					// End column drag
					if (d->DragCol)
					{
						GRect DragPos = d->DragCol->GetPos();
						GdcPt2 p(DragPos.x1 + (DragPos.X()/2), 0);
						PointToView(p);

						int OldIndex = d->DragCol->GetIndex();
						int Best = 100000000, NewIndex = OldIndex, i=0, delta;
						GListColumn *Col = 0;
						for (Col = Columns.First(); Col; Col = Columns.Next(), i++)
						{
							delta = abs(p.x - Col->d->Pos.x1);
							if (delta < Best)
							{
								Best = delta;
								NewIndex = i - (i > OldIndex ? 1 : 0);
							}
						}
						delta = abs(p.x - Columns.Last()->d->Pos.x2);
						if (delta < Best)
						{
							NewIndex = i;
						}

						Col = d->DragCol->GetColumn();
						if (OldIndex != NewIndex AND
							OnColumnReindex(Col, OldIndex, NewIndex))
						{
							Columns.Delete(Col);
							Columns.Insert(Col, NewIndex);
						}

						DeleteObj(d->DragCol);
					}

					Invalidate();
					break;
				}
			}

			GListItem *Item = HitItem(m.x, m.y);
			if (Item)
			{
				OnItemClick(Item, m);
			}

			if (IsCapturing())
			{
				Capture(false);
			}

			d->DragMode = DRAG_NONE;
		}

		Unlock();
	}
}

void GList::OnPulse()
{
	if (Lock())
	{
		if (IsCapturing())
		{
			GMouse m;
			GetMouse(m);

			if (m.y < 0 OR m.y >= Y())
			{
				switch (d->DragMode)
				{
					case SELECT_ITEMS:
					{
						int OverIndex = 0;
						GListItem *Over = 0;

						if (m.y < 0)
						{
							int Space = -m.y;

							int n = FirstVisible - 1;
							for (GListItem *i = Items[n]; i; i=Items.Prev(), n--)
							{
								GMeasureInfo Info;
								i->OnMeasure(&Info);
								if (Space > Info.y)
								{
									Space -= Info.y;
								}
								else
								{
									OverIndex = n;
									Over = i;
									break;
								}
							}

							if (NOT Over)
							{
								Over = Items.First();
								OverIndex = 0;
							}
						}
						else if (m.y >= Y())
						{
							int Space = m.y - Y();
							int n = LastVisible + 1;
							for (GListItem *i = Items[n]; i; i=Items.Next(), n++)
							{
								GMeasureInfo Info;
								i->OnMeasure(&Info);
								if (Space > Info.y)
								{
									Space -= Info.y;
								}
								else
								{
									OverIndex = n;
									Over = i;
									break;
								}
							}

							if (NOT Over)
							{
								Over = Items.Last();
								OverIndex = Items.Length()-1;
							}
						}

						int Min = min(d->DragData, OverIndex);
						int Max = max(d->DragData, OverIndex);
						int n = Min;
						for (GListItem *i = Items[Min]; i AND n <= Max; i=Items.Next(), n++)
						{
							i->Select(true);
						}

						if (Over)
						{
							Over->ScrollTo();
						}
						break;
					}
				}
			}
		}
		else
		{
			d->DragMode = DRAG_NONE;
			SetPulse();
		}

		Unlock();
	}
}

void GList::OnMouseMove(GMouse &m)
{
	if (Lock())
	{
		switch (d->DragMode)
		{
			case DRAG_COLUMN:
			{
				if (d->DragCol)
				{
					GdcPt2 p;
					PointToScreen(p);

					GRect r = d->DragCol->GetPos();
					r.Offset(-p.x, -p.y); // to view co-ord

					r.Offset(m.x - d->DragCol->GetOffset() - r.x1, 0);
					if (r.x1 < 0) r.Offset(-r.x1, 0);
					if (r.x2 > X()-1) r.Offset((X()-1)-r.x2, 0);

					r.Offset(p.x, p.y); // back to screen co-ord
					d->DragCol->SetPos(r, true);
					r = d->DragCol->GetPos();
				}
				break;
			}
			case RESIZE_COLUMN:
			{
				GListColumn *c = Columns.ItemAt(d->DragData);
				if (c)
				{
					int OldWidth = c->Width();
					int NewWidth = m.x - c->d->Pos.x1;

					c->Width(max(NewWidth, 4));
				}
				break;
			}
			case CLICK_COLUMN:
			{
				bool Update = false;
				GListColumn *c = Columns.ItemAt(d->DragData);
				if (c)
				{
					if (abs(m.x - d->DragStart.x) > DRAG_THRESHOLD OR
						abs(m.y - d->DragStart.y) > DRAG_THRESHOLD)
					{
						OnColumnDrag(d->DragData, m);
					}
					else
					{
						bool Over = c->d->Pos.Overlap(m.x, m.y);
						if (m.Down() AND Over != c->d->Down)
						{
							c->d->Down = Over;
							Invalidate(&c->d->Pos);
						}
					}
				}
				break;
			}
			case SELECT_ITEMS:
			{
				int n=0;
				bool Selected = m.y < ItemsPos.y1;

				if (IsCapturing())
				{
					if (MultiItemSelect)
					{
						int Over = -1;
						
						GListItem *h = HitItem(m.x, m.y, &Over);

						/*
						if (m.y < ItemsPos.y1 AND FirstVisible == 0)
						{
							Over = 0;
						}
						else
						{
							Iterator<GListItem> Iter(&Items);
							int n = FirstVisible;
							for (GListItem *k=Iter[n]; k AND k->OnScreen(); k=Iter.Next())
							{
								if ((m.y >= k->Pos.y1) AND (m.y <= k->Pos.y2))
								{
									Over = n;
									break;
								}
								n++;
							}
						}
						*/

						if (Over >= 0)
						{
							n = 0;

							int Start = min(Over, d->DragData);
							int End = max(Over, d->DragData);

							ForAllItems(i)
							{
								i->Select(n >= Start AND n <= End);
								n++;
							}
						}
					}
					else
					{
						ForAllItems(i)
						{
							// i->Select(m.y >= i->Pos.y1 AND m.y <= i->Pos.y2);
							i->Select(i->Pos.Overlap(m.x, m.y));
						}
					}
				}
				break;
			}
			case CLICK_ITEM:
			{
				if (IsCapturing() AND
					abs(d->DragStart.x-m.x) > DRAG_THRESHOLD OR
					abs(d->DragStart.y-m.y) > DRAG_THRESHOLD)
				{
					OnItemBeginDrag(Items.ItemAt(d->DragData), m);
					d->DragMode = DRAG_NONE;
					Capture(false);
				}
				break;
			}
			default:
			{
				#ifndef WIN32
				GListColumn *Resize, *Over;
				HitColumn(m.x, m.y, Resize, Over);
				if (Resize)
				{
					SetCursor(LCUR_SizeHor);
				}
				#endif
				break;
			}
		}

		Unlock();
	}
}

int64 GList::Value()
{
	int n=0;
	ForAllItems(i)
	{
		if (i->Select())
		{
			return n;
		}
		n++;
	}
	return -1;
}

void GList::Value(int64 Index)
{
	int n=0;
	ForAllItems(i)
	{
		if (n == Index)
		{
			i->Select(true);
			Keyboard = n;
		}
		else
		{
			i->Select(false);
		}
		n++;
	}
}

void GList::SelectAll()
{
	if (Lock())
	{
		ForAllItems(i)
		{
			i->d->Selected = true;
		}

		Unlock();

		Invalidate();
	}
}

bool GList::Select(GListItem *Obj)
{
	bool Status = false;
	ForAllItems(i)
	{
		i->Select(Obj == i);
		if (Obj == i) Status = true;
	}
	return true;
}

GListItem *GList::GetSelection()
{
	GListItem *n = 0;

	if (Lock())
	{
		ForAllItems(i)
		{
			if (i->Select())
			{
				n = i;
				break;
			}
		}

		Unlock();
	}

	return n;
}

bool GList::GetSelection(List<GListItem> &l)
{
	ForAllItems(i)
	{
		if (i->Select())
		{
			l.Insert(i);
		}
	}

	return l.Length() > 0;
}

bool GList::GetAll(List<GListItem> &l)
{
	ForAllItems(i)
	{
		l.Insert(i);
	}

	return true;
}

bool GList::GetUpdateRegion(GListItem *i, GRegion &r)
{
	r.Empty();

	if (d->Mode == GListDetails)
	{
		if (i->Pos.Valid())
		{
			GRect u = i->Pos;
			u.y2 = ItemsPos.y2;
			r.Union(&u);
			
			return true;
		}
	}
	else if (d->Mode == GListColumns)
	{
		if (i->Pos.Valid())
		{
			GRect u = i->Pos;
			u.y2 = ItemsPos.y2;
			r.Union(&u);
			
			u.x1 = u.x2 + 1;
			u.y1 = ItemsPos.y1;
			r.Union(&u);
			
			return true;
		}
	}

	return false;
}

bool GList::Insert(GListItem *i, int Index, bool Update)
{
	bool Status = false;

	if (i AND Lock())
	{
		// Insert
		if (i->Parent != this)
		{
			bool First = Items.Length() == 0;
			
			i->Parent = this;
			i->Select(false);
			Items.Insert(i, Index);
			i->OnInsert();

			if (First)
			{
				Keyboard = 0;
				i->Select(true);
			}
			
			if (Update)
			{
				// Update screen
				Pour();
				GRegion Up;
				if (GetUpdateRegion(i, Up))
				{
					Up.y2 = Y();
					Invalidate(&Up);
				}

				// Notify
				SendNotify(GLIST_NOTIFY_INSERT);
			}
		}

		Status = true;

		Unlock();
	}

	return Status;
}

bool GList::Insert(List<GListItem> &l, int Index, bool Update)
{
	bool Status = false;

	if (Lock())
	{
		bool First = Items.Length() == 0;
		
		// Insert list of items
		for (GListItem *i=l.First(); i; i=l.Next())
		{
			if (i->Parent != this)
			{
				i->Parent = this;
				i->Select(false);
				Items.Insert(i, Index);
				i->OnInsert();
				if (Index >= 0) Index++;
				
				if (First)
				{
					First = false;
					Keyboard = 0;
					i->Select(true);
				}
			}
		}

		if (Update)
		{
			// Update screen
			Pour();
			Invalidate();

			// Notify
			SendNotify(GLIST_NOTIFY_INSERT);
		}

		Status = true;

		Unlock();
	}

	return Status;
}

bool GList::Delete()
{
	return Delete(Items.Current());
}

bool GList::Delete(int Index)
{
	return Delete(Items.ItemAt(Index));
}

bool GList::Delete(GListItem *i)
{
	bool Status = false;
	if (Lock())
	{
		if (Remove(i))
		{
			// Delete
			DeleteObj(i);
			Status = true;
		}
		Unlock();
	}

	return Status;
}

bool GList::Remove(GListItem *i)
{
	bool Status = false;

	if (Lock())
	{
		if (i AND i->GetList() == this)
		{
			GRegion Up;
			bool Visible = GetUpdateRegion(i, Up);
			bool Selected = i->Select();
			int Index = Items.IndexOf(i);
			int64 Pos = (VScroll) ? VScroll->Value() : 0;

			// Remove from list
			Items.Delete(i);
			i->OnRemove();
			i->Parent = 0;
			UpdateScrollBars();

			// Update screen
			if ((VScroll AND VScroll->Value() != Pos) OR
				Index < FirstVisible)
			{
				Invalidate(&ItemsPos);
			}
			else if (Visible)
			{
				Up.y2 = ItemsPos.y2;
				Invalidate(&Up);
			}

			// Notify
			GViewI *Note = (GetNotify())?GetNotify():GetParent();
			if (Note)
			{
				if (Selected)
				{
					OnItemSelect(0);
				}

				Note->OnNotify(this, GLIST_NOTIFY_DELETE);
			}

			Status = true;
		}

		Unlock();
	}

	return Status;
}

bool GList::HasItem(GListItem *Obj)
{
	return Items.HasItem(Obj);
}

int GList::IndexOf(GListItem *Obj)
{
	return Items.IndexOf(Obj);
}

GListItem *GList::ItemAt(int Index)
{
	return Items.ItemAt(Index);
}

void GList::ScrollToSelection()
{
	if (VScroll)
	{
		int n=0;
		int Vis = VisibleItems();
		ForAllItems(i)
		{
			if (i->Select())
			{
				if (n < FirstVisible OR n > LastVisible)
				{
					int k = n - (Vis/2);
					VScroll->Value(max(k, 0));
					Invalidate(&ItemsPos);
					break;
				}
			}
			n++;
		}
	}
}

void GList::Sort(GListCompareFunc Compare, int Data)
{
	if (Lock())
	{
		if (Compare)
		{
			Items.Sort(Compare, Data);
			Invalidate(&ItemsPos);
		}

		Unlock();
	}
}

void GList::Empty()
{
	if (Lock())
	{
		ForAllItems(i)
		{
			LgiAssert(i->Parent == this);
			i->Parent = 0;
			DeleteObj(i);
		}
		Items.Empty();
		
		FirstVisible = LastVisible = -1;

		if (VScroll)
		{
			VScroll->Value(0);
			VScroll->SetLimits(0, -1);
		}

		Invalidate();
		DeleteArray(d->KeyBuf);

		Unlock();
	}
}

void GList::RemoveAll()
{
	if (Lock())
	{
		for (GListItem *i = Items.First(); i; i = Items.Next())
		{
			i->OnRemove();
			i->Parent = 0;
		}
		Items.Empty();
		FirstVisible = LastVisible = -1;

		if (VScroll)
		{
			// these have to be in this order because
			// "SetLimits" can cause the VScroll object to
			// be deleted and becoming NULL
			VScroll->Value(0);
			VScroll->SetLimits(0, -1);
		}

		Invalidate();
		DeleteArray(d->KeyBuf);
		Unlock();
	}
}

void GList::OnPosChange()
{
	GLayout::OnPosChange();
}

void GList::UpdateScrollBars()
{
	if (VScroll)
	{
		int Vis = VisibleItems();
		int Max = MaxScroll();

		if (VScroll->Value() > max(Max, 0))
		{
			VScroll->Value(Max);
		}

		VScroll->SetPage(Vis);
		VScroll->SetLimits(0, Items.Length() - 1);
	}
	
	if (HScroll)
	{
		HScroll->SetPage(d->VisibleColumns);
		HScroll->SetLimits(0, d->Columns - 1);
	}
}

void GList::Pour()
{
	#if GLIST_POUR_PROFILE
	int Start = LgiCurrentTime();
	#endif

	// Layout all the elements
	GRect Client = GetClient();

	if (d->Mode == GListDetails)
	{
		if (ColumnHeaders)
		{
			ColumnHeader = Client;
			ColumnHeader.y2 = ColumnHeader.y1 + SysFont->Y("A") + 4;
			ItemsPos = Client;
			ItemsPos.y1 = ColumnHeader.y2 + 1;
		}
		else
		{
			ItemsPos = Client;
		}
		
		int n = 0;
		int y = ItemsPos.y1;
		int Max = MaxScroll();
		FirstVisible = (VScroll) ? VScroll->Value() : 0;
		if (FirstVisible < 0) FirstVisible = 0;
		if (FirstVisible > Max) FirstVisible = Max;
		LastVisible = 0x7FFFFFFF;
		CompletelyVisible = 0;
		bool SomeHidden = false;

		ForAllItems(i)
		{
			if (n < FirstVisible OR n > LastVisible)
			{
				i->Pos.Set(-1, -1, -2, -2);
				SomeHidden = true;
			}
			else
			{
				GMeasureInfo Info;
				
				i->OnMeasure(&Info);
				i->Pos.Set(ItemsPos.x1, y, ItemsPos.x2, y+Info.y-1);
				y = y+Info.y;

				if (i->Pos.y2 > ItemsPos.y2)
				{
					LastVisible = n;
					SomeHidden = true;
				}
				else
				{
					CompletelyVisible++;
				}
			}

			n++;
		}

		if (LastVisible >= Items.Length())
		{
			LastVisible = Items.Length() - 1;
		}

		SetScrollBars(false, SomeHidden);
		UpdateScrollBars();
	}
	else if (d->Mode == GListColumns)
	{
		ColumnHeader.ZOff(-1, -1);
		ItemsPos = Client;
		FirstVisible = 0;

		int CurX = 0;
		int CurY = 0;
		int MaxX = 16;
		GArray<GListItem*> Col;
		d->Columns = 1;
		d->VisibleColumns = 0;
		int64 ScrollX = HScroll ? HScroll->Value() : 0;
		int64 OffsetY = HScroll ? 0 : GScrollBar::GetScrollSize();
		FirstVisible = -1;
		
		int n = 0;
		ForAllItems(i)
		{
			GMeasureInfo Info;
			i->OnMeasure(&Info);

			if (d->Columns <= ScrollX OR
				CurX > ItemsPos.X())
			{
				i->Pos.ZOff(-1, -1);
				i->d->LayoutColumn = d->Columns;
				
				if (ItemsPos.y1 + CurY + Info.y > ItemsPos.y2 - OffsetY)
				{
					CurY = 0;
					d->Columns++;
					
					if (d->Columns > ScrollX AND
						CurX < ItemsPos.X())
					{
						goto FlowItem;
					}
				}
			}
			else
			{
				FlowItem:
				if (ItemsPos.y1 + CurY + Info.y > ItemsPos.y2 - OffsetY)
				{
					// wrap to next column
					for (int n=0; n<Col.Length(); n++)
					{
						Col[n]->Pos.x2 = CurX + MaxX - 1;
					}
					Col.Length(0);
					
					CurX += MaxX;
					CurY = 0;
					d->Columns++;
					if (CurX < ItemsPos.X())
					{
						d->VisibleColumns++;
					}
				}
				
				if (FirstVisible < 0)
					FirstVisible = n;
				LastVisible = n;
				
				i->d->LayoutColumn = d->Columns;
				i->Pos.ZOff(Info.x-1, Info.y-1);
				i->Pos.Offset(ItemsPos.x1 + CurX, ItemsPos.y1 + CurY);
				Col[Col.Length()] = i;
				MaxX = max(MaxX, Info.x);
				
				CompletelyVisible++;
			}

			CurY += Info.y;
			n++;
		}
		
		d->VisibleColumns = max(1, d->VisibleColumns);

		// pour remaining items...
		for (n=0; n<Col.Length(); n++)
		{
			Col[n]->Pos.x2 = CurX + MaxX - 1;
		}
		Col.Length(0);
		if (CurX + MaxX < ItemsPos.X())
		{
			d->VisibleColumns++;
		}

		// printf("%u - ScrollX=%i VisCol=%i Cols=%i\n", (uint32)LgiCurrentTime(), ScrollX, d->VisibleColumns, d->Columns);
		SetScrollBars(d->VisibleColumns < d->Columns, false);
		UpdateScrollBars();
	}
	
	#if GLIST_POUR_PROFILE
	printf("GList::Pour() %i ms\n", LgiCurrentTime() - Start);
	#endif
}

void GList::OnPaint(GSurface *pDC)
{
	#if GLIST_ONPAINT_PROFILE
	int Start = LgiCurrentTime();
	#endif
	
	if (Lock())
	{
		COLOUR Back = Enabled() ? LC_WORKSPACE : LC_MED;
		Pour();

		// Check icon column status then draw
		if (AskImage() AND NOT IconCol)
		{
			IconCol = NEW(GListColumn(this, 0, 18));
			if (IconCol)
			{
				IconCol->Resizable(false);
				IconCol->Type(GIC_ASK_IMAGE);
			}
		}
		else if (NOT AskImage() AND IconCol) DeleteObj(IconCol);

		// Draw column headings
		if (ColumnHeaders AND ColumnHeader.Valid())
		{
			// Draw columns
			GRect cr = ColumnHeader;
			int cx = cr.x1;
			pDC->ClipRgn(&cr);

			if (IconCol)
			{
				cr.x1 = cx;
				cr.x2 = cr.x1 + IconCol->Width() - 1;
				IconCol->d->Pos = cr;
				IconCol->OnPaint(pDC, cr);
				cx += IconCol->Width();
			}

			// Draw other columns
			for (GListColumn *c = Columns.First(); c; c = Columns.Next())
			{
				cr.x1 = cx;
				cr.x2 = cr.x1 + c->Width() - 1;
				c->d->Pos = cr;
				c->OnPaint(pDC, cr);
				cx += c->Width();
			}

			// Draw ending peice
			cr.x1 = cx;
			cr.x2 = ColumnHeader.x2 + 2;

			if (cr.Valid())
			{
				if (GApp::SkinEngine)
				{
					GSkinState State;
					State.pScreen = pDC;
					State.Rect = cr;
					State.Enabled = Enabled();
					GApp::SkinEngine->OnPaint_ListColumn(0, 0, &State);
				}
				else
				{
					LgiWideBorder(pDC, cr, RAISED);
					pDC->Colour(LC_MED, 24);
					pDC->Rectangle(&cr);
				}
			}

			pDC->ClipRgn(0);
		}

		// Draw items
		if (NOT Buf)
		{
			Buf = NEW(GMemDC);
		}

		GRect r = ItemsPos;
		int n = FirstVisible;
		int LastY = r.y1;
		if (NOT Focus())
		{
			_Selection = GdcMixColour(Back, LC_SELECTION);
		}
		else
		{
			_Selection = LC_SELECTION;
		}

		GRegion Rgn(ItemsPos);
		for (GListItem *i = Items.ItemAt(n); i; i = Items.Next(), n++)
		{
			if (i->Pos.Valid())
			{
				// tell the item what colour to use
				if (i->Select())
				{
					i->Fore = LC_SEL_TEXT;
					i->Back = _Selection;
				}
				else
				{
					i->Fore = LC_TEXT;
					i->Back = Back;
				}

				#if DOUBLE_BUFFER_PAINT
				if (Buf->X() < i->Pos.X() OR
					Buf->Y() < i->Pos.Y())
				{
					Buf->Create(i->Pos.X(), i->Pos.Y(), GdcD->GetBits());
				}

				GRect r = i->Pos;
				r.Offset(-r.x1, -r.y1);
				i->OnPaint(Buf, r);
				pDC->Blt(i->Pos.x1, i->Pos.y1, Buf, &r);
				#else
				i->OnPaint(pDC, i->Pos);
				#endif
				Rgn.Subtract(&i->Pos);

				LastY = i->Pos.y2 + 1;
			}
			else
			{
				break;
			}
		}
		
		pDC->Colour(Back, 24);
		for (GRect *w=Rgn.First(); w; w=Rgn.Next())
		{
			pDC->Rectangle(w);
		}

		Unlock();
	}

	#if GLIST_ONPAINT_PROFILE
	printf("GList::OnPaint()=%i ms\n", LgiCurrentTime() - Start);
	#endif
}

void GList::OnFocus(bool b)
{
	GListItem *s = GetSelection();
	if (NOT s)
	{
		s = Items.First();
		if (s)
		{
			s->Select(true);
		}
	}

	for (GListItem *i = Items.ItemAt(FirstVisible); i; i = Items.Next())
	{
		if (i->Pos.Valid() AND
			i->d->Selected)
		{
			Invalidate(&i->Pos);
		}
	}

	GLayout::OnFocus(b);
	
	if (NOT b AND IsCapturing())
	{
		Capture(false);
	}
}

void GList::UpdateAllItems()
{
	ForAllItems(i)
	{
		i->d->EmptyDisplay();
	}
	
	Invalidate();
}

void GList::OnColumnClick(int Col, GMouse &m)
{
	d->ColClick = Col;
	d->ColMouse = m;
	SendNotify(GLIST_NOTIFY_COLS_CLICK);
}

bool GList::GetColumnClickInfo(int &Col, GMouse &m)
{
	if (d->ColClick >= 0)
	{
		Col = d->ColClick;
		m = d->ColMouse;
		return true;
	}
	
	return false;	
}

void GList::ResizeColumnsToContent(int Border)
{
	for (GListColumn *c=Columns.First(); c; c=Columns.Next())
	{
		if (c->Resizable())
		{
			c->Width(c->GetContentSize() + Border);
		}
	}
}
