#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <ctype.h>

#include "Lgi.h"
#include "GProcess.h"
#include "GSkinEngine.h"
#include "GArray.h"
#include "LgiWinManGlue.h"

////////////////////////////////////////////////////////////////

#if HAS_SHARED_MIME

#include "GFilterUtils.h"
#include "mime-types.h"
class GSharedMime : public GLibrary
{
public:
	GSharedMime() :
		#ifdef _DEBUG
		GLibrary("libsharedmime1d")
		#else
		GLibrary("libsharedmime1")
		#endif
	{
	}
	
	DynFunc0(int, mimetypes_init);
	DynFunc1(const char*, mimetypes_set_default_type, const char *, default_type);	
	DynFunc2(const char*, mimetypes_get_file_type, const char*, pathname, mimetypes_flags, flags);
	DynFunc2(const char*, mimetypes_get_data_type, const void*, data, int, length);
	DynFunc3(bool, mimetypes_decode, const char *, type, char **, media_type, char **, sub_type);
	DynFunc2(char *, mimetypes_convert_filename, const char *, pathname, const char *, mime_type);
	DynFunc3(bool, mimetypes_add_mime_dir, const char *, path, bool, high_priority, bool, rescan);
	DynFunc2(const char *, mimetypes_get_type_info, const char *, mime_type, const char *, lang);
};

#endif

#ifdef XWIN
/////////////////////////////////////////////////////////////////////////////
//
// Attempts to cleanup and call drkonqi to process the crash
//
void LgiCrashHandler(int Sig)
{
	// Don't get into an infinite loop
   	signal(SIGSEGV, SIG_DFL);

	// Our pid
	int MyPid = getpid();	
	printf("LgiCrashHandler trigger MyPid=%i\n", MyPid);

	// Fork to run the crash handler
	pid_t Pid = fork();
	if (Pid <= 0)
	{
		// Pass our state down to the crash handler...
		XObject o;
		int Args = 0;
		char *Arg[32];
		char SigName[16], PidName[16];
		ZeroObj(Arg);

		sprintf(SigName, "%i", Sig);	
		sprintf(PidName, "%i", MyPid);

		Arg[Args++] = "drkonqi";
		Arg[Args++] = "--display";
		Arg[Args++] = XDisplayString(o.XDisplay());
		Arg[Args++] = "--appname";
        Arg[Args++] = LgiApp->AppWnd ? LgiApp->AppWnd->Name() : (char*)"Lgi";
        Arg[Args++] = "--signal";
        Arg[Args++] = SigName;
        Arg[Args++] = "--pid";
        Arg[Args++] = PidName;
        
		setgid(getgid());
		setuid(getuid());

		execvp("drkonqi", Arg);
	}
	else
	{
		// Wait for child to exit
		waitpid(Pid, NULL, 0);
		_exit(253);
	}
}
#endif

/////////////////////////////////////////////////////////////////////////////
typedef GArray<GAppInfo*> AppArray;

class GAppPrivate
{
public:
	// Common
	GXmlTag *Config;
	GFileSystem *FileSystem;
	GdcDevice *GdcSystem;
	OsAppArguments Args;
	GLibrary *SkinLib;
	GHashTable MimeToApp;
	#if HAS_SHARED_MIME
	GSharedMime *Sm;
	#endif
	GLibrary *WmLib;

	GAppPrivate()
	{
		WmLib = 0;
		FileSystem = 0;
		GdcSystem = 0;
		Config = 0;
		SkinLib = 0;
		#if HAS_SHARED_MIME
		Sm = 0;
		#endif
	}

	~GAppPrivate()
	{
		DeleteObj(WmLib);
		DeleteObj(SkinLib);
		#if HAS_SHARED_MIME
		DeleteObj(Sm);
		#endif
		
		for (void *p = MimeToApp.First(); p; p = MimeToApp.Next())
		{
			AppArray *a = (AppArray*)p;
			a->DeleteObjects();
			DeleteObj(a);
		}
	}
};
/////////////////////////////////////////////////////////////////////////////
GSkinEngine *GApp::SkinEngine = 0;
GApp *TheApp = 0;
GMouseHook *GApp::MouseHook = 0;

GApp::GApp(char *AppMime, OsAppArguments &AppArgs, GAppArguments *Args) :
	OsApplication(AppArgs.Args, AppArgs.Arg)
{
	TheApp = this;
	d = NEW(GAppPrivate);
	
	// We want our printf's NOW!
	setvbuf(stdout,(char *)NULL,_IONBF,0); // print mesgs immediately.

	#if defined XWIN
	// Check for X11
	if (NOT XDisplay())
	{
		return;
	}
	#elif defined YWIN
	#endif

	// Setup the file and graphics sub-systems
	d->FileSystem = NEW(GFileSystem);
	d->GdcSystem = NEW(GdcDevice);

	srand(LgiCurrentTime());
	LgiInitColours();
	AppWnd = 0;

	SetAppArgs(AppArgs);
	MouseHook = NEW(GMouseHook);

	#if defined XWIN
	// Setup the SIGSEGV signal to call the KDE crash handler
	if (NOT GetOption("nch") AND
		LgiGetWindowManager() == WM_Kde)
	{
		signal(SIGSEGV, LgiCrashHandler);
	}
	#endif
	
	// System font setup
	SystemNormal = 0;
	GFontType SysFontType;

	if (SysFontType.GetSystemFont("System"))
	{
		SystemNormal = SysFontType.Create();

		if (SystemNormal)
		{
			SystemNormal->Transparent(true);
		}
		
		SystemBold = SysFontType.Create();
		if (SystemBold)
		{
			SystemBold->Bold(true);
			SystemBold->Transparent(true);
			SystemBold->Create();
		}
	}
	else
	{
		printf("%s:%i - Couldn't get system font setting.\n", __FILE__, __LINE__);
	}

	if (NOT SystemNormal)
	{
		LgiMsg(0, "Error: Couldn't create system font.", "Lgi Error: GApp::GApp", MB_OK);
		LgiExitApp();
	}
	
	if (NOT GetOption("noskin"))
	{
		// Load library
		char Name[64];
		d->SkinLib = NEW(GLibrary("liblgiskin"
			#ifdef _DEBUG
			"d"
			#endif
			));
		if (d->SkinLib)
		{
			if (d->SkinLib->IsLoaded())
			{
				Proc_CreateSkinEngine CreateSkinEngine =
					(Proc_CreateSkinEngine)d->SkinLib->GetAddress(LgiSkinEntryPoint);
				if (CreateSkinEngine)
				{
					SkinEngine = CreateSkinEngine(this);
				}
			}
			else
			{
				DeleteObj(d->SkinLib);
			}
		}
	}
}

GApp::~GApp()
{
	DeleteObj(AppWnd);
	DeleteObj(SystemNormal);
	DeleteObj(SystemBold);
	
	DeleteObj(MouseHook);
	DeleteObj(d->FileSystem);
	DeleteObj(d->GdcSystem);
	DeleteObj(d->Config);
	DeleteObj(GFontSystem::Me);
	DeleteObj(d);
	TheApp = 0;
}

GApp *GApp::ObjInstance()
{
	return TheApp;
}

bool GApp::IsOk()
{
	bool Status = 	(this != 0) AND
					(d != 0)
					/*
					#ifdef XWIN
					AND (XDisplay() != 0)
					#endif
					*/
					;
					
	LgiAssert(Status);
	return Status;
}

GMouseHook *GApp::GetMouseHook()
{
	return MouseHook;
}

int GApp::GetMetric(int Metric)
{
	switch (Metric)
	{
		case LGI_MET_DECOR_X:
		{
			return 12;
		}
		case LGI_MET_DECOR_Y:
		{
			return 26;
		}
	}

	return 0;
}

GViewI *GApp::GetFocus()
{
	#if defined XWIN
	XWidget *w = XObject::XApp()->GetFocus();
	if (w)
	{
		return (GView*)w->_GetWndPtr();
	}
	#endif
	
	return 0;
}

OsThreadId GApp::GetGuiThread()
{
	#if defined XWIN
	return XApplication::GetGuiThread();
	#endif
	
	return 0;
}

OsProcessId GApp::GetProcessId()
{
	return getpid();
}

OsAppArguments *GApp::GetAppArgs()
{
	return IsOk() ? &d->Args : 0;
}

void GApp::SetAppArgs(OsAppArguments &AppArgs)
{
	if (IsOk())
	{
		d->Args = AppArgs;
	}
}

bool GApp::Run(bool Loop)
{
	if (Loop)
	{
		OnCommandLine();
		#if defined XWIN
		XApplication::exec();
		#elif defined YWIN
		OsApplication::Run();
		#endif
	}
	else
	{
		#if defined XWIN
		processEvents();
		#endif
	}


	return 0;
}

void GApp::OnReceiveFiles(GArray<char*> &Files)
{
	if (AppWnd)
	{
		AppWnd->OnReceiveFiles(Files);
	}
}

void GApp::SetConfig(GXmlTag *Tag)
{
	if (IsOk() AND Tag)
	{
		GXmlTag *Old = GetConfig(Tag->Tag);
		if (Old)
		{
			Old->RemoveTag();
			DeleteObj(Old);
		}

		if (NOT d->Config)
		{
			d->Config = NEW(GXmlTag("Config"));
		}
		if (d->Config)
		{
			d->Config->InsertTag(Tag);
		}
	}
}

GXmlTag *GApp::GetConfig(char *Tag)
{
	if (IsOk() AND NOT d->Config)
	{
		char File[] = "lgi.conf";
		char Path[256];
		if (LgiGetExePath(Path, sizeof(Path)))
		{
			if (Path[strlen(Path)-1] != DIR_CHAR) strcat(Path, DIR_STR);
			strcat(Path, File);

			if (NOT FileExists(Path))
			{
				char *f = LgiFindFile(File);
				if (f)
				{
					strcpy(Path, f);
					DeleteArray(f);
				}
			}

			if (FileExists(Path))
			{
				d->Config = NEW(GXmlTag("Config"));
				if (d->Config)
				{
					GFile f;
					if (f.Open(Path, O_READ))
					{
						GXmlTree t;
						t.Read(d->Config, &f, 0);
					}
				}
			}
		}
	}

	if (Tag AND d->Config)
	{
		return d->Config->GetTag(Tag);
	}

	return 0;
}

char *GApp::GetArgumentAt(int n)
{
	return n >= 0 AND n < d->Args.Args ? NewStr(d->Args.Arg[n]) : 0;
}

bool GApp::GetOption(char *Option, char *Dest, int DestLen)
{
	GArray<char> Buf;
	if (GetOption(Option, Buf))
	{
		if (Dest)
		{
			if (DestLen > 0)
			{
				strsafecpy(Dest, &Buf[0], DestLen);
			}
			else return false;
		}
		return true;
	}
	return false;
}

bool GApp::GetOption(char *Option, GArray<char> &Buf)
{
	if (IsOk() AND Option)
	{
		int OptLen = strlen(Option);
		for (int i=1; i<d->Args.Args; i++)
		{
			char *a = d->Args.Arg[i];

			if (strchr("-/\\", a[0]))
			{
				if (strncmp(a+1, Option, OptLen) == 0)
				{
					char *Arg = 0;
					if (strlen(a+1+OptLen) > 0)
					{
						Arg = a + 1 + OptLen;
					}
					else if (i < d->Args.Args - 1)
					{
						Arg = d->Args.Arg[i + 1];
					}
					
					if (Arg)
					{
						if (strchr("\'\"", *Arg))
						{
							char Delim = *Arg++;
							char *End = strchr(Arg, Delim);
							if (End)
							{
								int Len = (int)End-(int)Arg;
								if (Len > 0)
								{
									Buf.Length(Len + 1);
									memcpy(&Buf[0], Arg, Len);
									Buf[Len] = 0;
								}
								else return false;
							}
							else return false;
						}
						else
						{
							Buf.Length(strlen(Arg)+1);
							strcpy(&Buf[0], Arg);
						}
					}

					return true;
				}
			}
		}
	}

	return false;
}

void GApp::OnCommandLine()
{
	char WhiteSpace[] = " \r\n\t";

	GArray<char*> Files;

	for (int i=1; i<GetAppArgs()->Args; i++)
	{
		char *a = GetAppArgs()->Arg[i];
		if (FileExists(a))
		{
			Files.Add(NewStr(a));
		}
	}

	// call app
	if (Files.Length() > 0)
	{
		OnReceiveFiles(Files);
	}

	// clear up
	Files.DeleteArrays();
}

void GApp::Exit(int Code)
{
	// printf("\nGApp::Exit(%i) called.\n", Code);

	if (Code)
	{
		// hard exit
		::exit(Code);
	}
	else
	{
		// soft exit
		#if defined XWIN
		XApplication::exit(0);
		#endif
	}
}


bool GApp::GetFileMimeType(char *File, char *Mime, int BufLen)
{
	bool Status = false;
	char Full[300] = "";

	if (NOT FileExists(File))
	{
		// Look in the path
		GToken p(getenv("PATH"), LGI_PATH_SEPARATOR);
		for (int i=0; i<p.Length(); i++)
		{
			LgiMakePath(Full, sizeof(Full), p[i], File);
			if (FileExists(Full))
			{
				File = Full;
				break;
			}
		}		
	}

	#if HAS_SHARED_MIME
	// freedesktop.org rocks...
	if (NOT d->Sm)
	{
		// Not loaded, go and try to load it...
		d->Sm = NEW(GSharedMime);
		if (d->Sm AND d->Sm->IsLoaded())
		{
			d->Sm->mimetypes_init();
		}
	}
	if (d->Sm AND d->Sm->IsLoaded())
	{
		// Loaded...
		char *m = (char*)d->Sm->mimetypes_get_file_type(File, MIMETYPES_CHECK_ALL);
		if (m)
		{
			#if HAS_FILE_CMD
			if (stricmp(m, "application/x-not-recognised") != 0)
			#endif
			{
				strcpy(Mime, m);
				return true;
			}
		}
		else
		{
			printf("%s:%i - mimetypes_get_file_type failed for '%s'\n", __FILE__, __LINE__, File);
		}
	}
	else
	{
		printf("%s:%i - Shared Mime not loaded!!!\n", __FILE__, __LINE__);
	}
	#endif

	#if HAS_FILE_CMD
	// doh! not installed... :(
	GStringPipe Output;
	char Args[256];
	sprintf(Args, "-i \"%s\"", File);
	GProcess p;
	if (p.Run("file", Args, 0, true, 0, &Output))
	{
		char *Out = Output.NewStr();
		if (Out)
		{
			char *s = strchr(Out, ':');
			if (s AND strchr(Out, '/'))
			{
				s += 2;
				
				char *e = s;
				while (*e AND (isalpha(*e) OR strchr("-_/", *e))) e++;
				*e = 0;
				
				strcpy(Mime, s);
				Status = true;
			}
			DeleteArray(Out);
		}
	}
	#endif
	
	return Status;
}

bool GApp::GetAppsForMimeType(char *Mime, GArray<GAppInfo*> &Apps)
{
	// Find alternative version of the MIME type (e.g. x-type and type).
	char AltMime[256];
	strcpy(AltMime, Mime);
	char *s = strchr(AltMime, '/');
	if (s)
	{
		s++;
		
		int Len = strlen(s) + 1;
		if (strnicmp(s, "x-", 2) == 0)
		{
			memmove(s, s+2, Len - 2);
		}
		else
		{
			memmove(s+2, s, Len);
			s[0] = 'x';
			s[1] = '-';
		}
	}

	LgiGetAppsForMimeType(Mime, Apps);
	LgiGetAppsForMimeType(AltMime, Apps);
	
	return Apps.Length() > 0;

	/*
	if (NOT d->MimeToApp.Length())
	{
		// printf("%s:%i - Building MimeToApp.\n", __FILE__, __LINE__);

		GToken Paths;
		if (_GetKdePaths(Paths, "apps"))
		{
			// No default... look in the apps dir
			List<char> f, e;
			e.Insert("*.desktop");
			for (int i=0; i<Paths.Length(); i++)
			{
				LgiRecursiveFileSearch(Paths[i], &e, &f);
			}

			for (char *File=f.First(); File; File=f.Next())
			{
				if (stristr(File, "kde-cervisia"))
				{
					continue;
				}
				
				char *DesktopFile = ReadTextFile(File);
				if (DesktopFile)
				{
					char MimeTypes[512];
					char AppPath[256] = "";

					// this is an app that supports the MimeTypes...
					if (_GetIniField("Desktop Entry", "MimeType", DesktopFile, MimeTypes, sizeof(MimeTypes)) AND
						_GetIniField("Desktop Entry", "Exec", DesktopFile, AppPath, sizeof(AppPath)))
					{
						// look through the types..
						GToken t(MimeTypes, ";");
						for (int i=0; i<t.Length(); i++)
						{
							// Fliter out invalid mime types
							if (strchr(t[i], '/'))
							{
								// Check for existing app array
								AppArray *Group = (AppArray*) d->MimeToApp.Find(t[i]);
								if (NOT Group)
								{
									// Not there so create it...
									d->MimeToApp.Add(t[i], Group = NEW(AppArray));
								}									
								if (Group)
								{
									// Check the application is not in the array already
									bool Has = false;
									for (int n=0; n<Group->Length(); n++)
									{
										GAppInfo *Existing = (*Group)[n];
										if (Existing AND Existing->Path)
										{
											char *a = strchr(AppPath, ' ');
											int alen = a ? (int)a-(int)AppPath : strlen(AppPath);

											char *b = strchr(Existing->Path, ' ');
											int blen = b ? (int)b-(int)Existing->Path : strlen(Existing->Path);

											if (alen == blen AND
												strncmp(AppPath, Existing->Path, alen) == 0)
											{
												Has = true;
												break;
											}
										}
									}
									
									if (NOT Has)
									{
										// It's not, so add it...
										GAppInfo *a = NEW(GAppInfo);
										if (a)
										{
											(*Group)[Group->Length()] = a;											
											a->Path = NewStr(AppPath);
											
											char s[256];
											
											// store the user visible name
											if (_GetIniField("Desktop Entry", "Name", DesktopFile, s, sizeof(s)))
											{
												a->Name = NewStr(s);
											}
											
											// store the icon file...
											if (_GetIniField("Desktop Entry", "Icon", DesktopFile, s, sizeof(s)))
											{
												a->Icon = NewStr(s);
											}
										}
									}
								}
							}
						}
					}

					DeleteArray(DesktopFile);
				}
			}
		}
		else
		{
			printf("LgiGetAppForMimeType: Couldn't get apps path(s)\n");
		}
	}
	
	AppArray *p = (AppArray*)d->MimeToApp.Find(Mime);
	if (p)
	{
		for (int i=0; i<p->Length(); i++)
		{
			Apps[i] = (*p)[i];
		}
		
		return true;
	}
	*/
}

GLibrary *GApp::GetWindowManagerLib()
{
	if (NOT d->WmLib)
	{
		WindowManager Wm = LgiGetWindowManager();
		switch (Wm)
		{
			case WM_Kde:
				d->WmLib = NEW(GLibrary("liblgikde3"
					#ifdef _DEBUG
					"d"
					#endif
					));
				break;
			case WM_Gnome:
				d->WmLib = NEW(GLibrary("liblgignome2"
					#ifdef _DEBUG
					"d"
					#endif
					));
				break;
			default:
				d->WmLib = NEW(GLibrary("liblgiother"
					#ifdef _DEBUG
					"d"
					#endif
					));
				break;
		}
		
		if (d->WmLib)
		{
			if (d->WmLib->IsLoaded())
			{
				Proc_LgiWmInit WmInit = (Proc_LgiWmInit) d->WmLib->GetAddress("LgiWmInit");
				if (WmInit)
				{
					WmInitParams Params;
					Params.Dsp = XObject::XDisplay();
					Params.Args = d->Args.Args;
					Params.Arg = d->Args.Arg;
					WmInit(&Params);
				}
				else
				{
					printf("%s:%i - Failed to find method 'LgiWmInit' in WmLib.\n", __FILE__, __LINE__);
				}
			}
		}
	}
	
	return d->WmLib AND d->WmLib->IsLoaded() ? d->WmLib : 0;
}

