#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <ctype.h>

#include "Lgi.h"
#include "GProcess.h"
#include "GSkinEngine.h"
#include "GArray.h"
#include "GToken.h"

////////////////////////////////////////////////////////////////
typedef GArray<GAppInfo*> AppArray;

void OnSigPipe(int i)
{
}

class GAppPrivate
{
public:
	// Common
	GXmlTag *Config;
	GFileSystem *FileSystem;
	GdcDevice *GdcSystem;
	OsAppArguments Args;
	GLibrary *SkinLib;
	GHashTable MimeToApp;

	GAppPrivate()
	{
		FileSystem = 0;
		GdcSystem = 0;
		Config = 0;
		SkinLib = 0;
	}

	~GAppPrivate()
	{
		DeleteObj(SkinLib);
		
		for (void *p = MimeToApp.First(); p; p = MimeToApp.Next())
		{
			AppArray *a = (AppArray*)p;
			a->DeleteObjects();
			DeleteObj(a);
		}
	}
};
/////////////////////////////////////////////////////////////////////////////
GSkinEngine *GApp::SkinEngine = 0;
GApp *TheApp = 0;
GMouseHook *GApp::MouseHook = 0;

GApp::GApp(char *AppMime, OsAppArguments &AppArgs, GAppArguments *ObjArgs) :
	OsApplication(AppArgs.Args, AppArgs.Arg)
{
	TheApp = this;
	d = NEW(GAppPrivate);

	// Catch and ignore SIGPIPE
	signal(SIGPIPE, OnSigPipe);

	// We want our printf's NOW!
	setvbuf(stdout,(char *)NULL,_IONBF,0); // print mesgs immediately.

	#if defined XWIN
	// Check for X11
	if (NOT XDisplay())
	{
		return;
	}
	#elif defined YWIN
	#endif

	// Setup the file and graphics sub-systems
	d->FileSystem = NEW(GFileSystem);
	d->GdcSystem = NEW(GdcDevice);

	srand(LgiCurrentTime());
	LgiInitColours();
	AppWnd = 0;

	SetAppArgs(AppArgs);
	MouseHook = NEW(GMouseHook);

	#if defined XWIN
	// Setup the SIGSEGV signal to call the KDE crash handler
	if (NOT GetOption("nch") AND
		LgiGetWindowManager() == WM_Kde)
	{
		signal(SIGSEGV, LgiCrashHandler);
	}
	#endif
	
	// System font setup
	SystemNormal = 0;
	GFontType SysFontType;

	if (SysFontType.GetSystemFont("System"))
	{
		SystemNormal = SysFontType.Create();
		if (SystemNormal)
		{
			SystemNormal->Transparent(true);
		}
		else
		{
			printf("%s:%i - Error creating system font.\n", __FILE__, __LINE__);
		}
		
		SystemBold = SysFontType.Create();
		if (SystemBold)
		{
			SystemBold->Bold(true);
			SystemBold->Transparent(true);
			SystemBold->Create();
		}
		else
		{
			printf("%s:%i - Error creating bold version of the system font.\n", __FILE__, __LINE__);
		}
	}
	else
	{
		printf("%s:%i - Couldn't get system font setting.\n", __FILE__, __LINE__);
	}

	if (NOT SystemNormal)
	{
		LgiMsg(0, "Error: Couldn't create system font.", "Lgi Error: GApp::GApp", MB_OK);
		LgiExitApp();
	}
	
	if (NOT GetOption("noskin"))
	{
		// Load library
		char Name[64];
		#ifdef _DEBUG
		char *Build = "d";
		#else
		char *Build = "";
		#endif
		
		sprintf(Name, "lgiskin%s.%s", Build, LGI_LIBRARY_EXT);
		d->SkinLib = NEW(GLibrary(Name));
		if (d->SkinLib)
		{
			if (d->SkinLib->IsLoaded())
			{
				Proc_CreateSkinEngine CreateSkinEngine =
					(Proc_CreateSkinEngine)d->SkinLib->GetAddress(LgiSkinEntryPoint);
				if (CreateSkinEngine)
				{
					SkinEngine = CreateSkinEngine(this);
				}
			}
			else
			{
				DeleteObj(d->SkinLib);
			}
		}
	}
}

GApp::~GApp()
{
	DeleteObj(AppWnd);
	DeleteObj(SystemNormal);
	DeleteObj(SystemBold);
	
	DeleteObj(MouseHook);
	DeleteObj(d->FileSystem);
	DeleteObj(d->GdcSystem);
	DeleteObj(d->Config);
	DeleteObj(GFontSystem::Me);
	DeleteObj(d);
	TheApp = 0;
}

GApp *GApp::ObjInstance()
{
	return TheApp;
}

bool GApp::IsOk()
{
	bool Status = 	(this != 0) AND
					(d != 0)
					/*
					#ifdef XWIN
					AND (XDisplay() != 0)
					#endif
					*/
					;
					
	LgiAssert(Status);
	return Status;
}

GMouseHook *GApp::GetMouseHook()
{
	return MouseHook;
}

int GApp::GetMetric(int Metric)
{
	switch (Metric)
	{
		case LGI_MET_DECOR_X:
		{
			return 12;
		}
		case LGI_MET_DECOR_Y:
		{
			return 26;
		}
	}

	return 0;
}

GViewI *GApp::GetFocus()
{
	#if defined XWIN
	QWidget *w = QObject::QApp()->GetFocus();
	if (w)
	{
		return (GView*)w->_GetWndPtr();
	}
	#endif
	
	return 0;
}

OsThreadId GApp::GetGuiThread()
{
	#if defined XWIN
	return QApplication::GetGuiThread();
	#endif
	
	return 0;
}

OsProcessId GApp::GetProcessId()
{
	return getpid();
}

OsAppArguments *GApp::GetAppArgs()
{
	return IsOk() ? &d->Args : 0;
}

void GApp::SetAppArgs(OsAppArguments &AppArgs)
{
	if (IsOk())
	{
		d->Args = AppArgs;
	}
}

bool GApp::Run(bool Loop)
{
	if (Loop)
	{
		OnCommandLine();

		printf("Starting RunApplicationEventLoop...\n");
		RunApplicationEventLoop();
		printf("Finished RunApplicationEventLoop...\n");
	}
	else
	{
		EventRef theEvent;
		EventTargetRef theTarget;
		theTarget = GetEventDispatcherTarget();

		while
		(
			ReceiveNextEvent
			(
				0,
				NULL,
				0.001, // kEventDurationForever,
				true,
				&theEvent
			)
			==
			noErr
		)
		{
			SendEventToEventTarget (theEvent, theTarget);
			ReleaseEvent(theEvent);
		}	
	}

	return 0;
}

void GApp::OnReceiveFiles(GArray<char*> &Files)
{
	if (AppWnd)
	{
		AppWnd->OnReceiveFiles(Files);
	}
}

void GApp::SetConfig(GXmlTag *Tag)
{
	/*
	if (IsOk() AND Tag)
	{
		GXmlTag *Old = GetConfig(Tag->Name);
		if (Old)
		{
			d->Config->Values.Delete(Old);
			DeleteObj(Old);
		}

		if (NOT d->Config)
		{
			d->Config = NEW(Xml);
		}
		if (d->Config)
		{
			d->Config->Values.Insert(Tag);
		}
	}
	*/
}

GXmlTag *GApp::GetConfig(char *Tag)
{
	/*
	if (IsOk())
	{
		if (NOT d->Config)
		{
			char Path[256];
			if (LgiGetExePath(Path, sizeof(Path)))
			{
				if (Path[strlen(Path)-1] != DIR_CHAR) strcat(Path, DIR_STR);
				strcat(Path, "lgi.conf");
				if (FileExists(Path))
				{
					d->Config = NEW(Xml);
					if (d->Config)
					{
						if (d->Config->ParseXmlFile(Path) <= 0)
						{
							DeleteObj(d->Config);
						}
					}
				}
			}
		}

		if (Tag AND d->Config)
		{
			for (XmlTag *t=d->Config->Values.First(); t; t=d->Config->Values.Next())
			{
				if (t->Name AND
					stricmp(t->Name, Tag) == 0)
				{
					return t;
				}
			}
		}
	}
	*/

	return 0;
}

char *GApp::GetArgumentAt(int n)
{
	return n >= 0 AND n < d->Args.Args ? d->Args.Arg[n] : 0;
}

bool GApp::GetOption(char *Option, char *Dest, int DestLen)
{
	GArray<char> Buf;
	if (GetOption(Option, Buf))
	{
		if (Dest)
		{
			strsafecpy(Dest, &Buf[0], DestLen);
		}
		return true;
	}
	return false;
}

bool GApp::GetOption(char *Option, GArray<char> &Buf)
{
	if (IsOk() AND Option)
	{
		int OptLen = strlen(Option);
		for (int i=1; i<d->Args.Args; i++)
		{
			char *a = d->Args.Arg[i];

			if (strchr("-/\\", a[0]))
			{
				if (strncmp(a+1, Option, OptLen) == 0)
				{
					char *Arg = 0;
					if (strlen(a+1+OptLen) > 0)
					{
						Arg = a + 1 + OptLen;
					}
					else if (i < d->Args.Args - 1)
					{
						Arg = d->Args.Arg[i + 1];
					}
					
					if (Arg)
					{
						if (strchr("\'\"", *Arg))
						{
							char Delim = *Arg++;
							char *End = strchr(Arg, Delim);
							if (End)
							{
								int Len = (int)End-(int)Arg;
								if (Len > 0)
								{
									Buf.Length(Len + 1);
									memcpy(&Buf[0], Arg, Len);
									Buf[Len] = 0;
								}
								else return false;
							}
							else return false;
						}
						else
						{
							Buf.Length(strlen(Arg)+1);
							strcpy(&Buf[0], Arg);
						}
					}

					return true;
				}
			}
		}
	}

	return false;
}

void GApp::OnCommandLine()
{
	char WhiteSpace[] = " \r\n\t";

	GArray<char*> Files;

	for (int i=1; i<GetAppArgs()->Args; i++)
	{
		char *a = GetAppArgs()->Arg[i];
		if (FileExists(a))
		{
			Files.Add(NewStr(a));
		}
	}

	// call app
	if (Files.Length() > 0)
	{
		OnReceiveFiles(Files);
	}

	// clear up
	Files.DeleteArrays();
}

void GApp::Exit(int Code)
{
	if (Code)
	{
		// hard exit
		::exit(Code);
	}
	else
	{
		// soft exit
		QuitApplicationEventLoop();
	}
}


bool GApp::GetFileMimeType(char *File, char *Mime, int BufLen)
{
	bool Status = false;
	char Full[300] = "";

	if (NOT FileExists(File))
	{
		// Look in the path
		GToken p(getenv("PATH"), LGI_PATH_SEPARATOR);
		for (int i=0; i<p.Length(); i++)
		{
			LgiMakePath(Full, sizeof(Full), p[i], File);
			if (FileExists(Full))
			{
				File = Full;
				break;
			}
		}		
	}
	
	return Status;
}

bool GApp::GetAppsForMimeType(char *Mime, GArray<GAppInfo*> &Apps)
{
	// Find alternative version of the MIME type (e.g. x-type and type).
	char AltMime[256];
	strcpy(AltMime, Mime);
	char *s = strchr(AltMime, '/');
	if (s)
	{
		s++;
		
		int Len = strlen(s) + 1;
		if (strnicmp(s, "x-", 2) == 0)
		{
			memmove(s, s+2, Len - 2);
		}
		else
		{
			memmove(s+2, s, Len);
			s[0] = 'x';
			s[1] = '-';
		}
	}

	if (NOT d->MimeToApp.Length())
	{
		// printf("%s:%i - Building MimeToApp.\n", __FILE__, __LINE__);

	}
	
	AppArray *p = (AppArray*)d->MimeToApp.Find(Mime);
	if (p)
	{
		for (int i=0; i<p->Length(); i++)
		{
			Apps[i] = (*p)[i];
		}
		
		return true;
	}

	return false;
}
