/*
Copyright 2003,04 Aris Adamantiadis

This file is part of the SSH Library

The SSH Library is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or (at your
option) any later version.

The SSH Library is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
License for more details.

You should have received a copy of the GNU Lesser General Public License
along with the SSH Library; see the file COPYING.  If not, write to
the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
MA 02111-1307, USA. */

#ifndef _LIBSSH_H
#define _LIBSSH_H
#include <libssh/config.h>
#ifdef WIN32
	#define DIR_CHAR		'\\'
	#include <winsock2.h>
	#ifdef LIBSSH_EXPORTS
		#ifdef __cplusplus
		#define EXPORT extern "C" __declspec(dllexport)
		#else
		#define EXPORT __declspec(dllexport)
		#endif
	#else
		#ifdef __cplusplus
		#define EXPORT extern "C" __declspec(dllimport)
		#else
		#define EXPORT __declspec(dllimport)
		#endif
	#endif
#else
	#define DIR_CHAR		'/'
	#include <unistd.h>
	#include <sys/select.h> /* for fd_set * */
	#define EXPORT
#endif
#include <sys/types.h>
#define LIBSSH_VERSION "libssh-0.11"

#ifdef __cplusplus
extern "C" {
#endif

typedef struct string_struct STRING;
typedef struct buffer_struct BUFFER;
typedef struct public_key_struct PUBLIC_KEY;
typedef struct private_key_struct PRIVATE_KEY;
typedef struct ssh_options_struct SSH_OPTIONS;
typedef struct channel_struct CHANNEL;
typedef struct ssh_session SSH_SESSION;
typedef struct ssh_kbdint SSH_KBDINT;

/* integer values */
#ifdef WIN32
typedef unsigned char u8;
typedef unsigned short u16;
typedef unsigned int u32;
typedef unsigned __int64 u64;
#else
typedef u_int32_t u32;
typedef u_int16_t u16;
typedef u_int64_t u64;
typedef u_int8_t u8;
#endif

/* the offsets of methods */
#define KEX_ALGO 0
#define KEX_HOSTKEY 1
#define KEX_CRYPT_C_S 2
#define KEX_CRYPT_S_C 3
#define KEX_MAC_C_S 4
#define KEX_MAC_S_C 5
#define KEX_COMP_C_S 6
#define KEX_COMP_S_C 7
#define KEX_LANG_C_S 8
#define KEX_LANG_S_C 9

#define SSH_AUTH_SUCCESS 0
#define SSH_AUTH_DENIED 1
#define SSH_AUTH_PARTIAL 2
#define SSH_AUTH_INFO 3
#define SSH_AUTH_ERROR -1

#define SSH_SERVER_ERROR -1
#define SSH_SERVER_NOT_KNOWN 0
#define SSH_SERVER_KNOWN_OK 1
#define SSH_SERVER_KNOWN_CHANGED 2
#define SSH_SERVER_FOUND_OTHER 3

#ifndef MD5_DIGEST_LEN
    #define MD5_DIGEST_LEN 16
#endif
/* errors */

enum ssh_error {SSH_NO_ERROR, SSH_REQUEST_DENIED, SSH_INVALID_REQUEST, SSH_CONNECTION_LOST,SSH_FATAL,SSH_INVALID_DATA,SSH_EINTR};
EXPORT char *ssh_get_error(SSH_SESSION *session); /* returns a static char array */
enum ssh_error ssh_error_code(SSH_SESSION *session);
EXPORT void ssh_say(int priority,char *format,...);
EXPORT void ssh_set_verbosity(int num);

 /* There is a verbosity level */
 /* 3 : packet level */
 /* 2 : protocol level */
 /* 1 : functions level */
 /* 0 : important messages only */
 /* -1 : no messages */

/* in client.c */

EXPORT SSH_SESSION *ssh_connect(SSH_OPTIONS *options);
EXPORT void ssh_disconnect(SSH_SESSION *session);
EXPORT int ssh_service_request(SSH_SESSION *session,char *service);
EXPORT char *ssh_get_issue_banner(SSH_SESSION *session);
/* get copyright informations */
EXPORT const char *ssh_copyright();
/* string.h */

/* You can use these functions, they won't change */
/* makestring returns a newly allocated string from a char * ptr */
EXPORT STRING *string_from_char(char *what);
/* it returns the string len in host byte orders. str->size is big endian warning ! */
EXPORT int string_len(STRING *str);
EXPORT STRING *string_new(u32 size);
/* string_fill copies the data in the string. it does NOT check for boundary so allocate enough place with new_string */
/* right before */
EXPORT void string_fill(STRING *str,void *data,int len);
/* returns a newly allocated char array with the str string and a final nul caracter */
EXPORT char *string_to_char(STRING *str);
EXPORT STRING *string_copy(STRING *str);

/* deprecated */
EXPORT void ssh_crypto_init();

/* useful for debug */
EXPORT void ssh_print_hexa(char *descr,unsigned char *what, int len);
EXPORT void ssh_get_random(void *,int);

/* this one can be called by the client to see the hash of the public key before accepting it */
EXPORT int ssh_get_pubkey_hash(SSH_SESSION *session,char hash[MD5_DIGEST_LEN]);
EXPORT STRING *ssh_get_pubkey(SSH_SESSION *session);

/* deprecated */
EXPORT int pubkey_get_hash(SSH_SESSION *session,char hash[MD5_DIGEST_LEN]);

/* in connect.c */
EXPORT int ssh_fd_poll(SSH_SESSION *session);
EXPORT int ssh_select(CHANNEL **channels,CHANNEL **outchannels, int maxfd, fd_set *readfds, struct timeval *timeout);

EXPORT void publickey_free(PUBLIC_KEY *key);

/* in keyfiles.c */

EXPORT PRIVATE_KEY *privatekey_from_file(SSH_SESSION *session,char *filename,int type,char *passphrase);
EXPORT void private_key_free(PRIVATE_KEY *prv);
EXPORT STRING *publickey_from_file(char *filename,int *_type);
EXPORT STRING *publickey_from_next_file(SSH_SESSION *session,char **pub_keys_path,char **keys_path,
                            char **privkeyfile,int *type,int *count);
EXPORT int ssh_is_server_known(SSH_SESSION *session);
EXPORT int ssh_write_knownhost(SSH_SESSION *session);

/* in channels.c */

/* this one is deprecated */
EXPORT CHANNEL *open_session_channel(SSH_SESSION *session,int window,int maxpacket);
EXPORT CHANNEL *channel_open_forward(SSH_SESSION *session,char *remotehost, int remoteport, char *sourcehost, int localport);
EXPORT CHANNEL *channel_open_session(SSH_SESSION *session);
EXPORT void channel_free(CHANNEL *channel);
EXPORT int channel_request_pty(CHANNEL *channel);
EXPORT int channel_request_pty_size(CHANNEL *channel, char *term,int cols, int rows);
EXPORT int channel_change_pty_size(CHANNEL *channel,int cols,int rows);
EXPORT int channel_request_shell(CHANNEL *channel);
EXPORT int channel_request_subsystem(CHANNEL *channel, char *system);
EXPORT int channel_request_env(CHANNEL *channel,char *name, char *value);
EXPORT int channel_request_exec(CHANNEL *channel, char *cmd);
EXPORT int channel_request_sftp(CHANNEL *channel);
EXPORT int channel_write(CHANNEL *channel,void *data,int len);
EXPORT int channel_set_write_handler(CHANNEL *channel,
    void (*write_fct)(CHANNEL *channel, void *data, int len, void *userdefined),
    void *user);
EXPORT int channel_set_stderr_write_handler(CHANNEL *channel,
    void (*write_err_fct)(CHANNEL *channel, void *data, int len, void *userdefined),
    void *user);
EXPORT int channel_send_eof(CHANNEL *channel);
EXPORT int channel_read(CHANNEL *channel, BUFFER *buffer,int bytes,int is_stderr);
EXPORT int channel_poll(CHANNEL *channel, int is_stderr);
EXPORT int channel_close(CHANNEL *channel);
EXPORT int channel_read_nonblocking(CHANNEL *channel, char *dest, int len, int is_stderr);
EXPORT int channel_is_open(CHANNEL *channel);
/* in options.c */

EXPORT SSH_OPTIONS *options_new();
EXPORT SSH_OPTIONS *options_copy(SSH_OPTIONS *opt);
EXPORT int options_set_wanted_method(SSH_OPTIONS *opt,int method, char *list);
EXPORT void options_set_username(SSH_OPTIONS *opt,char *username);
EXPORT void options_set_port(SSH_OPTIONS *opt, unsigned int port);
EXPORT SSH_OPTIONS *ssh_getopt(int *argcptr, char **argv);
EXPORT void options_set_host(SSH_OPTIONS *opt, const char *host);
/* don't connect to host, use fd instead */
EXPORT void options_set_fd(SSH_OPTIONS *opt, int fd);
EXPORT void options_set_bindaddr(SSH_OPTIONS *opt, char *bindaddr);
EXPORT void options_set_identity(SSH_OPTIONS *opt, char *identity);
EXPORT void options_set_status_callback(SSH_OPTIONS *opt, void (*callback)(void *arg, float status), void *arg);
EXPORT void options_set_timeout(SSH_OPTIONS *opt, long seconds, long usec);
EXPORT void options_set_ssh_dir(SSH_OPTIONS *opt, char *dir);
EXPORT void options_set_known_hosts_file(SSH_OPTIONS *opt, char *dir);
/* buffer.c */

EXPORT BUFFER *buffer_new();
EXPORT void buffer_free(BUFFER *buffer);
/* buffer_get returns a pointer to the begining of the buffer. no position is taken into account */
EXPORT void *buffer_get(BUFFER *buffer);
/* same here */
EXPORT int buffer_get_len(BUFFER *buffer);


/* in auth.c */
/* these functions returns AUTH_ERROR is some serious error has happened,
  AUTH_SUCCESS if success,
  AUTH_PARTIAL if partial success,
  AUTH_DENIED if refused */
EXPORT int ssh_userauth_none(SSH_SESSION *session,char *username);
EXPORT int ssh_userauth_password(SSH_SESSION *session,char *username,char *password);
EXPORT int ssh_userauth_offer_pubkey(SSH_SESSION *session, char *username,int type, STRING *publickey);
EXPORT int ssh_userauth_pubkey(SSH_SESSION *session, char *username, STRING *publickey, PRIVATE_KEY *privatekey);
EXPORT int ssh_userauth_autopubkey(SSH_SESSION *session);
EXPORT int ssh_userauth_kbdint(SSH_SESSION *session, char *user, char *submethods);
EXPORT int ssh_userauth_kbdint_getnprompts(SSH_SESSION *session);
EXPORT char *ssh_userauth_kbdint_getname(SSH_SESSION *session);
EXPORT char *ssh_userauth_kbdint_getinstruction(SSH_SESSION *session);
EXPORT char *ssh_userauth_kbdint_getprompt(SSH_SESSION *session, int i, char *echo);
EXPORT void ssh_userauth_kbdint_setanswer(SSH_SESSION *session, unsigned int i, char *answer);

#ifdef WIN32

typedef struct {
      char    *pw_name;      /* user name */
      char    *pw_passwd;    /* user password */
      int     pw_uid;        /* user id */
      int     pw_gid;        /* group id */
      char    *pw_gecos;     /* real name */
      char    *pw_dir;       /* home directory */
      char    *pw_shell;     /* shell program */
} passwd;
EXPORT char *getpass(char *buf);
passwd *getpwent(void);
EXPORT int getuid();
EXPORT int geteuid();

#define R_OK	1
#define W_OK	2
#define X_OK	4
#define F_OK	8
EXPORT int access(const char *pathname, int mode);

int network_read(int fd, void *buf, size_t count);
int network_write(int fd, void *buf, size_t count);

extern char *optarg;
extern int optind, opterr, optopt;

#define snprintf	_snprintf
#define vsnprintf	_vsnprintf
#define strcasecmp	lstrcmpiA
#define strncasecmp	_strnicmp

#else

#define network_read	read
#define network_write	write

#endif

#ifdef __cplusplus
} ;
#endif
#endif /* _LIBSSH_H */
